/* imidiacy - Record a MIDI file from the computer keyboard 
version 2007.9.06 Copyright (C) 2006-2007 Andrew Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 59 
Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program produces standard MIDI files based on presses of the PC
keyboard.  The events Note Off, Note On, Aftertouch, and tempo change
are supported.  Output is to the monitor and soundcard or PC speaker.
 
For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/


/* Standard C includes */
#include <stdio.h>
#include <math.h>
#include <string.h>

/* library-dependent */
#include "allegro.h"
#include "midi_types.h"

/* Something to play back */ 
SAMPLE *waveform ; 

/* It's necessary to allocate a huge, static array of all the drum 
sounds.  Hope you've got RAM */
static SAMPLE *ptr_drum[NUM_DRUMSOUNDS] ;

/* A table of voice numbers, mapping MIDI note to the voice.  I don't 
know if voice numbers are allocated sequentially, this array may or may 
not be necessary */ 
int voice[128] ; 
float pitch[128] ;
int voice_count ;

/**********************************************************************/

/* play a wav sample, given a Midi note number and a velocity. */
void play_note(int note, int velocity) 
{
  float frequency ;

/* calculate the proper frequency if not already in the array.  This is 
a proportion by which to accelerate the playback, and not the real 
frequency of the musical tone. */

  if ( pitch[note] == 0 ) 
    pitch[note] = pow(2.0, (note - 69) / 12.0 ) ;
    
/* Turn off if still sounding */
  if ( voice[note] >= 0 ) {
    voice_stop(voice[note]) ;
    deallocate_voice(voice[note]) ;
  }

/* allocate voice and set the desired properties */
  voice[note] = allocate_voice(waveform) ;
  if (voice[note] >= 0) {
    frequency = pitch[note] * waveform->freq ; 
    voice_set_frequency(voice[note], frequency) ;

/* The Allegro volume settings run to 255 rather than 127 as in Midi */
    voice_set_volume(voice[note], velocity * 2) ;
    voice_start(voice[note]) ;
  }
  else {
    fprintf(stderr, "Couldn't allocate voice for MIDI note %d.\n", note) ;
  } /* End If voice allocation */
  
  return ;    
} 

/******************************************************************************/


/* Play a drum sound, given Midi note number and a velocity.  The sounds 
from freepats drumset 0 are used, regardless of what drum kit may have 
been selected. */ 

void play_drum(int note, int velocity) {
  int index ;
  
/* There are only 40 keys presently mapped to notes.  This means that 
only a selection of the 56 in the kit can represent drum sounds. */
/* Turn off if still sounding */
  if ( voice[note] >= 0 ) {
    voice_stop(voice[note]) ;
    deallocate_voice(voice[note]) ;
  }

/* allocate voice and set the desired properties.  Here 25 is the lowest 
drum in freepats.  The relation between the note numbers and the indices 
in the voice array is irregular */
  if ( (note >= 25) && (note <= 27) ) {
    index = note - 25 ;
  }
  else if ( (note >= 31) && (note <= 82) ) {
    index = note - 28 ;
  }
  else if (note == 84) {
    index = 55 ;
  }
  else {
    return ;  /* No sound available for this note */
  } 

/* F'n allocate_voice barfs on bad input, so make sure that it does not 
get sent a null pointer */
  if (!ptr_drum[index]) {
    fprintf(stderr, "No sample loaded for note %d.\n", note) ;
    return ;
  }
  
  if ( (voice[note] = allocate_voice(ptr_drum[index])) >= 0) {
/* The Allegro volume settings run to 255 rather than 127 as in Midi */
    voice_set_volume(voice[note], velocity * 2) ;
    voice_start(voice[note]) ;
  }
  else {
    fprintf(stderr, "Couldn't allocate voice for MIDI drum %d.\n", note) ;
  } /* End If voice allocation */
  
  return ;    
} /* End Function play_drum */


/******************************************************************************/

/* Turn off a sounding sample, deallocate */
void hush_note(int note) 
{  
  if ( voice[note] >= 0) {
    voice_stop(voice[note]) ;
    deallocate_voice(voice[note]) ;
    voice[note] = -1 ;
  }
} 

/******************************************************************************/
/* change the volume of a sounding voice */
void adjust_volume(int note, int velocity)
{
  if (voice[note] >= 0) 
    voice_set_volume(voice[note], velocity * 2) ;
  else
    fprintf(stderr, "adjust_volume: voice not allocated") ;
}

/******************************************************************************/

/* Initialize the sound driver and load a sample, depending on the 
selected instrument.  */

int init_sound(int patch, int channel) 
{
  char *filename ;
  char pathname[80] ;

  if (install_sound(DIGI_AUTODETECT, MIDI_NONE, "") != 0) {
    fprintf(stderr, "Couldn't set a sound driver\nAllegro message: %s\n", 
             allegro_error) ;
/* Letting negative one denote NO sound */
    return -1 ;
  }

  if (channel == 9) {
    char prefix[40] ;
    bool ok_directory = FALSE ; 
    int i ; 

/* Try various directories for the presence of the wav files */
    strcpy(prefix, "audiodata/") ;

/* Destination from after changing the prefix */
SeekFiles: 
    strcpy(pathname, prefix) ;
    strcat(pathname, "snare_roll.wav") ;
    ptr_drum[0] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "snap.wav") ;
    ptr_drum[1] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "high_q.wav") ;
    ptr_drum[2] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "sticks.wav") ;
    ptr_drum[3] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "square_click.wav") ;
    ptr_drum[4] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "metronome_click.wav") ;
    ptr_drum[5] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "metronome_bell.wav") ;
    ptr_drum[6] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "kick_1.wav") ;
    ptr_drum[7] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "kick_2.wav") ;
    ptr_drum[8] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "stick_rim.wav") ;
    ptr_drum[9] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "snare_1.wav") ;
    ptr_drum[10] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "clap_hand.wav") ;
    ptr_drum[11] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "snare_2.wav") ;
    ptr_drum[12] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "tom_low_2.wav") ;
    ptr_drum[13] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "hi_hat_closed.wav") ;
    ptr_drum[14] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "tom_low_1.wav") ;
    ptr_drum[15] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "hi_hat_pedal.wav") ;
    ptr_drum[16] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "tom_mid_2.wav") ;
    ptr_drum[17] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "hi_hat_open.wav") ;
    ptr_drum[18] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "tom_mid_1.wav") ;
    ptr_drum[19] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "tom_high_2.wav") ;
    ptr_drum[20] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "cymbal_crash_1.wav") ;
    ptr_drum[21] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "tom_high_1.wav") ;
    ptr_drum[22] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "cymbal_ride_1.wav") ;
    ptr_drum[23] = load_sample(pathname) ;
 
    strcpy(pathname, prefix) ;
    strcat(pathname, "cymbal_chinese.wav") ;
    ptr_drum[24] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "cymbal_ride_bell.wav") ;
    ptr_drum[25] = load_sample(pathname) ;
 
    strcpy(pathname, prefix) ;
    strcat(pathname, "tombourine.wav") ;
    ptr_drum[26] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "cymbal_splash.wav") ;
    ptr_drum[27] = load_sample(pathname) ;
 
    strcpy(pathname, prefix) ;
    strcat(pathname, "cow_bell.wav") ;
    ptr_drum[28] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "cymbal_crash_2.wav") ;
    ptr_drum[29] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "vibra_slap.wav") ;
    ptr_drum[30] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "cymbal_ride_2.wav") ;
    ptr_drum[31] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "bongo_high.wav") ;
    ptr_drum[32] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "bongo_low.wav") ;
    ptr_drum[33] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "conga_high_1_mute.wav") ;
    ptr_drum[34] = load_sample(pathname) ;
 
    strcpy(pathname, prefix) ;
    strcat(pathname, "conga_high_2_open.wav") ;
    ptr_drum[35] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "conga_low.wav") ;
    ptr_drum[36] = load_sample(pathname) ;
 
    strcpy(pathname, prefix) ;
    strcat(pathname, "timbale_high.wav") ;
    ptr_drum[37] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "timbale_low.wav") ;
    ptr_drum[38] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "agogo_high.wav") ;
    ptr_drum[39] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "agogo_low.wav") ;
    ptr_drum[40] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "cabasa.wav") ;
    ptr_drum[41] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "maracas.wav") ;
    ptr_drum[42] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "whistle_1_high_short.wav") ;
    ptr_drum[43] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "whistle_2_low_long.wav") ;
    ptr_drum[44] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "guiro_1_short.wav") ;
    ptr_drum[45] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "guiro_2_long.wav") ;
    ptr_drum[46] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "claves.wav") ;
    ptr_drum[47] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "wood_block_1.wav") ;
    ptr_drum[48] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "wood_block_2.wav") ;
    ptr_drum[49] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "cuica_1_mute.wav") ;
    ptr_drum[50] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "cuica_2_open.wav") ;
    ptr_drum[51] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "triangle_1_mute.wav") ;
    ptr_drum[52] = load_sample(pathname) ;

    strcpy(pathname, prefix) ;
    strcat(pathname, "triangle_2_open.wav") ;
    ptr_drum[53] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "shaker.wav") ;
    ptr_drum[54] = load_sample(pathname) ;
  
    strcpy(pathname, prefix) ;
    strcat(pathname, "belltree.wav") ;
    ptr_drum[55] = load_sample(pathname) ;

/* Check if ANY of these files were found */
    for (i = 0 ; i < NUM_DRUMSOUNDS ; ++i) 
      ok_directory = ok_directory || ptr_drum[i] ;

    if (!ok_directory) {
      if (!strncmp(prefix, "audiodata/", 11) ) {
        strcpy(prefix, "imidiacy/audiodata/") ;
        goto SeekFiles ;
      }
      else if (!strncmp(prefix, "imidiacy/audiodata/", 20) ) {
        strcpy(prefix, "") ;
        goto SeekFiles ; 
      }
/* Give up and return a failure value */
      else {
        return -2 ;
      }
    } /* End if (!ok_directory) */
  }

/* Not a drum channel */
  else {
    switch (patch) {
      case 0 :
        filename = "acpiano.wav" ;
        break ;
      case 1 :
        filename = "britepno.wav" ;
        break ;
      case 2 :
        filename = "synpiano.wav" ;
        break ;
      case 3 :
        filename = "honky.wav" ;
        break ;
      case 4 :
        filename = "epiano1.wav" ;
        break ;
      case 5 :
        filename = "epiano2.wav" ;
        break ;
      case 6 :
        filename = "hrpschrd.wav" ;
        break ;
      case 7 :
        filename = "clavinet.wav" ;
        break ;
      case 8 :
        filename = "celeste.wav" ;
        break ;
      case 9 :
        filename = "glocken.wav" ;
        break ;
      case 10 :
        filename = "musicbox.wav" ;
        break ;
      case 11 :
        filename = "vibes.wav" ;
        break ;
      case 12 :
        filename = "marimba.wav" ;
        break ;
      case 13 :
        filename = "xylophon.wav" ;
        break ;
      case 14 :
        filename = "tubebell.wav" ;
        break ;
      case 15 :
        filename = "santur.wav" ;
        break ;
      case 16 :
        filename = "hammond.wav" ;
        break ;
      case 17 :
        filename = "percorg.wav" ;
        break ;
      case 18 :
        filename = "rockorg.wav" ;
        break ;
      case 19  :
        filename = "church.wav" ;
        break ;
      case 20 :
        filename = "reedorg.wav" ;
        break ;
      case 21 :
        filename = "accordn.wav" ;
        break ;
      case 22 :
        filename = "harmonca.wav" ;
        break ;
      case 23 :
        filename = "concrtna.wav" ;
        break ;
      case 24 :
        filename = "nyguitar.wav" ;
        break ;
      case 25 :
        filename = "acguitar.wav" ;
        break ;
      case 26 :
        filename = "jazzgtr.wav" ;
        break ;
      case 27 :
        filename = "cleangtr.wav" ;
        break ;
      case 28 :
        filename = "mutegtr.wav" ;
        break ;
      case 29 :
        filename = "odguitar.wav" ;
        break ;
      case 30 :
        filename = "distgtr.wav" ;
        break ;
      case 31 :
        filename = "gtrharm.wav" ;
        break ;
      case 32 :
        filename = "acbass.wav" ;
        break ;
      case 33 :
        filename = "fngrbass.wav" ;
        break ;
      case 34 :
        filename = "pickbass.wav" ;
        break ;
      case 35 :
        filename = "fretless.wav" ;
        break ;
      case 36 :
        filename = "slapbas1.wav" ;
        break ;
      case 37 :
        filename = "slapbas2.wav" ;
        break ;
      case 38 :
        filename = "synbass1.wav" ;
        break ;
      case 39 :
        filename = "synbass2.wav" ;
        break ;
      case 40 :
        filename = "violin.wav" ;
        break ;
      case 41 :
        filename = "viola.wav" ;
        break ;
      case 42 :
        filename = "cello.wav" ;
        break ;
      case 43 :
        filename = "contraba.wav" ;
        break ;
      case 44 :
        filename = "tremstr.wav" ;
        break ;
      case 45 :
        filename = "pizzcato.wav" ;
        break ;
      case 46 :
        filename = "harp.wav" ;
        break ;
      case 47 :
        filename = "timpani.wav" ;
        break ;
      case 48 :
        filename = "marcato.wav" ;
        break ;
      case 49 :
        filename = "slowstr.wav" ;
        break ;
      case 50 :
        filename = "synstr1.wav" ;
        break ;
      case 51 :
        filename = "synstr2.wav" ;
        break ;
      case 52 :
        filename = "choir.wav" ;
        break ;
      case 53 :
        filename = "doo.wav" ;
        break ;
      case 54 :
        filename = "voices.wav" ;
        break ;
      case 55 :
        filename = "orchhit.wav" ;
        break ;
      case 56 :
        filename = "trumpet.wav" ;
        break ;
      case 57 :
        filename = "trombone.wav" ;
        break ;
      case 58 :
        filename = "tuba.wav" ;
        break ;
      case 59 :
        filename = "mutetrum.wav" ;
        break ;
      case 60 :
        filename = "frenchrn.wav" ;
        break ;
      case 61 :
        filename = "hitbrass.wav" ;
        break ;
      case 62 :
        filename = "synbras1.wav" ;
        break ;
      case 63 :
        filename = "synbras2.wav" ;
        break ;
      case 64 :
        filename = "sprnosax.wav" ;
        break ;
      case 65 :
        filename = "altosax.wav" ;
        break ;
      case 66 :
        filename = "tenorsax.wav" ;
        break ;
      case 67 :
        filename = "barisax.wav" ;
        break ;
      case 68 :
        filename = "oboe.wav" ;
        break ;
      case 69 :
        filename = "englhorn.wav" ;
        break ;
      case 70 :
        filename = "bassoon.wav" ;
        break ;
      case 71 :
        filename = "clarinet.wav" ;
        break ;
      case 72 :
        filename = "piccolo.wav" ;
        break ;
      case 73 :
        filename = "flute.wav" ;
        break ;
      case 74 :
        filename = "recorder.wav" ;
        break ;
      case 75 :
        filename = "woodflut.wav" ;
        break ;
      case 76 :
        filename = "bottle.wav" ;
        break ;
      case 77 :
        filename = "shakazul.wav" ;
        break ;
      case 78 :
        filename = "whistle.wav" ;
        break ;
      case 79 :
        filename = "ocarina.wav" ;
        break ;
      case 80 :
        filename = "sqrwave.wav" ;
        break ;
      case 81 :
        filename = "sawwave.wav" ;
        break ;
      case 82 :
        filename = "calliope.wav" ;
        break ;
      case 83 :
        filename = "chiflead.wav" ;
        break ;
      case 84 :
        filename = "charang.wav" ;
        break ;
      case 85 :
        filename = "voxlead.wav" ;
        break ;
      case 86 :
        filename = "lead5th.wav" ;
        break ;
      case 87 :
        filename = "basslead.wav" ;
        break ;
      case 88 :
        filename = "newage.wav" ;
        break ;
      case 89 :
        filename = "warmpad.wav" ;
        break ;
      case 90 :
        filename = "polysyn.wav" ;
        break ;
      case 91 :
        filename = "ghostie.wav" ;
        break ;
      case 92 :
        filename = "bowglass.wav" ;
        break ;
      case 93 :
        filename = "metalpad.wav" ;
        break ;
      case 94 :
        filename = "halopad.wav" ;
        break ;
      case 95 :
        filename = "sweeper.wav" ;
        break ;
      case 96 :
        filename = "aurora.wav" ;
        break ;
      case 97 :
        filename = "soundtrk.wav" ;
        break ;
      case 98 :
        filename = "crystal.wav" ;
        break ;
      case 99 :
        filename = "atmosphr.wav" ;
        break ;
      case 100 :
        filename = "freshair.wav" ;
        break ;
      case 101 :
        filename = "unicorn.wav" ;
        break ;
      case 102 :
        filename = "echovox.wav" ;
        break ;
      case 103 :
        filename = "startrak.wav" ;
        break ;
      case 104 :
        filename = "sitar.wav" ;
        break ;
      case 105 :
        filename = "banjo.wav" ;
        break ;
      case 106 :
        filename = "shamisen.wav" ;
        break ;
      case 107 :
        filename = "koto.wav" ;
        break ;
      case 108 :
        filename = "kalimba.wav" ;
        break ;
      case 109 :
        filename = "bagpipes.wav" ;
        break ;
      case 110 :
        filename = "fiddle.wav" ;
        break ;
      case 111 :
        filename = "shannai.wav" ;
        break ;
      case 112 :
        filename = "carillon.wav" ;
        break ;
      case 113 :
        filename = "agogo.wav" ;
        break ;
      case 114 :
        filename = "steeldrm.wav" ;
        break ;
      case 115 :
        filename = "woodblk.wav" ;
        break ;
      case 116 :
        filename = "taiko.wav" ;
        break ;
      case 117 :
        filename = "toms.wav" ;
        break ;
      case 118 :
        filename = "syntom.wav" ;
        break ;
      case 119 :
        filename = "revcym.wav" ;
        break ;
      case 120 :
        filename = "fx-fret.wav" ;
        break ;
      case 121 :
        filename = "fx-blow.wav" ;
        break ;
      case 122 :
        filename = "seashore.wav" ;
        break ;
      case 123 :
        filename = "jungle.wav" ;
        break ;
      case 124 :
        filename = "telephon.wav" ;
        break ;
      case 125 :
        filename = "helicptr.wav" ;
        break ;
      case 126 : 
        filename = "applause.wav" ;
        break ;
      case 127 :
        filename = "pistol.wav" ;
        break ;
      default:
        fprintf(stderr, "patch out of bounds\n") ;
        filename = "acpiano.wav" ;
    } /* End Switch */ 

/* Load the sample, trying different locations */
    strcpy(pathname, "audiodata/") ;
    strcat(pathname, filename) ;
    waveform = load_sample(pathname) ;
    if (!waveform) {
      waveform = load_sample(filename) ;
      if (!waveform) {
        strcpy(pathname, "imidiacy/audiodata/") ;
        strcat(pathname, filename) ;
        waveform = load_sample(pathname) ;
        if (!waveform) {
          fprintf(stderr, "Couldn't load wave file\n") ;
          return -2 ;
        }
      }
    }
  } /* End If Drum or Instrument channel */

/* Initialize the voice[] array to negative to denote that voices have 
not been allocated */
  { 
    int i ;
    for (i = 0 ; i < 128 ; ++i) 
      voice[i] = -1 ;
  }

  return 0 ;  /* denotes success */
}

/******************************************************************************/

/* Free memory holding the audio sample */
void end_sound() 
{ 
  int i ;

  if (waveform) {
    destroy_sample(waveform) ;
  }

  for (i = 0 ; i < NUM_DRUMSOUNDS ; ++i) {
    if (ptr_drum[i]) {
      destroy_sample(ptr_drum[i]) ;
    }
  } 
}

/******************************************************************************/

