/* imidiacy - Record a MIDI file from the computer keyboard 
version 2007.9.06 Copyright (C) 2006-2007 Andrew Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 59 
Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program produces standard MIDI files based on presses of the PC
keyboard.  The events Note Off, Note On, Aftertouch, and tempo change
are supported.  Output is to the monitor and soundcard or PC speaker.
 
For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "allegro.h"
#include "midi_types.h"

/* Global variables */
volatile unsigned long counter ;
bool h ;  /* State of the harmony-mode checkbox */

/* Represents the y position of the beginning and end of the roll punch 
(vertical line) of each note.  This array contains heads of linked 
lists. */ 
struct punchtype *punch[40] ;

/**********************************************************************/
/* version dependent code to do not much */
void prepare_timing()
{
  LOCK_FUNCTION(timer) ;
  install_int_ex(timer, BPS_TO_TIMER(TIMER_FREQUENCY)) ;
}

/**********************************************************************/
/* to be called at the end of record() */
void end_timing() 
{
  remove_int(timer) ;
}

/**********************************************************************/

/* conserve CPU cycles */
void delay(unsigned int milliseconds) 
{
  rest(milliseconds) ;
}

/**********************************************************************/
void legal(char *string)
{
  int choice ;
  choice = alert (string, 
                 "Free according to the GNU General Public License", 
                 "There is ABSOLUTELY NO WARRANTY",
                 "&OK", "&Quit", 
                  'o', 'q') ;
 
  if (choice == 2) exit(EXIT_SUCCESS) ; 
} /* End Function legal */

/**********************************************************************/
/* start a new line in the piano-roll representation of the sound */

void add_note(int note, int velocity, int key_sig, int channel) 
{
  (void) key_sig ;              /* not needed for this representation */
  int x, y, note_shift ;
  int color ;
  struct punchtype *ptr_present, *ptr_new ;

/* Keep in mind that note, presently, may run only from 57 to 96, or on 
a drum track, from LOW_DRUM to LOW_DRUM + 40.  Then, remember that 
SCROLL_MARGIN is redefined as a percentage of the screen width, rather 
than a number of pixels, so to be adjustible to different resolutions.  
I'm going to go ahead and hard-code the 40.  Keep in mind its meaning! 
*/
  note_shift = (channel != 9) ? LOW_NOTE : LOW_DRUM ;
  x = ((note - note_shift) / 40.F ) * ((100.F - 2.F * SCROLL_MARGIN) * (SCREEN_W / 100.F)) + (SCROLL_MARGIN * SCREEN_W / 100.F) ;
  y = SCREEN_H - 1 ;

/*  This is intended to become more red as the volume increases */
  color = makecol(2 * velocity, 128, 255 - (2 * velocity) ) ;

/* Bail out if these coordinates are bad. */
  assert ( (x >= 0) && (x < SCREEN_W) && (y >= 0) && (y < SCREEN_H) ) ;
  putpixel(screen, x, y, color) ;
  
/* Start a new punch node */
/* Allocate and connect the node */
  ptr_new = (struct punchtype *)malloc(sizeof(struct punchtype)) ;
  if (ptr_new == NULL) {
    perror("Memory allocation error.") ;
    exit(EXIT_FAILURE) ;
  }

  ptr_present = punch[note - note_shift]  ;
  if (ptr_present == NULL) {
    punch[note - note_shift] = ptr_new ;
    ptr_present = ptr_new ;
    ptr_present -> ptr_next = NULL ;
  }
  else {
//    fprintf(stderr, "This shouldn't execute the first note.\n") ;
    while ( ptr_present -> ptr_next != NULL) {
      ptr_present = ptr_present -> ptr_next ;
    }
    ptr_present -> ptr_next = ptr_new ;
    ptr_new -> ptr_next = NULL ;
  }  /* End If */

  /* Put some values in the node */
  ptr_new -> y_head = y ;
  ptr_new -> y_tail = 0 ;
  ptr_new -> ptr_next = NULL ;

}  /* End function add_note */

/**********************************************************************/
/* Put a black pixel at the base, to stop drawing a note */
void del_note(int note, int velocity, int channel)
{ 
  int x, y, note_shift ;
  int color ;
  struct punchtype *ptr_present ;

/* use the exact same locator formula, to make sure it's the correct 
pixel being changed. */ 

  note_shift = (channel != 9 ) ? LOW_NOTE : LOW_DRUM ;
  x = ((note - note_shift) / 40.F ) * ((100.F - 2.F * SCROLL_MARGIN) * (SCREEN_W / 100.F)) + (SCROLL_MARGIN * SCREEN_W / 100.F) ;
  y = SCREEN_H - 1 ;
  color = makecol(0, 0, 0) ;      /* Black */

/* Bail out if these coordinates are bad */
  assert ( (x >= 0) && (x < SCREEN_W) && (y >= 0) && (y < SCREEN_H) ) ;
  putpixel(screen, x, y, color) ;

/* Find the last node in the linked list */
  ptr_present = punch[note - note_shift]  ;
//  fprintf(stderr, "punch[] has the value %X \n", punch[note-array_offset]) ;   /* A node should always have been allocated */  
  while ( ptr_present -> ptr_next != NULL) {
    ptr_present = ptr_present -> ptr_next ;
  }  /* Wend */

  /* Put some values in the node */
  ptr_present -> y_tail = y ;

}  /* End function del_note() */

/**********************************************************************/ 
/* Move pixels in the graphical display up by one. */

void scroll_roll() 
{ 

/* This function changes only the pixels in the piano roll view which
have to be moved, which is faster than blitting a large portion of the
screen.  It does not allow for color changes on the same line, but
aftertouch events are not supported by the cheap synthesizers anyway. */

  int i, x, color ;
  struct punchtype *ptr_present ;

  for (i = 0 ; i < 40 ; ++i) {
    ptr_present = punch[i] ;

    while (ptr_present != NULL) {

/* Copy a pixel up one for the head of a punch */
      if (ptr_present -> y_head > ((STATUS_LINES +1) * text_height(font)) ) {
        x = (i / 40.F) * ((100.F - 2.F * SCROLL_MARGIN) * (SCREEN_W / 100.F)) + (SCROLL_MARGIN * SCREEN_W / 100.F) ;
        color = getpixel(screen, x, ptr_present -> y_head ) ; 
        --(ptr_present -> y_head) ;
        putpixel(screen, x, ptr_present -> y_head, color) ;
      } /* End If */

/* Add a black pixel at the tail of a punch */ 
      if (ptr_present -> y_tail > ((STATUS_LINES +1) * text_height(font)) ) {
        x = (i / 40.F) * ((100.F - 2.F * SCROLL_MARGIN) * (SCREEN_W / 100.F)) + (SCROLL_MARGIN * SCREEN_W / 100.F) ;
        color = makecol(0, 0, 0) ;    /* Black */
        --(ptr_present -> y_tail) ;
        putpixel(screen, x, ptr_present -> y_tail, color) ;
      }

/* Free a node if its tail has scrolled out of bounds, remembering zero 
indicates that the note has not ended yet */
      else if (ptr_present -> y_tail > 0) {
/* Only the oldest node should be erased */
//        fprintf(stderr, "ptr_present=%X should equal punch[i]=%X\n", ptr_present, punch[i]) ;
        punch[i] = ptr_present -> ptr_next ;
        free(ptr_present) ;
      } /* End If */

/* Advance to the next node */
      ptr_present = ptr_present -> ptr_next ;
    } /* Wend (ptr_present != NULL) */
  }  /* Next x position */

} /* End Function pscroll() */

/**********************************************************************/ 
/* adjust the position of the slider indicator for tempo or velocity */
void slider(int linenumber, int old_setting, int new_setting) 
{
  int x, y, left_width, right_width, range, min_setting ;
  float quotient ;

  if (linenumber == TEMPO_LINE) 
  {
     left_width =  text_length(font, "Adagio <") ;
     right_width = text_length(font, "> Allegro") ;
     range = MAX_BPM - MIN_BPM ;
     min_setting = MIN_BPM ;
  }
  else if (linenumber == VELOCITY_LINE)
  {
    left_width = text_length(font, "Piano <") ;
    right_width = text_length(font, "> Forte") ;
    range = 127 - 0 ;
    min_setting = 0 ;
  }
  else return ;

/* Erase the old setting mark */
  quotient = (float)(old_setting - min_setting) / (float)range ;
  x = left_width + (int)((float)(SCREEN_W - (left_width + right_width) - 1) * quotient) ;
  y = text_height(font) * (linenumber + 1) ;

/* Bounds check */
  assert ( (x >= 0) && (x < SCREEN_W) && (y >= 0) && (y < SCREEN_H) ) ;
  textout_ex(screen, font, "-", x, y, makecol(128, 128, 128), makecol(0, 0, 0)) ;

/* Draw a new setting mark */
  quotient = (float)(new_setting - min_setting) / (float)range ;
  x = left_width + (int)((float)(SCREEN_W - (left_width + right_width) - 1) * quotient) ;
  y  = text_height(font) * (linenumber + 1) ;

  if ( (x >= 0) && (x < SCREEN_W) && (y >= 0) && (y < SCREEN_H) )
    textout_ex(screen, font, "|", x, y, makecol(128, 128, 128), makecol(0, 0, 0)) ;
  
} /* End Function slider */

/**********************************************************************/ 
/* Adjust the position of the legato slider */ 

void sliderf(int linenumber, float old_setting, float new_setting) 
{
  int x, y, left_width, right_width ;
  float range, min_setting ;

  if (linenumber == LEGATO_LINE) 
  {
     left_width =  text_length(font, "Staccato <") ;
     right_width = text_length(font, "> Legato") ;
     range = 1.0 - 0.5 ;
     min_setting = 0.5 ;
  }
  else return ;

/* Erase the old setting mark */
  x = left_width + (int)((float)(SCREEN_W - (left_width + right_width) - 1) * ( (old_setting - min_setting) / range) );
  y  = text_height(font) * (linenumber + 1) ;

/* Bounds check */
  assert( (x >= 0) && (x < SCREEN_W) && (y >= 0) && (y < SCREEN_H) ) ;
  textout_ex(screen, font, "-", x, y, makecol(128, 128, 128), makecol(0, 0, 0)) ;

/* Draw a new setting mark */
  x = left_width + (int)((float)(SCREEN_W - (left_width + right_width) - 1) * ( (new_setting - min_setting) / range) );
  y  = text_height(font)* (linenumber + 1) ;

  if ( (x >= 0) && (x < SCREEN_W) && (y >= 0) && (y < SCREEN_H) )
    textout_ex(screen, font, "|", x, y, makecol(128, 128, 128), makecol(0, 0, 0)) ;
  
  return ;
} /* End Function sliderf() */

/**********************************************************************/ 
/* Set up the main display for recording, keeping it simple for now */

void draw_display(int bpm, int velocity, float legato, char *version_string)
{
  int i ;
  char help_line[] = 
"Backspace=End Up/Down=Volume Left/Right=Tempo Home/End=Legato PgUp/PgDn=f/p." ;
  struct punchtype *ptr_present, *ptr_old ;

/* Clear the mess off the screen */
  clear_bitmap(screen) ;

/* print the version and help strings */
  textout_ex(screen, font, version_string, 5, 1 * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;
  textout_ex(screen, font, help_line, 5, 2 * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;

/* Print the Tempo line */
  for (i = 0 ; i < SCREEN_W ; i = i + text_length(font, "-") )
  {
    textout_ex(screen, font, "-", i, (TEMPO_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;
  }
  textout_ex(screen, font, "Adagio <", 0, (TEMPO_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;
  textout_right_ex(screen, font, "> Allegro", SCREEN_W - 1, (TEMPO_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;

/* Print the Velocity line */
  for (i = 0 ; i < SCREEN_W ; i = i + text_length(font, "-") )
  {
    textout_ex(screen, font, "-", i, (VELOCITY_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;
  }
  textout_ex(screen, font, "Piano <", 0, (VELOCITY_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;
  textout_right_ex(screen, font, "> Forte", SCREEN_W - 1, (VELOCITY_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;

/* Print the Legato line */
  for (i = 0 ; i < SCREEN_W ; i = i + text_length(font, "-") )
  {
    textout_ex(screen, font, "-", i, (LEGATO_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;
  }
  textout_ex(screen, font, "Staccato <", 0, (LEGATO_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;
  textout_right_ex(screen, font, "> Legato", SCREEN_W - 1, (LEGATO_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;

/* Print a metronome line */
  textout_ex(screen, font, "Metronome:", 0, (METRONOME_LINE + 1) * text_height(font), makecol(128, 128, 128), makecol(0, 0, 0)) ;

/* Set the initial slider positions */
  slider(TEMPO_LINE, bpm, bpm) ;
  slider(VELOCITY_LINE, velocity, velocity) ;
  sliderf(LEGATO_LINE, legato, legato) ;

/* Set the timer back to zero (only necessary if many takes are done) */
  counter = 0 ;

/* Clear out the piano-roll punches from any previous takes */
  for (i = 0 ; i < 40 ; ++i) {
    ptr_present = punch[i] ;
    if (ptr_present == NULL) 
      continue ;

/* Advance along the linked list */
    while (ptr_present != NULL) {
      ptr_old = ptr_present ;
      ptr_present = ptr_present -> ptr_next ;
      free(ptr_old) ;
    } /* Wend */
    punch[i] = NULL ;
  } /* Next note */

} /* End Function draw_display() */

/**********************************************************************/

char *chan_getter(int index, int *size)
{
  static char *strings[] = {"0", "1", "2", "3", "4", "5", "6", "7", "8", 
"9 (Drums)", "10", "11", "12", "13", "14", "15"} ;

  if (index < 0) 
  {
    *size = 16 ;
    return NULL ;
  }
  else 
    return strings[index] ;
} /* End Function *chan_getter() */

/**********************************************************************/
char *ksig_getter(int index, int *size)
{
  char *strings[] = {"7 Flats - Cb", "6 Flats - Gb", "5 Flats - Db", "4 Flats - Ab", "3 Flats - Eb", "2 Flats - Bb", "1 Flat - F", "Natural - C", "1 Sharp - G", "2 Sharps - D", "3 Sharps - A", "4 Sharps - E", "5 Sharps - B", "6 Sharps - F#", "7 Sharps - C#"} ;

  if (index < 0)
  { 
    *size = 15 ;
    return NULL ;
  }
  else
    return strings[index] ;
} /* End Function *ksig_getter() */

/**********************************************************************/
char *tsig_getter(int index, int *size)
{
  static char *strings[] = {"2/4", "3/4", "4/4", "5/4", "6/4", "7/4"} ;

  if (index < 0)
  {
    *size = 6 ;
    return NULL ;
  }
  else
    return strings[index] ;
} /* End Function tsig_getter() */

/**********************************************************************/

/* Ask user:  keep the take? */
int query_keep() 
{ 
  bool retval = TRUE ;
  unsigned char a ;

  clear_bitmap(screen) ;
  clear_keybuf() ;
  textout_ex(screen, font, "Keep this take?  (Y/N) ", 0.10 * SCREEN_W, 0.10 * SCREEN_H, makecol(128, 128, 128), -1) ;
  do {
    rest(100) ; /* Save processer cycles */
    if (keyboard_needs_poll() == TRUE ) {
      if (poll_keyboard() != 0) fprintf(stderr, "Couldn't read keyboard.\n") ;
    }

    if ( key[KEY_Y] ) 
    {
      a = 'Y' ;
      retval = TRUE ;
    }
    else if ( key[KEY_N] ) 
    {
      a = 'N' ;
      retval = FALSE ;
    }
    else 
      a = '\0' ;
  } while (!a) ;

  return retval ;
}  /* End Function query_keep */

/**********************************************************************/

/* return 0 for quit, 1 for another take, and 2 to change settings */
int query_another() 
{
  unsigned char a ;
  int retval = 2 ;

  textout_ex(screen, font, "Another take, change settings, or quit?  (T/S/Q) ", 0.10 * SCREEN_W, 0.20 * SCREEN_H, makecol(128,128, 128), -1) ;
  do {
    rest(100) ; /* Don't waste processer cycles */
    if (keyboard_needs_poll() == TRUE ) {
      if (poll_keyboard() != 0) fprintf(stderr, "Couldn't read keyboard.\n") ;
    }

    if ( key[KEY_T] ) {
      a = 'T' ;
      retval = 1 ;
    }
    else if ( key[KEY_S] ) {
      a = 'S' ;
      retval = 2 ;
    }
    else if ( key[KEY_Q] ) {
      a = 'Q' ;
      retval = 0 ;
    }
    else {
      a = '\0' ;
    }  /* End If */
  } while (!a) ;
  
  return retval ;
} /* End Function query_another */

/**********************************************************************/
char *drum_getter(int index, int *size)
{

/* These are names of different drum kits which are not necessarily 
available for most players!  These names come from Eric A. Welsh */

  char *strings[] = {"Standard", "Room", "Power", "Electric", 
"Analog", "Dance", "Jazz", "Brush", "Orchestral", "Ethnic", "Kick & Snare"} ;

  if (index < 0)
  {
    *size = 11 ;
    return NULL ;
  }
  else
    return strings[index] ;
} /* End Function *drum_getter */

/**********************************************************************/ 
/* This function is needed for the Allegro menu system, and is cribbed 
straight from the documentation. */

char *inst_getter(int index, int *size)
{

/* Names of the patches available to timidity.  Some of these voices are 
not available in freepats.  But it seems like a good idea to make them 
available anyway, so that this limitation lies outside of the program 
rather than inherent in it.  Names here follow freepats where available, 
Yamaha, elsewhere. */

  char *strings[] = 

  { 
"Piano", "Brite Piano", "Electric G Piano", "Honky-tonk Piano", 
"Electric Piano 1", "Electric Piano 2", "Harpsichord", "Clavinet", 
"Celesta", "Glockenspiel", "Music Box", "Vibraphone", "Marimba", 
"Xylophone", "Tubular Bells", "Dulcimer", "Hammond Organ", "Percussive \
Organ", "Rock Organ", "Church Organ", "Reed Organ", "Accordion", 
"Harmonica", "Tango Accordion", "Nylon Guitar", "Steel Guitar", "Jazz \
Guitar", "Clean Electric Guitar", "Muted Electric Guitar", "Overdriven \
Guitar", "Distortion Guitar", "Guitar Harmonics", "Acoustic Bass", 
"Finger Bass", "Pick Bass", "Fretless Bass", "Slap Bass 1", "Slap Bass \
2", "Synth Bass 1", "Synth Bass 2", "Violin", "Viola", "Cello", 
"Contrabass", "Tremolo Strings", "Pizzicato Strings", "Harp", "Timpani", 
"String Ensemble", "String Ensemble 2", "Synth Strings 1", "Synth \
Strings 2", "Choir Aahs", "Voice Oohs", "Synth Voice", "Orchestra Hit", 
"Trumpet", "Trombone", "Tuba", "Muted Trumpet", "French Horn", "Brass \
Section", "Synth Brass 1", "Synth Brass 2", "Soprano Sax", "Alto Sax", 
"Tenor Sax", "Baritone Sax", "Oboe", "English Horn", "Bassoon", 
"Clarinet", "Piccolo", "Flute", "Recorder", "Panpipe", "Bottle Blow", 
"Shakuhachi", "Whistle", "Ocarina", "Square Wave", "Sawtooth Lead", 
"Calliope Lead", "Chiff Lead", "Charang Lead", "Voice Lead", "Fifth \
Lead", "Bass+Lead", "New Age", "Warm Pad", "Polysynth Pad", "Choir Pad", 
"Bowed Pad", "Metallic Pad", "Halo Pad", "Sweep Pad", "Rain", 
"Soundtrack", "Crystal", "Atmosphere", "Brightness", "Goblins-Unicorn", 
"Echo Voice", "Sci-Fi", "Sitar", "Banjo", "Shamisen", "Koto", "Kalimba", 
"Bagpipe", "Fiddle", "Shanai", "Tinkle Bell", "Agogo", "Steel Drums", 
"Wood Block", "Taiko Drum", "Melodic Drum", "Synth Drum", "Reverse \
Cymbal", "Fret Noise", "Breath Noise", "Seashore", "Bird Tweet", 
"Telephone Ring", "Helicopter", "Applause", "Gunshot"
  } ;

  if (index < 0) 
  {
    *size = 128 ;
    return NULL ;
  }
  else 
    return strings[index] ;
  
} /* End Function inst_getter() */

/**********************************************************************/
/* a custom dialog procedure to capture the return value of d_check_proc */

int my_check_proc(int msg, DIALOG *d, int c) 
{
  int retval ;
  retval = d_check_proc(msg, d, c) ;

  if (msg == MSG_KEY) {
    h = !(h) ;
//DEBUG :
//fprintf(stderr,"check_proc: msg = %i, c = %i, retval = %i \n", msg, c, retval) ;
  }

  return D_O_K ;
}

/**********************************************************************/
/* a custom dialog procedure, which prints the value of a GUI slider as
a number */
int gui_slider_callback(void *dp3, int d2)
{
  char slider_type ;
  slider_type = *(char *)dp3 ;

  if (slider_type == 'T') {
    char buffer[4] ;
    snprintf(buffer, 4, "%3d", d2 + MIN_BPM) ;
    active_dialog[14].dp = buffer ;
    if (object_message(&active_dialog[14], MSG_DRAW, 0) != D_O_K)
      fprintf(stderr, "Unlucky return value from object_message \n") ;
  }
  else if (slider_type == 'V') {
    char buffer[4] ;
    snprintf(buffer, 4, "%3d", d2) ;
    active_dialog[15].dp = buffer ;
    if (object_message(&active_dialog[15], MSG_DRAW, 0) != D_O_K)
      fprintf(stderr, "Bad return value from object_message() \n") ;
  }
  else if (slider_type == 'L') {
     char buffer[5] ;
     snprintf(buffer, 5, "%3d%%", d2 + 50) ;
     active_dialog[16].dp = buffer ;
     if (object_message(&active_dialog[16], MSG_DRAW, 0) != D_O_K)
       fprintf(stderr, "Bad return value from object_message() \n") ;
  }
  else {
    fprintf(stderr, "slider_callback(): %c is not recognized\n", slider_type) ;
  }

  return D_O_K ;
}
/**********************************************************************/

/* dialog which must come before the main settings dialog */

int query_channel()
{
  int channel ;

  DIALOG channel_dialog[] =
  {
/* (dialog proc)     (x)   (y)   (w)   (h) (fg)(bg) (key) (flags)     (d1) (d2)    (dp)                   (dp2) (dp3) */
/* First clear the screen, 				array element 0 */
   {d_clear_proc, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL},
/* Prompt for the MIDI channel, 			array element 1 */
  {d_text_proc, 0.1 * SCREEN_W, 0.4 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "What MIDI channel, 0 to 15 ?", NULL, NULL},
/* Listbox for channel					array element 2 */
   {d_list_proc, 0.6 * SCREEN_W, 0.4 * SCREEN_H, 0.20 * SCREEN_W, 0.30 * SCREEN_H, 0, 0, 0, D_EXIT, 0, 0, (void *)chan_getter, NULL, NULL},
/* A kind explanation of what this means                array element 3 */
   {d_ctext_proc, 0.5 * SCREEN_W, 0.9 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "Selecting a channel determines whether melody or drums will be heard.", NULL, NULL },
/* Yield processor time					array element 4 */
   {d_yield_proc, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL},
/* Terminate the list 					array element 5 */
   {NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL}
  } ; /* End DIALOG array */

/* set up colors */
  gui_bg_color = makecol(0, 0, 0);
  gui_fg_color = makecol(200, 240, 200);
  set_dialog_color (channel_dialog, gui_fg_color, gui_bg_color);

/* black color for d_clear_proc */
  channel_dialog[0].bg = makecol(0, 0, 0);
   
/* shift the dialog 2 pixels away from the border */
  position_dialog (channel_dialog, 2, 2);

  do_dialog(channel_dialog, -1) ;
  channel = channel_dialog[2].d1 ;

  return channel ;
} /* End Function channel_dialog() */

/**********************************************************************/
/* a dialog sequence for collecting data */

void query_settings(struct S_V *setting_vars)
{

/* Characters to mark which slider is which, to be addressed in the dp3 field */
  char tempo_marker = 'T' ;
  char velocity_marker = 'V' ;
  char legato_marker = 'L' ;

DIALOG setting_dialog[] = 
{
/* (dialog proc)     (x)   (y)   (w)   (h) (fg)(bg) (key) (flags)     (d1) (d2)    (dp)                   (dp2) (dp3) */
/* First clear the screen, 				array element 0 */
  {d_clear_proc, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL},
/* Prompt for an initial tempo, 			array element 1 */
  {d_text_proc, 0.05 * SCREEN_W, 0.05 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "Initial tempo in beats per minute", NULL, NULL},
/* Prompt for an initial velocity, 			array element 2 */
  {d_text_proc, 0.05 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "Please enter initial volume", NULL, NULL},
/* Prompt for an intital legato level, 			array element 3 */
  {d_text_proc, 0.05 * SCREEN_W, 0.15 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "Please enter initial legato", NULL, NULL},
/* Prompt for the key signature, 			array element 4 */
  {d_text_proc, 0.05 * SCREEN_W, 0.20 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "Which key signature?", NULL, NULL},
/* Prompt for the time signature, 			array element 5 */
  {d_text_proc, 0.05 * SCREEN_W, 0.35 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "How many beats per measure?", NULL, NULL},
/* Prompt for patch 					array element 6 */
  {d_text_proc, 0.05 * SCREEN_W, 0.50 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "What voice to play? ", NULL, NULL},
/* Slider for tempo, (needs to be transformed)		array element 7 */
  {d_slider_proc, 0.50 * SCREEN_W, 0.05 * SCREEN_H, 0.25 * SCREEN_W, 0.02 * SCREEN_H, 0, 0, 0, 0, MAX_BPM - MIN_BPM, 60, NULL, gui_slider_callback, &tempo_marker},
/* Slider for velocity, 				array element 8 */
  {d_slider_proc, 0.50 * SCREEN_W, 0.10 * SCREEN_H, 0.25 * SCREEN_W, 0.02 * SCREEN_H, 0, 0, 0, 0, 127, 64, NULL, gui_slider_callback, &velocity_marker},
/* Slider for legato, (needs to be transformed)		array element 9 */
  {d_slider_proc, 0.50 * SCREEN_W, 0.15 * SCREEN_H, 0.25 * SCREEN_W, 0.02 * SCREEN_H, 0, 0, 0, 0, 50, 25, NULL, gui_slider_callback, (void *)&legato_marker},
/* Listbox for key signature				array element 10 */
  {d_list_proc, 0.60 * SCREEN_W, 0.20 * SCREEN_H, 0.20 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 0, 0, 7, 0, (void *)ksig_getter, NULL, NULL},
/* Listbox for time signature 				array element 11 */
  {d_list_proc, 0.60 * SCREEN_W, 0.35 * SCREEN_H, 0.07 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 0, 0, 2, 0, (void *)tsig_getter, NULL, NULL},
/* Listbox for patch 					array element 12 */
  {d_text_list_proc, 0.35 * SCREEN_W, 0.50 * SCREEN_H, 0.6 * SCREEN_W, 0.25 * SCREEN_H, 0, 0, 0, 0, 0, 0, (void *)inst_getter, NULL, NULL},
/* Checkbox for whether to use harmony mode             array element 13 */
  {my_check_proc, 0.05 * SCREEN_W, 0.80 * SCREEN_H, 0.3 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 'h', 0, 0, 0, "&Harmony mode:", NULL, NULL},
/* Display the value of the tempo slider                array element 14*/
  {d_rtext_proc, 0.95 * SCREEN_W, 0.05 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "120", NULL, NULL}, 
/* Display the value of the velocity slider                array element 15*/
  {d_rtext_proc, 0.95 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, " 64", NULL, NULL}, 
/* Display the value of the legato slider                array element 16*/
  {d_rtext_proc, 0.95 * SCREEN_W, 0.15 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, " 75%", NULL, NULL}, 
/* Click when done, 					array element 17 */
  {d_button_proc, 0.85 * SCREEN_W, 0.85 * SCREEN_H, 0.10 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 'd', D_EXIT, 0, 0, "&Done", NULL, NULL},
/* Yield processor time					array element 18 */
  {d_yield_proc, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL},
/* Terminate the list 					array element 19 */
  {NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL}
  } ; /* End DIALOG array */

/* If this is a drum channel, change the callback */
  if (setting_vars->channel == 9 ) 
    setting_dialog[12].dp = (void *)drum_getter ;

/* set up colors */
  gui_bg_color = makecol(0, 0, 0);
  gui_fg_color = makecol(200, 240, 200);
  set_dialog_color (setting_dialog, gui_fg_color, gui_bg_color);

/* black color for d_clear_proc */
  setting_dialog[0].bg = makecol(0, 0, 0);
   
/* shift the dialog 2 pixels away from the border */
  position_dialog (setting_dialog, 2, 2);

/* set default values */
  h = FALSE ;

/* do the dialog */
  do_dialog(setting_dialog, -1);

/* Extract the data */
  setting_vars->bpm = setting_dialog[7].d2 + MIN_BPM ;
  setting_vars->velocity = setting_dialog[8].d2 ;
  setting_vars->legato = 0.5 + (float)setting_dialog[9].d2 / 100.0 ;
  setting_vars->key_sig = setting_dialog[10].d1 - 7 ;
  setting_vars->time_sig = setting_dialog[11].d1 + 2 ;

  if (setting_vars->channel == 9) { 
/* The patch numbers required for the drum kit are irregular */
    switch (setting_dialog[12].d1) {
      case 0 :
        setting_vars->patch = 0 ;
        break ;
      case 1 :
        setting_vars->patch = 8 ;
        break ;
      case 2 :
        setting_vars->patch = 16 ;
        break ;
      case 3 :
        setting_vars->patch = 24 ;
        break ;
      case 4 :
        setting_vars->patch = 25 ;
        break ;
      case 5 :
        setting_vars->patch = 26 ;
        break ;
      case 6 :
        setting_vars->patch = 32 ;
        break ;
      case 7 :
        setting_vars->patch = 40 ;
        break ;
      case 8 :
        setting_vars->patch = 48 ;
        break ;
      case 9 :
        setting_vars->patch = 49 ;
        break ;
      case 10 :
        setting_vars->patch = 50 ;
        break ;
      default :
        setting_vars->patch = 0 ;
    }  /* End Switch( what drum kit?) */
  }
  else {
    setting_vars->patch = setting_dialog[12].d1 ;
  } /* End If */

/* Harmony should be an option in the Allegro version */
  setting_vars->harmony = h ;

  clear_bitmap(screen) ;
} /* End function query_settings() */

/**************************************************************************/

void query_details(struct D_V *detail_vars)
{
  int n  = 0 ;
  char text_string[STRING_SIZE] = "",
       copyright_string[STRING_SIZE] = "",
       title_string[STRING_SIZE] = "tune01" ;
       
  DIALOG detail_dialog[] = 
  {
/* First clear the screen, 				array element 0 */
    {d_clear_proc, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL},
/* Prompt for title 					array element 1 */
    {d_text_proc, 0.05 * SCREEN_W, 0.05 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "Title for this track? ", NULL, NULL},
/* Prompt for copyright 				array element 2 */
    {d_text_proc, 0.05 * SCREEN_W, 0.15 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "Author and date? ", NULL, NULL},
/* Prompt for text 					array element 3 */
    {d_text_proc, 0.05 * SCREEN_W, 0.25 * SCREEN_H, 0, 0, 0, 0, 0, 0, 0, 0, "Any comments? ", NULL, NULL},
/* Entry field for title 				array element 4 */
    {d_edit_proc, 0.50 * SCREEN_W, 0.05 * SCREEN_H, 0.45 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 0, 0, STRING_SIZE -1, 0, title_string, NULL, NULL},
/* Entry field for copyright 				array element 5 */
    {d_edit_proc, 0.50 * SCREEN_W, 0.15 * SCREEN_H, 0.45 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 0, 0, STRING_SIZE -1, 0, copyright_string, NULL, NULL},
/* Entry field for text 				array element 6 */
    {d_edit_proc, 0.50 * SCREEN_W, 0.25 * SCREEN_H, 0.45 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 0, 0, STRING_SIZE -1, 0, text_string, NULL, NULL},
/* Click when done, 					array element 7 */
    {d_button_proc, 0.85 * SCREEN_W, 0.85 * SCREEN_H, 0.10 * SCREEN_W, 0.10 * SCREEN_H, 0, 0, 'd', D_EXIT, 0, 0, "&Done", NULL, NULL},
/* Yield processor time					array element 8 */
    {d_yield_proc, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL},
/* Terminate the list 					array element 9 */
    {NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL}
  } ; /* End DIALOG array */

/* set up colors */
  gui_bg_color = makecol(0, 0, 0);
  gui_fg_color = makecol(200, 240, 200);
  set_dialog_color (detail_dialog, gui_fg_color, gui_bg_color);

/* black color for d_clear_proc */
  detail_dialog[0].bg = makecol(0, 0, 0);
   
/* shift the dialog 2 pixels away from the border */
  position_dialog(detail_dialog, 2, 2);

  do_dialog(detail_dialog, -1) ;

/* Extract data from the structure */
  strcpy(detail_vars->text, detail_dialog[6].dp) ;
  strcpy(detail_vars->copyright, detail_dialog[5].dp) ;
  strcpy(detail_vars->title, detail_dialog[4].dp) ;

  if (detail_vars->channel == 9) {
    strcpy(detail_vars->instrument, drum_getter(detail_vars->patch, &n)) ;
  }
  else {
    strcpy(detail_vars->instrument, inst_getter(detail_vars->patch, &n)) ;
  } /* End If */

  clear_bitmap(screen) ;
} /* End Function query_details */

/**********************************************************************/ 

void metronome(unsigned long present_time, int time_sig)
{
  int i, x, y, beat ;
/* a one-character string for compatibility with textout() */
  char beat_string[] = "1" ; 

  y = text_height(font) * (METRONOME_LINE + 1) ;

/* Erase the old mark. */
  for (i = 0 ; i <= time_sig ; ++i)
  {
    x = text_length(font, "Metronome:  ") + text_length(font, "1") * i ;
    textout_ex(screen, font, " ", x, y, makecol(0, 0, 0), makecol(0, 0, 0) ) ;
  }
 
/* calculate the present beat of the measure.  Draw a new mark */
  beat = ((present_time / TICKS_PER_BEAT) % time_sig) + 1 ;
  x = text_length(font, "Metronome:  ") + text_length(font, "1") * beat ;
  beat_string[0] = (char) (beat + (int)('1') - 1 ) ;  

  textout_ex(screen, font, beat_string, x, y, makecol(128,128,128), makecol(0, 0, 0)) ;

} /* End Function metronome() */

/**********************************************************************/ 
/* Collect information on which keys have been pressed */

int get_input(int channel)
{
  static bool first_event = TRUE ;  /* whether to initialize prior_keystate */
  static int queue_begin = 0 , queue_end = 0 ;             /* array indices */
  static int event_queue[MAX_QUEUE_LENGTH] ;           /* a revolving queue */
  static bool lapped = FALSE ;        /* has end index met the begin index? */

/* (0 < note < 128) should indicate NoteOn, and higher numbers should 
represent controllers. */
  static bool prior_keystate[139] ;
  int event, i, retval ;
  bool b ;
 
/* Remember to set all keys to initially not pressed */
  if (first_event) { 
    for (i = 1 ; i < 139 ; ++i)    /* Taking there to be 10 controllers + 128 */
      prior_keystate[i] = FALSE ;
    first_event = FALSE ;
  }

/* The array event_queue is a ring buffer indexed with a beginning 
marker and an end marker.  These represent the next vacant element, and 
the next unreturned element, respectively.  First, read the keyboard. */
  if (keyboard_needs_poll()) {
    if (poll_keyboard() != 0) fprintf(stderr, "Error reading the keyboard!\n") ;
  }

/* some very repetious code for each recognized keystroke */
  b = (bool)key[KEY_SPACE] ;
  if (b && (!prior_keystate[REST]) ) {
    event_queue[queue_end] = REST ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[REST] = b ;

  b = (bool)key[KEY_DOWN] ;
  if (b && (!prior_keystate[DEC_VELOCITY]) )  {
    event_queue[queue_end] = DEC_VELOCITY ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[DEC_VELOCITY] = b ;
  
  b = (bool)key[KEY_UP] ;
  if (b && (!prior_keystate[INC_VELOCITY]) ) {
    event_queue[queue_end] = INC_VELOCITY ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[INC_VELOCITY] = b ; 

  b = (bool)key[KEY_LEFT] ;
  if (b && (!prior_keystate[DEC_TEMPO]) ) {
    event_queue[queue_end] = DEC_TEMPO ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[DEC_TEMPO] = b ;

  b = (bool)key[KEY_RIGHT] ;
  if (b && (!prior_keystate[INC_TEMPO]) )  {
    event_queue[queue_end] = INC_TEMPO ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[INC_TEMPO] = b ;

  b = (bool)key[KEY_PGDN] ;
  if (b && (!prior_keystate[PIANO]) )  {
    event_queue[queue_end] = PIANO ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[PIANO] = b ;

  b = (bool)key[KEY_PGUP] ;
  if (b && (!prior_keystate[FORTE]) )  {
    event_queue[queue_end] = FORTE ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[FORTE] = b ;

  b = (bool)key[KEY_HOME] ;
  if (b && (!prior_keystate[STACCATO]) ) {
    event_queue[queue_end] = STACCATO ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[STACCATO] = b ;

  b = (bool)key[KEY_END] ;
  if (b && (!prior_keystate[LEGATO]) )  {
    event_queue[queue_end] = LEGATO ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[LEGATO] = b ;

  b = (bool)(key[KEY_BACKSPACE] || key[KEY_ESC] || key[KEY_F1]) ;
  if (b && (!prior_keystate[QUIT]) ) {
    event_queue[queue_end] = QUIT ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
  }
  prior_keystate[QUIT] = b ;

/* Now consider the note events, even more tedious */
  b = (bool)key[KEY_Z] ;
  if (b != prior_keystate[57])   {
    event_queue[queue_end] = b ? 57 : -57 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[57] = b ;
  }

  b = (bool)key[KEY_X] ;
  if (b != prior_keystate[58])   {
    event_queue[queue_end] = b ? 58 : -58 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[58] = b ;
  }

  b = (bool)key[KEY_C] ;
  if (b != prior_keystate[59]) {
    event_queue[queue_end] = b ? 59 : -59 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[59] = b ;
  }

  b = (bool)key[KEY_V] ;
  if (b != prior_keystate[60])  {
    event_queue[queue_end] = b ? 60 : -60 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[60] = b ;
  }

  b = (bool)key[KEY_B] ;
  if (b != prior_keystate[61])   {
    event_queue[queue_end] = b ? 61 : -61 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[61] = b ;
  }

  b = (bool)key[KEY_N] ;
  if (b != prior_keystate[62])   {
    event_queue[queue_end] = b ? 62 : -62 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[62] = b ;
  }

  b = (bool)key[KEY_M] ;
  if (b != prior_keystate[63])  {
    event_queue[queue_end] = b ? 63 : -63 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[63] = b ;
  }

  b = (bool)key[KEY_COMMA] ;
  if (b != prior_keystate[64]) {
    event_queue[queue_end] = b ? 64 : -64 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[64] = b ;
  }

  b = (bool)key[KEY_STOP] ;
  if (b != prior_keystate[65]) {
    event_queue[queue_end] = b ? 65 : -65 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[65] = b ;
  }

  b = (bool)key[KEY_SLASH] ;
  if (b != prior_keystate[66]) {
    event_queue[queue_end] = b ? 66 : -66 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[66] = b ;
  }

  b = (bool)key[KEY_A] ;
  if (b != prior_keystate[67]) {
    event_queue[queue_end] = b ? 67 : -67 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[67] = b ;
  }

  b = (bool)key[KEY_S] ;
  if (b != prior_keystate[68]) {
    event_queue[queue_end] = b ? 68 : -68 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[68] = b ;
  }
  
  b = (bool)key[KEY_D] ;
  if (b != prior_keystate[69]) {
    event_queue[queue_end] = b ? 69 : -69 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[69] = b ;
  }

  b = (bool)key[KEY_F] ;
  if (b != prior_keystate[70]) {
    event_queue[queue_end] = b ? 70 : -70 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[70] = b ;
  }

  b = (bool)key[KEY_G] ;
  if (b != prior_keystate[71]) {
    event_queue[queue_end] = b ? 71 : -71 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[71] = b ;
  }

  b = (bool)key[KEY_H] ;
  if (b != prior_keystate[72]) {
    event_queue[queue_end] = b ? 72 : -72 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[72] = b ;
  }

  b = (bool)key[KEY_J] ;
  if (b != prior_keystate[73]) {
    event_queue[queue_end] = b ? 73 : -73 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[73] = b ;
  }

  b = (bool)key[KEY_K] ;
  if (b != prior_keystate[74]) {
    event_queue[queue_end] = b ? 74 : -74 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[74] = b ;
  }
  
  b = (bool)key[KEY_L] ;
  if (b != prior_keystate[75]) {
    event_queue[queue_end] = b ? 75 : -75 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[75] = b ;
  }
  
  b = (bool)key[KEY_COLON] ;
  if (b != prior_keystate[76])  {
    event_queue[queue_end] = b ? 76 : -76 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[76] = b ;
  }

  b = (bool)key[KEY_Q] ;
  if (b != prior_keystate[77]) {
    event_queue[queue_end] = b ? 77 : -77 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[77] = b ;
  }

  b = (bool)key[KEY_W] ;
  if (b != prior_keystate[78])  {
    event_queue[queue_end] = b ? 78 : -78 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[78] = b ;
  }
  
  b = (bool)key[KEY_E] ;
  if (b != prior_keystate[79]) {
    event_queue[queue_end] = b ? 79 : -79 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[79] = b ;
  }
  
  b = (bool)key[KEY_R] ;
  if (b != prior_keystate[80]) {
    event_queue[queue_end] = b ? 80 : -80 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[80] = b ;
  }

  b = (bool)key[KEY_T] ;
  if (b != prior_keystate[81])  {
    event_queue[queue_end] = b ? 81 : -81 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[81] = b ;
  }

  b = (bool)key[KEY_Y] ;
  if (b != prior_keystate[82]) {
    event_queue[queue_end] = b ? 82 : -82 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[82] = b ;
  }
  
  b = (bool)key[KEY_U] ;
  if (b != prior_keystate[83]) {
    event_queue[queue_end] = b ? 83 : -83 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[83] = b ;
  }
 
  b = (bool)key[KEY_I] ;
  if (b != prior_keystate[84]) {
    event_queue[queue_end] = b ? 84 : -84 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[84] = b ;
  }
  
  b = (bool)key[KEY_O] ;
  if (b != prior_keystate[85])  {
    event_queue[queue_end] = b ? 85 : -85 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[85] = b ;
  }
  
  b = (bool)key[KEY_P] ;
  if (b != prior_keystate[86]) {
    event_queue[queue_end] = b ? 86 : -86 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[86] = b ;
  }
  
  b = (bool)key[KEY_1] ;
  if (b != prior_keystate[87]) {
    event_queue[queue_end] = b ? 87 : -87 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[87] = b ;
  }
 
  b = (bool)key[KEY_2] ;
  if (b != prior_keystate[88]) {
    event_queue[queue_end] = b ? 88 : -88 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[88] = b ;
  }
  
  b = (bool)key[KEY_3] ;
  if (b != prior_keystate[89]) {
    event_queue[queue_end] = b ? 89 : -89 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[89] = b ;
  }
  
  b = (bool)key[KEY_4] ;
  if (b != prior_keystate[90]) {
    event_queue[queue_end] = b ? 90 : -90 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[90] = b ;
  }
  
  b = (bool)key[KEY_5] ;
  if (b != prior_keystate[91]) {
    event_queue[queue_end] = b ? 91 : -91 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[91] = b ;
  }
  
  b = (bool)key[KEY_6] ;
  if (b != prior_keystate[92]) {
    event_queue[queue_end] = b ? 92 : -92 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[92] = b ;
  }
  
  b = (bool)key[KEY_7] ;
  if (b != prior_keystate[93]) {
    event_queue[queue_end] = b ? 93 : -93 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[93] = b ;
  }
  
  b = (bool)key[KEY_8] ;
  if (b != prior_keystate[94]) {
    event_queue[queue_end] = b ? 94 : -94 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[94] = b ;
  }

  b = (bool)key[KEY_9] ;
  if (b != prior_keystate[95]) {
    event_queue[queue_end] = b ? 95 : -95 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[95] = b ;
  }

  b = (bool)key[KEY_0] ;
  if (b != prior_keystate[96]) {
    event_queue[queue_end] = b ? 96 : -96 ;
    queue_end = (queue_end + 1) % MAX_QUEUE_LENGTH ;
    if (queue_end == queue_begin) lapped = TRUE ;
    prior_keystate[96] = b ;
  }

/* Now take an event out of the queue */
  if (lapped) {
    queue_begin = queue_end ;
    lapped = FALSE ;
  }
  else if (queue_end == queue_begin) return NO_KEYSTROKE ;
  
  event = event_queue[queue_begin] ;
  queue_begin = (queue_begin + 1) % MAX_QUEUE_LENGTH ;
  
//DEBUG: see what's in the event queue
//  fprintf(stderr, "Events in queue are %d, %d, %d, %d\n", event_queue[0], event_queue[1], event_queue[2], event_queue[3]) ;
//  fprintf(stderr, "queue_begin = %d, queue_end = %d\n", queue_begin, queue_end) ;

/* For a drum channel, modify the note number by a constant offset.  
It's tidier to do this here, than to make the change throughout 
record(). */
  if ( (channel == 9) && (event >= 1) && (event <= 127) ) {
    retval = event + LOW_DRUM - 57 ;
  }
  else if ( (channel == 9) && (event >= -127) && (event <= -1) ) {
    retval = event - LOW_DRUM + 57 ;
  }
  else {
    retval = event ;
  }

  return retval ;

} /* End Function get_input */

/**********************************************************************/ 
/* the timer interrupt callback function -- not as simple as it looks */

void timer()
{
  ++counter ;
}
END_OF_FUNCTION(timer)

/**********************************************************************/ 
/* get the time, measured in milliseconds.  This is easy, because the 
callback function does all the real work */ 

unsigned long get_time() 
{
  unsigned long msec_now ;

  msec_now  = 1000 * ( (float) counter / TIMER_FREQUENCY) ;
  return msec_now ;
}

/**********************************************************************/
/* A more thorough setup function */
void ready_screen() 
{
  int width, height, depth ;    /* video resolution and color depth */

  allegro_init() ;
  install_keyboard() ;
/* Turn off key repeats, if it becomes necesary  set_keyboard_rate(0, 0) ; */
  install_timer() ;

/* Try to set a compatible color depth */
  if ( (depth = desktop_color_depth()) != 0) {
    set_color_depth(depth) ;  
  }
  else {
    set_color_depth(8) ;
  }

/* Try to set the preferred graphics resolution */
  if (get_desktop_resolution(&width, &height) ) { /* Nonzero denotes an error */
    width = WIDTH1 ;
    height = HEIGHT1 ;
  }

/* Try setting a graphics mode in order of preference */
  if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, width * 0.9, height * 0.9, 0, 0) == 0) {
    ;   /* Nothing required to be done if successful */      }
  else if (set_gfx_mode(GFX_AUTODETECT, width, height, 0, 0) == 0) {
    ; /* Do nothing special on success */      }
  else if (set_gfx_mode(GFX_AUTODETECT, WIDTH1, HEIGHT1, 0, 0) == 0) {
    ; /* Do nothing on success */         }
  else if (set_gfx_mode(GFX_AUTODETECT, WIDTH2, HEIGHT2, 0, 0) == 0) {
    ; /* Do nothing on success */         }
  else if (set_gfx_mode(GFX_AUTODETECT, WIDTH3, HEIGHT3, 0, 0) == 0) {
    ; /* Do nothing on success */         }
  else if (set_gfx_mode(GFX_SAFE, width, height, 0, 0) == 0) {
    ; /* Do nothing on success */      }
  else {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0) ;
    allegro_message("Couldn't set any graphics mode\n %s\n", allegro_error) ;
    exit(EXIT_FAILURE) ;
  }

/* set the window title to something useful */
  set_window_title("imidiacy MIDI recorder") ;

/* Can't remember why this is here instead of in function prepare_timing() */
  LOCK_VARIABLE(counter) ;
} 

/**********************************************************************/
/* An easy cleanup function */
void cleanup_screen() 
{
  int i ;
  struct punchtype *ptr_present, *ptr_old ;

/* Clear out the piano-roll punches from any previous takes */
  for (i = 0 ; i < 40 ; ++i) {
    ptr_present = punch[i] ;
    if (ptr_present == NULL) 
      continue ;

/* Advance along the linked list */
    while (ptr_present != NULL) {
      ptr_old = ptr_present ;
      ptr_present = ptr_present -> ptr_next ;
      free(ptr_old) ;
    } /* Wend */
    punch[i] = NULL ;
  } /* Next note */

/* don't need to call remove_timer()  */
/* don't need to call remove_keyboard() */
  allegro_exit() ;
}

/**********************************************************************/
/* An easy pause function */
void press_any_key()
{
  if ( keypressed() ) 
    clear_keybuf() ;
  textout_ex(screen, font, "Press any key to continue...", 0.25 * SCREEN_W, 0.25 * SCREEN_H, makecol(128, 128, 128), makecol(0, 0, 0)) ;
  (void) readkey() ;  
  clear_keybuf() ;
}
/**********************************************************************/

