/* imidiacy - Record a MIDI file from the computer keyboard 
version 2007.9.06 Copyright (C) 2006-2007 Andrew Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 59 
Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program produces standard MIDI files based on presses of the PC
keyboard.  The events Note Off, Note On, Aftertouch, and tempo change
are supported.  Output is to the monitor and soundcard or PC speaker.
 
For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/

/* 
This preprocessor instruction is intended to define type bool, if it has 
not been defined already in curses.h, which (get ready for this one) has 
bool as a macro for NCURSES_BOOL as a typedef for unsigned char, UNLESS 
stdbool.h is #include'd, in which case NCURSES_BOOL is a macro for bool, 
defined in stdbool.  It makes sense for me just to include stdbool.h, 
but the logical constants will be left capitalized, which matches usage 
in the ncurses manual.
*/
#ifndef CURSES_H
  #undef TRUE
  #define TRUE 1
  #undef FALSE
  #define FALSE 0
  #include <stdbool.h>
#endif

/* Storage type for the events, somewhat resembling the file format.  
It's an ugly compromise between the Channel and Meta formats, and does 
not exactly suit either one.  If type is 0X8, 0X9, or 0XA, then four 
bits of it are merged with four bits of the channel variable and written 
as one byte to the file.  If type is 0X51, then the redundant data must 
be inserted to the file: 255(meta) 51(tempo) 3(length following).  Then 
the next three bytes have a meaning different from the tag: they denote 
the tempo in microseconds per quarter note.

This structure type is named DBL_NODE because it contains a pointer to 
both the previous and next nodes in the sequence of events.  This allows 
for editing earlier events without reading all the way from the 
beginning of the list.
*/

struct double_linked_list {
  unsigned char delta_time[4] ;
  unsigned char type ;
  unsigned char channel ;
  unsigned char note ;
  unsigned char velocity ;
  struct double_linked_list *ptr_prev ;
  struct double_linked_list *ptr_next ;
} ;
typedef struct double_linked_list DBL_NODE ;

/* A structure for conveniently passing user input back to the main 
program.  S_V means Setting Variables */

struct S_V {
  int bpm ;
  int velocity ;
  float legato ;
  int channel ;
  int key_sig ;
  int time_sig ;
  bool harmony ;
  int patch ;
} ;

/* A structure for passing back user input.  D_V means Detail Variables */
struct D_V {
  char *text ;
  char *copyright ;
  char *title ;
  char *instrument ;
  int channel ;
  int patch ;
} ;

/* A structure for storing the beginning and end of the "punches" in the 
piano roll view.  */
struct punchtype {
  unsigned short y_head ;
  unsigned short y_tail ;
  struct punchtype *ptr_next ;
} ;

/* Initialize global constants */ 
/* give names to the values returned by get_input() These have changed */
#define NO_KEYSTROKE    0
#define DEC_VELOCITY  128
#define INC_VELOCITY  129
#define DEC_TEMPO     130
#define INC_TEMPO     131
#define PIANO         132
#define FORTE         133
#define STACCATO      134
#define LEGATO        135
#define QUIT          136
#define KEY_IGNORED   137
#define REST          138

/* All the silly exit codes have been eliminated in favor of C99 
standard EXIT_SUCCESS and EXIT_FAILURE */

/* misc constants, some of which may be adjusted as desired */
#define MAX_BPM 180           /* fastest allowable b.p.m. (tempo, measured in beats per minute) */
#define MIN_BPM 60            /* slowest allowable b.p.m. */
#define BPM_STEP 6            /* how much to change tempo per keystroke (measured in user-oriented b.p.m., not the microseconds per quarter note used internally. */ 
#define VEL_PIANO 25          /* "piano" velocity preset level */
#define VEL_FORTE 100         /* "forte" velocity preset level */ 
#define VEL_STEP 10           /* how much to change velocity per keystroke */
#define LEG_PERCENTAGE 105    /* divided by 100, determines how much to change the legato setting per keystroke */
#define TOL_METRONOME 5       /* number of "ticks" past a beat it's acceptable to beep */
#define TICKS_PER_BEAT 120    /* the programmer has chosen this figure arbitrarily */ 
#define STRING_SIZE 128       /* size to allocate for title, author, etc */
#define DELAY_DURATION 2      /* time in milliseconds to pause before reading another event */
#define MICROSECONDS_PER_MINUTE  60000000      /* conversion factor */
#define NUM_DRUMSOUNDS 56     /* How many drum samples to allocate */
#define LOW_DRUM 36           /* For Allegro drums, which note number maps to key 'Z' */
#define LOW_NOTE 57           /* For Allegro, the note number mapped to 'Z' */

/* Trying to get all the curses-related definitions put in one place. */
#define STATUS_LINES     6       /* how much header text appears on the ouput screen, above the note-display area */
#define TEMPO_LINE       2       /* where tempo slider appears */
#define VELOCITY_LINE    3       /* where velocity slider appears */
#define LEGATO_LINE      4       /* where legato slider appears */
#define METRONOME_LINE   5       /* where the timekeeping appears */
#define LOWEST_SCREEN_NOTE  21 /* which midi note number shall appear on the left of the output screen */

/* Section for definitions related to the Allegro library functions */
#define MAX_QUEUE_LENGTH   64          /* how many events to remember */
#define WIDTH1            640          /* preferred graphics resolutions */
#define HEIGHT1           480          /* most preferred */
#define WIDTH2            800
#define HEIGHT2           600          /* less preferred */
#define WIDTH3           1024      /* Hi-res is unneeded for this application */
#define HEIGHT3           768
#define TIMER_FREQUENCY   240      /* in Hertz, how often to adjust the timer */
#define SCROLL_MARGIN      10      /* How much room to leave in the piano roll view, as a percentage of the width  */
#define SCROLL_DELAY       24      /* in ms per pixel */
#define TOL_SCROLL          7     /* in ms.  This should be greater than the DELAY_DURATION */


/* Function declarations */
long record(long tempo, int velocity, float legato, int channel, int key_sig, int time_sig, bool use_sound, bool harmony) ;
DBL_NODE *add_node(DBL_NODE *ptr_active) ;
int write_file(long tempo, int channel, int key_sig, int time_sig, int patch, int take, long count_bytes) ;
void handle_signal(int signum) ;
int write_var_bytes(unsigned long number, unsigned char *midi_fmt) ;
int count_var_bytes(unsigned long number) ;
unsigned long read_var_bytes(unsigned char *midi_fmt) ;
void draw_display(int bpm, int velocity, float legato, char *version_string) ;
void add_note(int note, int velocity, int key_sig, int channel) ;
void del_note(int note, int velocity, int channel) ;
void slider(int linenumber, int old_setting, int new_setting) ;
void sliderf(int linenumber, float old_setting, float new_setting) ;
void query_details(struct D_V *detail_vars) ;
void query_settings(struct S_V *setting_vars) ;
int query_keep() ;
int query_another() ;
int query_channel() ;
void play_note(int note, int velocity) ;
void hush_note() ;
void play_drum(int note, int velocity) ;
void adjust_volume(int note, int velocity) ;
int init_sound() ;
void erase_list() ;
void end_sound() ;
void metronome(unsigned long present_time, int time_sig) ;
int get_input(int channel) ;
void cleanup_screen() ;
void ready_screen() ;
void press_any_key() ;
unsigned long get_time() ;
void delay(unsigned int milliseconds) ;
void timer() ;
void scroll_roll() ;
void prepare_timing() ;
void end_timing() ;
void legal(char *disclaimer) ;
