!---- ------------------------------------------------------------------
! Find the median of an array of real numbers.  A reduced version of the 
! partitioning sort (quick sort) is used.  The original array is unchanged.
!
! (C) 2011 Andy Allinger
! Free according to the terms of the Allegro Giftware license
! or, at your option, the GNU General Public License (GPL)
! some public domain subroutines by other programmers are included in this file
!---- ------------------------------------------------------------------
!      __________Variable___Description_____________Data_Type___________
!                 median     the median              Real
!                 X          array                   Real
!                 N          length                  Integer
!                 T          period                  Real
!                 FAULT      exit code               Integer
!                      
!  NOTE:  FAULT returns negative on normal exits, positive on error
!---- ------------------------------------------------------------------
      FUNCTION median(X, N, T, FAULT)
      
      IMPLICIT NONE
      INTEGER N, FAULT
      REAL X(N), T
      REAL median, MED3

!         Local variables
      REAL E, O ; DIMENSION E(N), O(N)               ! work arrays
      INTEGER I, J, K                                ! counters
      INTEGER HI, MID, LO, HIOLD, LOOLD, L2, H2      ! array offsets
      INTEGER nbits                             ! bits to represent n
      REAL PIVOT, LESSER, DIF
      LOGICAL SKIP, INO, INOOLD                      ! data is in O
      INTEGER SIP                               ! sort instead parameter
      REAL MIP                                ! maximum iteration parameter

!      can't use .EQ. operator on floating-point numbers
      REAL BIG, MACHEP, machine
      SAVE BIG, MACHEP
      DATA BIG, MACHEP / 1E30, -1.0 /
!   these parameters barely affect performance, adjust them however you like
      DATA SIP, MIP / 3, 1.62 /   

!---- ------------------------------------------------------------------
      IF (MACHEP < 0) THEN
         MACHEP = machine(1)
      END IF

!      clear error codes
      FAULT = 0

!---- ------------------------------------------------------------------
 
!      special cases for small N
      IF (N < 1) THEN
        median = 0.0
        FAULT = 1
        RETURN
      ELSE IF (N .EQ. 1) THEN
        median = X(1)
        FAULT = -1
        RETURN
      ELSE IF (N .EQ. 2) THEN
        IF (T > MACHEP) THEN
          LESSER = MIN(X(1), X(2))
          median = MAX(X(1), X(2))
          GO TO 80
        END IF
        median = (X(1) + X(2)) / 2.0
        FAULT = -2
        RETURN
      ELSE IF (N .EQ. 3) THEN
        median = MED3(X(1), X(2), X(3))
        FAULT = -3
        RETURN
      END IF

!---- ------------------------------------------------------------------
*         bits required to represent n
       i = BIT_SIZE(n)
  10   i = i - 1
       IF (.NOT. BTEST(n, i)) GO TO 10
       nbits = i + 1

!       find the middle of the array, initial guess, sorting limits
      MID = (N / 2) + 1
      LO = 1 ; HI = N
      PIVOT = MED3(X(LO), X(MID), X(HI))
      LOOLD = LO ; HIOLD = HI ; L2 = LOOLD ; H2 = HIOLD
      INO = .FALSE. ; INOOLD = INO
      LESSER = -BIG ; SKIP = .TRUE.
      
!      sort data from X into work array E
      DO J = 1, N

!        sort values larger than PIVOT to the rear
        IF ( X(J) > PIVOT) THEN
          E(HI) = X(J)
          HI = HI - 1
          
!         values less than or equal to PIVOT come to the front        
        ELSE
          E(LO) = X(J)
          LO = LO + 1
        END IF
      END DO
        
!      find which partition 
!        HI and LO have crossed.  test HI < MID ?
      IF (HI < MID) THEN
        HI = HIOLD
        GO TO 30
        
!          the lesser than or equal to partition         
      ELSE IF (HI > MID) THEN

!            check for equality of all values from MID to HI
        DO K = MID, HI
          IF ( ABS(E(K) - PIVOT) > MACHEP) THEN
            LO = LOOLD
            GO TO 30
          END IF
        END DO
      END IF

!         found it in one pass 
      median = PIVOT
      FAULT = -4
      IF (.NOT. BTEST(N,0)) GO TO 60
      RETURN
      
!---- ------------------------------------------------------------------
!      alternate between putting data into O and putting data into E
!          Beginning of the main loop
!---- ------------------------------------------------------------------
  30  DO I = 1, INT(MIP * nbits) + 1   ! no significance to this limit

!          mark location where a value less than pivot could be found
        IF (LO .NE. LOOLD) THEN
          L2 = LOOLD
          H2 = HIOLD
          INOOLD = INO
        END IF
        LOOLD = LO ; HIOLD = HI
        INO = .NOT. INO 
        
!           odd i
        IF (INO) THEN
!           PIVOT = MED3(E(LO), E( (LO+HI)/2 ), E(HI))  ! standard method
          PIVOT = E(LO + MOD(N*I, (HI-LO)) )    ! alternative pivoting
!            error check
      IF ((MOD(N*I, (HI-LO)) + LO) > HI) FAULT = 2
          DO J = LO, HI                                  ! sort
            IF ( E(J) > PIVOT) THEN
              O(HI) = E(J)
              HI = HI - 1
            ELSE
              O(LO) = E(J)
              LO = LO + 1
            END IF
          END DO  ! next j

          IF (HI < MID) THEN                     !  find which partition
            HI = HIOLD
            GO TO 40
          ELSE IF (HI > MID) THEN
            DO K = MID, HI
              IF ( ABS(O(K) - PIVOT) > MACHEP ) THEN            
                LO = LOOLD
                GO TO 40
              END IF
            END DO
          END IF
          median = PIVOT
          FAULT = -5
          IF (.NOT. BTEST(N,0) ) GO TO 60
          RETURN
        
!---- ------------------------------------------------------------------
!         even i        
        ELSE   
!            use median pivoting on the even passes
!             better for almost-sorted data           
          PIVOT = MED3(O(LO), O((HI+LO)/2), O(HI))
          
          DO J = LO, HI        ! sort
            IF ( O(J) > PIVOT) THEN
              E(HI) = O(J)
              HI = HI - 1
            ELSE
              E(LO) = O(J)
              LO = LO + 1
            END IF
          END DO  ! next j

!      find which partition
          IF (HI < MID) THEN
            HI = HIOLD
            GO TO 40
          ELSE IF (HI > MID) THEN
            DO K = MID, HI
              IF ( ABS(E(K) - PIVOT) > MACHEP ) THEN
                LO = LOOLD
                GO TO 40
              END IF
            END DO
          END IF
          median = PIVOT
          FAULT = -6
          IF (.NOT. BTEST(N,0) ) GO TO 60
          RETURN
        END IF        ! even or odd i
            
!---- ------------------------------------------------------------------
  40    CONTINUE 
        IF ( (HI-LO) < SIP) GO TO 50
      END DO  ! next i
!      PRINT *, 'Went the maximum (', (I-1), ') iterations.'

!---- ------------------------------------------------------------------
!          sort the stubborn
  50  IF (INO) THEN 
        CALL SHELL(O(LO), (HI-LO+1))
        median = O(MID)
      ELSE
        CALL SHELL(E(LO), (HI-LO+1))
        median = E(MID)
      END IF
!      print *, 'Call to Shell, N=',(HI-LO+1), ' yields median =',median

!        odd N finished at this point
      FAULT = -7
      IF (BTEST(N,0)) RETURN
!      even N:  average with the next-less value
      IF (MID > LO) THEN
        IF (INO) THEN
          LESSER = O(MID-1)
        ELSE
          LESSER = E(MID-1)
        END IF
        IF (T > MACHEP) GO TO 80
        median = (median + LESSER) / 2
        FAULT = -8
        RETURN

!---- ------------------------------------------------------------------
!        if MID = LO, seek LESSER in the old work array
      ELSE IF (LO .NE. LOOLD) THEN
!        PRINT *, 'Searching old work array...'
        IF (INO) THEN
          CALL SHELL(E(LOOLD), (HIOLD-LOOLD+1))
          LESSER = E(MID-1)
        ELSE
          CALL SHELL(O(LOOLD), (HIOLD-LOOLD+1))
          LESSER = O(MID-1)
        END IF

        IF (T > MACHEP) GO TO 80
        median = (median + LESSER) / 2
        FAULT = -9
        RETURN
      END IF
      GO TO 70
      
!---- ------------------------------------------------------------------
!       Even N need special treatment.  Find the next lesser number in the 
!      array after median
  60  CONTINUE

!            look for the next lesser value than PIVOT
      IF (MID > LOOLD) THEN
!        PRINT *, 'Searching current work array:...'
        IF (INO) THEN
          DO J = LOOLD, MID
!            print *, 'o(',J,') =', O(J)
            IF ( SKIP .AND. ABS(O(J) - median) .LE. MACHEP ) THEN
              SKIP = .FALSE.
              CYCLE
            END IF
            IF ( (O(J) > LESSER) .AND. (O(J) .LE. median)) LESSER = O(J)
          END DO
        ELSE
          DO J = LOOLD, MID
!            print *, 'e(',J,') =', E(J)
            IF ( SKIP .AND. ABS(E(J) - median) .LE. MACHEP ) THEN
              SKIP = .FALSE.
              CYCLE
            END IF
            IF ( (E(J) > LESSER) .AND. (E(J) .LE. median)) LESSER = E(J)
          END DO
        END IF
*       print *, 'Went', I, ' iter: found pivot ',PIVOT, 'less ', LESSER
    
        IF (T > MACHEP) GO TO 80
        median = (median + LESSER) / 2
        FAULT = -10
        RETURN
      END IF

!---- ------------------------------------------------------------------      
!        Search an older work array
  70  CONTINUE

!      PRINT *, 'Searching older (L2) work array'
      IF (INOOLD) THEN
        CALL SHELL(O(L2), (H2-L2+1))
        LESSER = O(MID-1)
      ELSE
        CALL SHELL(E(L2), (H2-L2+1))
        LESSER = E(MID-1)
      END IF
      
      IF (T > MACHEP) GO TO 80
      median = (median + LESSER) / 2
      FAULT = -11
      RETURN        

*-----------------------------------------------------------------------
  80   CONTINUE      ! modulo arithmetic
*        split the difference      
       DIF = MOD(median - LESSER, T)
       IF (DIF > ABS(T/2) ) DIF = DIF - ABS(T) 
       median = LESSER + DIF/2
      
*        correct sign
       IF (LESSER .GE. 0 .AND. median < 0) median = median + ABS(T)
       FAULT = -12

      END  ! of median


!---- ------------------------------------------------------------------
*       rank - a variation of partial sorting to find an arbitrary element
*      of an array were the array to be sorted
*             X                              array
*             N                              length of array
*             R                              position of element
*             fault                          error code
!---- ------------------------------------------------------------------      
      FUNCTION rank(X, N, R, FAULT)
      
      IMPLICIT NONE
      INTEGER N, R, FAULT
      REAL X(N)
      REAL rank, MED3

!         Local variables
      REAL E, O ; DIMENSION E(N), O(N)            ! work arrays
      INTEGER I, J, K                             ! counters
      INTEGER HI, LO, HIOLD, LOOLD                ! array offsets
      REAL PIVOT, machine
      INTEGER nbits                             ! bits required to represent n
      LOGICAL INO                                 ! data is in O
      INTEGER SIP                                 ! sort instead parameter
      REAL MIP                                    ! maximum iteration parameter

!      can't use .EQ. operator on floating-point numbers
      REAL MACHEP 

!   these parameters barely affect performance, adjust them however you like
      DATA SIP, MIP / 3, 1.62 /   

!---- ------------------------------------------------------------------
      MACHEP = machine(1)
      FAULT = 0                       ! clear error codes

!      validate
      IF (R > N .OR. R < 1) THEN
        rank = 0.0
        FAULT = 2
        RETURN
      END IF 
 
!      special cases for small N
      IF (N < 1) THEN
        rank = 0.0
        FAULT = 1
        RETURN
      ELSE IF (N .EQ. 1) THEN
        rank= X(1)
        RETURN
      ELSE IF (N .EQ. 2) THEN
        IF (R .EQ. 1) THEN
          rank = MIN(X(1), X(2))
        ELSE IF (R .EQ. 2) THEN
          rank = MAX(X(1), X(2))
        ELSE 
          STOP 'rank: unexpected error'
        END IF
        RETURN
      ELSE IF (N .EQ. 3) THEN
        IF (R .EQ. 1) THEN
          rank = MIN(X(1), X(2), X(3))
        ELSE IF (R .EQ. 2) THEN
          rank = MED3(X(1), X(2), X(3))
        ELSE IF (R .EQ. 3) THEN
          rank = MAX(X(1), X(2), X(3))
        ELSE
          STOP 'rank: unexpected error'
        END IF
        RETURN
      END IF

!---- ------------------------------------------------------------------
*         bits required to represent n
       i = BIT_SIZE(n)
  10   i = i - 1
       IF (.NOT. BTEST(n, i)) GO TO 10
       nbits = i + 1

!           find pivot and limits
      LO = 1 ; HI = N ; LOOLD = LO ; HIOLD = HI
      PIVOT = MED3(X(LO), X(R), X(HI))
      INO = .FALSE.
      
!      sort data from X into work array E
      DO J = 1, N

!        sort values larger than PIVOT to the rear
        IF ( X(J) > PIVOT) THEN
          E(HI) = X(J)
          HI = HI - 1
          
!         values less than or equal to PIVOT come to the front        
        ELSE
          E(LO) = X(J)
          LO = LO + 1
        END IF
      END DO
        
!      find which partition 
!        HI and LO have crossed.  test HI < R ?
      IF (HI < R) THEN
        HI = HIOLD
        GO TO 30
        
!          the lesser than or equal to partition         
      ELSE IF (HI > R) THEN

!            check for equality of all values from R to HI
        DO K = R, HI
          IF ( ABS(E(K) - PIVOT) > MACHEP) THEN
            LO = LOOLD
            GO TO 30
          END IF
        END DO
      END IF

!         found it in one pass 
      rank = PIVOT
      RETURN
      
!---- ------------------------------------------------------------------
!      alternate between putting data into O and putting data into E
!          Beginning of the main loop
!---- ------------------------------------------------------------------
  30  DO I = 1, INT(MIP * nbits) + 1   ! no significance to this limit

        LOOLD = LO ; HIOLD = HI
        INO = .NOT. INO 
        
!           odd i
        IF (INO) THEN
!           PIVOT = MED3(E(LO), E( (LO+HI)/2 ), E(HI))  ! standard method
          PIVOT = E(LO + MOD(N*I, (HI-LO)) )    ! alternative pivoting
!            error check
      IF ((MOD(N*I, (HI-LO)) + LO) > HI) FAULT = 2
          DO J = LO, HI                                  ! sort
            IF ( E(J) > PIVOT) THEN
              O(HI) = E(J)
              HI = HI - 1
            ELSE
              O(LO) = E(J)
              LO = LO + 1
            END IF
          END DO  ! next j

          IF (HI < R) THEN                     !  find which partition
            HI = HIOLD
            GO TO 40
          ELSE IF (HI > R) THEN
            DO K = R, HI
              IF ( ABS(O(K) - PIVOT) > MACHEP ) THEN            
                LO = LOOLD
                GO TO 40
              END IF
            END DO
          END IF
          rank = PIVOT
          RETURN
        
!---- ------------------------------------------------------------------
!         even i        
        ELSE   
!            use median pivoting on the even passes
!             better for almost-sorted data           
          PIVOT = MED3(O(LO), O((HI+LO)/2), O(HI))
          
          DO J = LO, HI        ! sort
            IF ( O(J) > PIVOT) THEN
              E(HI) = O(J)
              HI = HI - 1
            ELSE
              E(LO) = O(J)
              LO = LO + 1
            END IF
          END DO  ! next j

!      find which partition
          IF (HI < R) THEN
            HI = HIOLD
            GO TO 40
          ELSE IF (HI > R) THEN
            DO K = R, HI
              IF ( ABS(E(K) - PIVOT) > MACHEP ) THEN
                LO = LOOLD
                GO TO 40
              END IF
            END DO
          END IF
          rank = PIVOT
          RETURN
        END IF        ! even or odd i
            
!---- ------------------------------------------------------------------
  40    CONTINUE 
        IF ( (HI-LO) < SIP) GO TO 70
      END DO  ! next i

!---- ------------------------------------------------------------------
!          sort the stubborn
  70  IF (INO) THEN 
        CALL SHELL(O(LO), (HI-LO+1))
        rank = O(R)
      ELSE
        CALL SHELL(E(LO), (HI-LO+1))
        rank = E(R)
      END IF

      RETURN
      END     ! of rank

!-----------------------------------------------------------------------
!       find the median of 3 numbers
!-----------------------------------------------------------------------
      FUNCTION MED3 (A, B, C)
      
      IMPLICIT NONE
      REAL A, B, C, MED3
      
      IF (A > B) THEN
        IF (B > C) THEN
          MED3 = B 
        ELSE 
          IF (A > C) THEN
            MED3 = C
          ELSE 
            MED3 = A
          END IF
        END IF
      ELSE
        IF (B .LE. C) THEN
          MED3 = B
        ELSE 
          IF (A > C) THEN
            MED3 = A
          ELSE
            MED3 = C
          END IF
        END IF
      END IF
      
      RETURN
      END     ! of MED3


*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
*      get the mode of an array of real numbers
*         X                    an array         Real
*         n                    length           Integer
*         fault                error code       Integer   (not implemented)
*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      FUNCTION mode(X, n, fault)
      
      IMPLICIT NONE
      INTEGER n, fault
      REAL X, mode
      DIMENSION X(n)
      
      INTEGER i, r, bound, bins, tie, numties, pop, bigpop
      REAL Z, MACHEP, machine
      DIMENSION Z(n), bound(n), tie(n), pop(n)
      SAVE MACHEP ; DATA MACHEP /-1/
      
*      begin executable statements
      fault = 0
      IF (MACHEP < 0) MACHEP = machine(1)
      
*      special case
      IF (n .EQ. 1) THEN
        mode = X(1)
        RETURN
      END IF            
      
*      initialize      
      bound(1) = 1
      bins = 1
      bigpop = 0
      
*         copy data to Z to protect X      
      DO i = 1, n
        Z(i) = X(i)
      END DO
      
      CALL ssort(Z, n)
      
*       find duplicates, mark positions and populations
      DO i = 2, n
        IF ( ABS(Z(i) - Z(i-1)) < MACHEP ) CYCLE
        r = i - bound(bins)
        pop(bins) = r
        IF (r > bigpop) bigpop = r
        bins = bins + 1
        bound(bins) = i
      END DO
      pop(bins) = n - bound(bins) + 1
      IF (pop(bins) > bigpop) bigpop = pop(bins)
      
*         compare populations to biggest population, in case of ties      
      numties = 0
      DO i = 1, bins
        IF (pop(i) < bigpop) CYCLE
        numties = numties + 1
        tie(numties) = i
      END DO

!         check:
!      PRINT *, 'The mode is a ', numties, ' way tie.'
      
*         choose a random popular value      
      r = MIN(INT(RAND(0) * numties) +1, numties)
      mode = Z(bound(tie(r)))

      END  ! of mode

*++++++++++++++++++++++++++End of file average.f++++++++++++++++++++++++
