*-----------------------------------------------------------------------
*        EIGENVALUES AND EIGENVECTORS OF SYMMETRIC REAL MATRICES
*                           from EISPACK
*          revised by Alfred Morris, Naval Surface Warfare Center
*                     This file is public domain.
*-----------------------------------------------------------------------
*  fewer arguments, T made automatic, reformatted comments  19 Dec 2011 AJA
*-----------------------------------------------------------------------
*     A     the matrix
*     M     row dimension of A - (how much space is allocated for the array)
*     N     order of A - (how much space the matrix occupies)
*     W     eigenvalues
*     Z     eigenvectors
*     IERR  error code
*-----------------------------------------------------------------------
      SUBROUTINE SEIGV (A, M, N, W, Z, IERR)

       REAL A(M,N), W(N), Z(M,N), T(N)
       CALL TRED2 (M, N, A, W, T, Z)
       CALL TQL2 (M, N, W, T, Z, IERR)

      END

*-----------------------------------------------------------------------
*
*     This subroutine is a translation of the algol procedure TRED2,
*     Num. Math. 11, 181-195(1968) By Martin, Reinsch, and Wilkinson.
*     Handbook for Auto. Comp., Vol.II-Linear Algebra, 212-226(1971).
*
*     This subroutine reduces a real symmetric matrix to a
*     symmetric tridiagonal matrix using and accumulating
*     orthogonal similarity transformations.
*
*     On Input-
*
*        NM must be set to the row dimension of two-dimensional
*          array parameters as declared in the calling program
*          dimension statement,
*
*        N is the order of the matrix,
*
*        A contains the real symmetric input matrix.  Only the
*          lower triangle of the matrix need be supplied.
*
*     On Output-
*
*        D contains the diagonal elements of the tridiagonal matrix,
*
*        E contains the subdiagonal elements of the tridiagonal
*          matrix in its last N-1 positions.  E(1) is set to zero,
*
*        Z contains the orthogonal transformation matrix
*          produced in the reduction,
*
*        A and Z may coincide.  If distinct, A is unaltered.
*
*-----------------------------------------------------------------------
      SUBROUTINE TRED2(NM,N,A,D,E,Z)

      INTEGER I,J,K,L,N,II,NM,JP1
      REAL A(NM,N),D(N),E(N),Z(NM,N)
      REAL F,G,H,HH,SCALE

      DO 100 I = 1, N

         DO 100 J = 1, I
            Z(I,J) = A(I,J)
  100 CONTINUE

      IF (N .EQ. 1) GO TO 320
*     ********** FOR I=N STEP -1 UNTIL 2 DO -- **********
      DO 300 II = 2, N
         I = N + 2 - II
         L = I - 1
         H = 0.0
         SCALE = 0.0
         IF (L < 2) GO TO 130
*     ********** Scale row (algol tol then not needed) **********
         DO 120 K = 1, L
  120    SCALE = SCALE + ABS(Z(I,K))

         IF (SCALE .NE. 0.0) GO TO 140
  130    E(I) = Z(I,L)
         GO TO 290

  140    DO 150 K = 1, L
            Z(I,K) = Z(I,K) / SCALE
            H = H + Z(I,K) * Z(I,K)
  150    CONTINUE

         F = Z(I,L)
         G = SQRT(H)
         IF (F .GE. 0.0) G = -G
         E(I) = SCALE * G
         H = H - F * G
         Z(I,L) = F - G
         F = 0.0

         DO 240 J = 1, L
            Z(J,I) = Z(I,J) / H
            G = 0.0
*     ********** form element of A*U **********
            DO 180 K = 1, J
  180       G = G + Z(J,K) * Z(I,K)

            JP1 = J + 1
            IF (L < JP1) GO TO 220

            DO 200 K = JP1, L
  200       G = G + Z(K,J) * Z(I,K)
*     ********** form element of P **********
  220       E(J) = G / H
            F = F + E(J) * Z(I,J)
  240    CONTINUE

         HH = F / (H + H)
*     ********** form reduced A **********
         DO 260 J = 1, L
            F = Z(I,J)
            G = E(J) - HH * F
            E(J) = G

            DO 260 K = 1, J
               Z(J,K) = Z(J,K) - F * E(K) - G * Z(I,K)
  260    CONTINUE

  290    D(I) = H
  300 CONTINUE

  320 D(1) = 0.0
      E(1) = 0.0
*     ********** accumulation of transformation matrices **********
      DO 500 I = 1, N
         L = I - 1
         IF (D(I) .EQ. 0.0) GO TO 380

         DO 360 J = 1, L
            G = 0.0

            DO 340 K = 1, L
  340       G = G + Z(I,K) * Z(K,J)

            DO 360 K = 1, L
               Z(K,J) = Z(K,J) - G * Z(K,I)
  360    CONTINUE

  380    D(I) = Z(I,I)
         Z(I,I) = 1.0
         IF (L < 1) GO TO 500

         DO 400 J = 1, L
            Z(I,J) = 0.0
            Z(J,I) = 0.0
  400    CONTINUE

  500 CONTINUE

      RETURN
      END ! of TRED2

*-----------------------------------------------------------------------
*
*     This subroutine is a translation of the algol procedure TQL2,
*     Num. Math. 11, 293-306(1968) By Bowlder, Martin, Reinsch, and
*     Wilkinson.
*     Handbook for Auto. Comp., Vol.II-Linear Algebra, 227-240(1971).
*
*     This subroutine finds the eigenvalues and eigenvectors
*     of a symmetric tridiagonal matrix by the QL method.
*     The eigenvectors of a full symmetric matrix can also
*     be found if TRED2 has been used to reduce this
*     full matrix to tridiagonal form.
*
*     On Input-
*
*        NM must be set to the row dimension of two-dimensional
*          array parameters as declared in the calling program
*          dimension statement,
*
*        N is the order of the matrix,
*
*        D contains the diagonal elements of the input matrix,
*
*        E contains the subdiagonal elements of the input matrix
*          in its last N-1 positions.  E(1) is arbitrary,
*
*        Z contains the transformation matrix produced in the 
*          reduction by TRED2, if performed.  If the eigenvectors
*          of the tridiagonal matrix are desired, Z must contain
*          the identity matrix.
*
*      On Output-
*
*        D contains the eigenvalues in ascending order.  If an
*          error exit is made, the eigenvalues are correct but
*          unordered for indices 1,2,...,IERR-1,
*
*        E has been destroyed,
*
*        Z contains orthonormal eigenvectors of the symmetric
*          tridiagonal (or full) matrix.  If an error exit is made,
*          Z contains the eigenvectors associated with the stored
*          eigenvalues,
*
*        IERR is set to       
*          zero       for normal return,
*          J          if the J-th eigenvalue has not been
*                     determined after 30 iterations.
*
*-----------------------------------------------------------------------
      SUBROUTINE TQL2(NM,N,D,E,Z,IERR)

      INTEGER I,J,K,L,M,N,II,L1,NM,MML,IERR
      REAL D(N),E(N),Z(NM,N)
      REAL B,C,F,G,H,P,R,S,MACHEP
      REAL machine

*     ********** MACHEP is a machine dependent parameter.  Assign
*                MACHEP the value U where U is the smallest positive
*                floating point number such that 1.0 + U > 1.0 
*                **********
      MACHEP = machine(1)

      IERR = 0
      IF (N .EQ. 1) GO TO 1001

      DO 100 I = 2, N
  100 E(I-1) = E(I)

      F = 0.0
      B = 0.0
      E(N) = 0.0

      DO 240 L = 1, N
         J = 0
         H = MACHEP * (ABS(D(L)) + ABS(E(L)))
         IF (B < H) B = H
*     ********** look for small sub-diagonal element **********
         DO 110 M = L, N
            IF (ABS(E(M)) .LE. B) GO TO 120
*     ********** E(N) is always zero, so there is no exit
*                through the bottom of the loop **********
  110    CONTINUE
  
*               added 2011/6/13  
         if (m > n) THEN 
           ierr = -1
           PRINT *, 'TQL2:  Couldn''t find small subdiagonal element'
           M = N
         END IF

  120    IF (M .EQ. L) GO TO 220
  130    IF (J .EQ. 30) GO TO 1000
         J = J + 1
*     ********** form shift **********
         L1 = L + 1
         G = D(L)
         P = (D(L1) - G) / (2.0 * E(L))
         R = SQRT(P*P+1.0)
         IF (P < 0.0) R = -R
         D(L) = E(L) / (P + R)
         H = G - D(L)

         DO 140 I = L1, N
  140    D(I) = D(I) - H

         F = F + H
*     ********** QL transformation **********
         P = D(M)
         C = 1.0
         S = 0.0
         MML = M - L
*     ********** FOR I=M-1 STEP -1 UNTIL L DO -- **********
         DO 200 II = 1, MML
            I = M - II
            G = C * E(I)
            H = C * P
            IF (ABS(P) < ABS(E(I))) GO TO 150
            C = E(I) / P
            R = SQRT(C*C+1.0)
            E(I+1) = S * P * R
            S = C / R
            C = 1.0 / R
            GO TO 160
  150       C = P / E(I)
            R = SQRT(C*C+1.0)
            E(I+1) = S * E(I) * R
            S = 1.0 / R
            C = C * S
  160       P = C * D(I) - S * G
            D(I+1) = H + S * (C * G + S * D(I))
*     ********** form vector **********
            DO 180 K = 1, N
               H = Z(K,I+1)
               Z(K,I+1) = S * Z(K,I) + C * H
               Z(K,I) = C * Z(K,I) - S * H
  180       CONTINUE

  200    CONTINUE

         E(L) = S * P
         D(L) = C * P
         IF (ABS(E(L)) > B) GO TO 130
  220    D(L) = D(L) + F
  240 CONTINUE
*     ********** order eigenvalues and eigenvectors **********
      DO 300 II = 2, N
         I = II - 1
         K = I
         P = D(I)

         DO 260 J = II, N
            IF (D(J) .GE. P) GO TO 260
            K = J
            P = D(J)
  260    CONTINUE

         IF (K .EQ. I) GO TO 300
         D(K) = D(I)
         D(I) = P

         DO 280 J = 1, N
            P = Z(J,I)
            Z(J,I) = Z(J,K)
            Z(J,K) = P
  280    CONTINUE

  300 CONTINUE

      GO TO 1001
*     ********** set error -- no convergence to an
*                eigenvalue after 30 iterations **********
 1000 IERR = L
 1001 RETURN
      END ! of TQL2


*-----------------------------------------------------------------------
*
*     This subroutine is a translation of the algol procedure IMTQL2,
*     Num. Math. 12, 377-383(1968) by Martin and Wilkinson,
*     as modified in Num. Math. 15, 450(1970) by Dubrulle,
*     Handbook for Auto. Comp., Vol.II-Linear Algebra, 241-248(1971).
*
*     This subroutine finds the eigenvalues and eigenvectors
*     of a symmetric tridiagonal matrix by the implicit QL method.
*     The eigenvectors of a full symmetric matrix can also
*     be found if TRED2 has been used to reduce this
*     full matrix to tridiagonal form.
*
*     On Input-
*
*        NM must be set to the row dimension of two-dimensional
*          array parameters as declared in the calling program
*          dimension statement,
*
*       N is the order of the matrix,
*
*        D contains the diagonal elements of the input matrix,
*
*        E contains the subdiagonal elements of the input matrix
*          in its last N-1 positions.  E(1) is arbitrary,
*
*        Z contains the transformation matrix produced in the
*          reduction by TRED2, if performed.  If the eigenvectors
*          of the tridiagonal matrix are desired, Z must contain
*          the identity matrix.
*
*      On Output-
*
*        D contains the eigenvalues in ascending order.  If an
*          error exit is made, the eigenvalues are correct but
*          unordered for indices 1,2,...,ierr-1,
*
*        E has been destroyed,
*
*        Z contains orthonormal eigenvectors of the symmetric
*          tridiagonal (or full) matrix.  If an error exit is made,
*          Z contains the eigenvectors associated with the stored
*          eigenvalues,
*
*        IERR is set to
*          zero       for normal return,
*          J          if the J-th eigenvalue has not been
*                     determined after 30 iterations.
*
*-----------------------------------------------------------------------
      SUBROUTINE IMTQL2(NM,N,D,E,Z,IERR)

      INTEGER I,J,K,L,M,N,II,NM,MML,IERR
      DOUBLE PRECISION D(N),E(N),Z(NM,N)
      DOUBLE PRECISION B,C,F,G,P,R,S,MACHEP
      DOUBLE PRECISION dmachine

*     ********** MACHEP is a machine dependent parameter.  Assign
*                MACHEP the value U where U is the smallest positive
*                floating point number such that 1.0 + U > 1.0 .
*                **********
      MACHEP = dmachine(1)

      IERR = 0
      IF (N .EQ. 1) GO TO 1001

      DO 100 I = 2, N
  100 E(I-1) = E(I)

      E(N) = 0.0

      DO 240 L = 1, N
         J = 0
*     ********** look for small sub-diagonal element **********
  105    DO 110 M = L, N
            IF (M .EQ. N) GO TO 120
            IF (ABS(E(M)) .LE. MACHEP * (ABS(D(M)) + ABS(D(M+1))))
     &         GO TO 120
  110    CONTINUE

  120    P = D(L)
         IF (M .EQ. L) GO TO 240
         IF (J .EQ. 30) GO TO 1000
         J = J + 1
*     ********** form shift **********
         G = (D(L+1) - P) / (2.0 * E(L))
         R = SQRT(G*G+1.0)
         IF (G  <  0.0) R = -R
         G = D(M) - P + E(L) / (G + R)
         S = 1.0
         C = 1.0
         P = 0.0
         MML = M - L
*     ********** FOR I=M-1 STEP -1 UNTIL L DO -- **********
         DO 200 II = 1, MML
            I = M - II
            F = S * E(I)
            B = C * E(I)
            IF (ABS(F) < ABS(G)) GO TO 150
            C = G / F
            R = SQRT(C*C+1.0)
            E(I+1) = F * R
            S = 1.0 / R
            C = C * S
            GO TO 160
  150       S = F / G
            R = SQRT(S*S+1.0)
            E(I+1) = G * R
            C = 1.0 / R
            S = S * C
  160       G = D(I+1) - P
            R = (D(I) - G) * S + 2.0 * C * B
            P = S * R
            D(I+1) = G + P
            G = C * R - B
*     ********** form vector **********
            DO 180 K = 1, N
               F = Z(K,I+1)
               Z(K,I+1) = S * Z(K,I) + C * F
               Z(K,I) = C * Z(K,I) - S * F
  180       CONTINUE

  200    CONTINUE

         D(L) = D(L) - P
         E(L) = G
         E(M) = 0.0
         GO TO 105
  240 CONTINUE
*     ********** order eigenvalues and eigenvectors **********
      DO 300 II = 2, N
         I = II - 1
         K = I
         P = D(I)

         DO 260 J = II, N
            IF (D(J) .GE. P) GO TO 260
            K = J
            P = D(J)
  260    CONTINUE

         IF (K .EQ. I) GO TO 300
         D(K) = D(I)
         D(I) = P

         DO 280 J = 1, N
            P = Z(J,I)
            Z(J,I) = Z(J,K)
            Z(J,K) = P
  280    CONTINUE

  300 CONTINUE

      GO TO 1001
*     ********** set error -- no convergence to an
*                eigenvalue after 30 iterations **********
 1000 IERR = L
 1001 RETURN
      END ! of IMTQL2


* the last routine is not from EISPACK
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
* bla - Basic Lanczos Algorithm 
*        find least eigenvectors of a symmetric matrix in triad format 
* 
*  by Andy Allinger, 2012, released to the public domain 
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*  This program may be used by any person for any purpose.
*  It comes with ABSOLUTELY NO WARRANTY.
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
* origin:  Cornelius Lanczos, 1950
*
* refer to:
*    en.wikipedia.org   "Lanczos Algorithm" and "Gram-Schmidt Orthogonalization"
*
*    Beresford N. Parlett and David S. Scott
*    "The Lanczos Algorithm with Selective Orthogonalization"
*    draft report, [1979]
*    Mathematics of Computation v.33 n.145 pp.217-238, January 1979
*
* also the subroutine packages:
*    LASO, by Scott and Parlett, 1983
*    MINVAL, by Richard Ray Underwood, 1975
*
* X must be symmetric such that X[i,j] = X[j,i]
* Its array representation however is X[i]   1 .LE. i .LE. m
*                                     IX[i]    "        "
*                                     JX[i]    "        "
* where each X value is a datum, IX is its row position, and JX its column.
* The entries may be given in any order.  For each pair of entries
*                 X[i,j], X[j,i]       i .NE. j
* one or the other may be specified, but don't include both.
*
* The random number generator is assumed to be already initialized.   
* Make sure to call SRAND() in the main program.
*
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*                                I/O LIST
*  Name             Type             In/Out        Description
*  X[m]             DOUBLE PRECISION in            nonzero entries of matrix 
*  IX[m]            INTEGER          in            row indices of X entries
*  JX[m]            INTEGER          in            column indices of X entries
*  m                INTEGER          in            length of X
*  n                INTEGER          in            order of matrix
*  nlim             INTEGER          in            row dimension of Z
*  kreqd            INTEGER          in            number desired eigenvalues
*  L[13@kreqd]      DOUBLE PRECISION out           eigenvalues
*  Z[nlim,13@kreqd] DOUBLE PRECISION out           eigenvectors
*  fault            INTEGER          out           error code     0: OK
*                                                          negative: incomplete
*                                                          positive: error
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
      SUBROUTINE bla(X, IX, JX, m, n, nlim, kreqd, L, Z, fault)
      
       IMPLICIT NONE
       INTEGER  IX, JX, m, n, nlim, kreqd, fault, luck
       PARAMETER (luck = 7)   ! no guarrantees in this algorithm, just wishes
       DOUBLE PRECISION X, L, Z
       DIMENSION X(m), IX(m), JX(m), 
     &         L(luck * kreqd), Z(nlim, luck * kreqd)

*            constants
       INTEGER jlim,                ! number of Lanczos steps allowed
     &         maxiter              ! maximum number of iterations
       PARAMETER (jlim = 500,
     &            maxiter = 1 000 000)
     
*            local variables
       INTEGER 
     &      iter,                  ! iteration count
     &      h, i, j, o, ii, jj,    ! counters
     &      jn,                    ! in case jlim < n.  jn = MIN(jlim, n)
     &      k, ktot,               ! number of convergent, acceptable vectors
     &      kg,                    ! number good vectors since restart
     &      ks, ksold, ky,         ! saved good vectors in Z and in Y
     &      v,                     ! columns in matrix Z 
     &      wlim                   ! # eigenvalues in latest decomposition
     
       DOUBLE PRECISION epsilon, tolg, tolex, tolgx   ! precisions
       DOUBLE PRECISION 
     &      A(jlim), aa,                  ! alpha
     &      asmall, abig, bbig, bsmall,   ! extreme values of alpha and beta
     &      B(0:jlim), bb,                ! beta
     &      E(0:jlim),                    ! backup copy of beta
     &      hold,                         ! temporary scalar
     &      dmachine,                     ! external function
     &      norm,                         ! length of a column vector
     &      p                             ! projection 
       DOUBLE PRECISION
     &      Q(n,jlim+1),                  ! Lanczos vectors
     &      r,                            ! residual B[j] @ |S[j,i]|
     &      S(jlim,jlim),                 ! eigenvectors of T
     &      U(n),                         ! temporary vector
     &      W(jlim), wlast,               ! eigenvalues
     &      wsmall, wbig,                 ! extreme eigenvalues
     &      xnorm,                        ! ||X||
     &      xi,
     &      Y(n,luck * kreqd),            ! good Ritz vectors
     &      zeta, eta, theta              ! error bound recurrance terms

       LOGICAL
     &      G(jlim),            ! whether a Ritz vector is good
     &      consec,             ! no gaps in sequence of convergent vectors
     &      test                ! whether a test run for multiplicities
     
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*               begin
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
       fault = 0
       iter = 0
       jn = MIN(jlim, n)
       v = luck * kreqd

       epsilon = dmachine(1)
       tolg = SQRT(epsilon)                               ! goodness tolerance
       
       ks = 0        ! nothing converged and saved yet
       ktot = 0      ! nothing accepted yet
       wlast = -dmachine(3)  ! least possible eigenvalue
       test = .FALSE.
       
*          vet input
       IF (kreqd .GE. jn) THEN
         fault = 4
         PRINT *, 'lanso: jlim too low'
         RETURN
       END IF       

*           find norm of X
       norm = 0
       DO i = 1, m
         hold = X(i)
         norm = norm + hold * hold
       END DO
       xnorm = SQRT(norm)       
       tolex = epsilon * xnorm
       tolgx = tolg * xnorm 
       wsmall = 0 ; wbig = 0

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  10   CONTINUE    ! choose q[1] as a random unit vector
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
       DO i = 1, n                     ! q[1] = random vector
         Q(i,1) = RAND(0) - 0.5
       END DO       

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
  20   CONTINUE      ! come here if q[1] is saved from a previous step         
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
       DO h = 1, ks        ! orthogonalize against saved converged vectors
         p = 0
         DO i = 1, n
           p = p + Q(i,1) * Z(i,h)
         END DO
         DO i = 1, n
           Q(i,1) = Q(i,1) - p * Z(i,h)
         END DO
       END DO
       norm = 0                     ! normalize q[1]
       DO i = 1, n
         hold = Q(i,1)
         norm = norm + hold * hold
       END DO
       norm = SQRT(norm)
       DO i = 1, n
         Q(i,1) = Q(i,1) / norm
       END DO

       wlim = 0
       B(0) = 0
        
*           no Ritz vectors are good
       DO i = 1, jn 
         G(i) = .FALSE.
       END DO
       k = 0          ! no accepted vectors
       ky = 0         ! no good vectors
       zeta = epsilon ; eta = epsilon ; theta = epsilon
              
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
*             main loop
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
       DO j = 1, jn
         iter = iter + 1
         PRINT *, 'beginning iteration ', iter
 
*          A Lanczos step
*            u[j]: X q[j] - b[j-1] q[j-1]
*            sparse symmetric matrix multiplication
         DO i = 1, n ; U(i) = 0 ; END DO
         DO i = 1, m
           U(IX(i)) = U(IX(i)) + X(i) * Q(JX(i),j)
           IF (IX(i) .EQ. JX(i)) CYCLE   ! don't double-count the diagonal
           U(JX(i)) = U(JX(i)) + X(i) * Q(IX(i),j)
         END DO
       
         IF (j > 1) THEN
           bb = B(j-1)
           DO i = 1, n
             U(i) = U(i) - bb * Q(i,j-1)
           END DO
         END IF

*            a[j]: q[j] . u[j]
         aa = 0
         DO i = 1, n
           aa = aa + Q(i,j) * U(i)
         END DO
         A(j) = aa

*            u[j]: u[j] - a[j] q[j]
         DO i = 1, n
           U(i) = U(i) - aa * Q(i,j)
         END DO

*            b[j]: ||u[j]||
         norm = 0
         DO i = 1, n
           hold = U(i)
           norm = norm + hold * hold
         END DO
         bb = SQRT(norm)
         B(j) = bb

*      if b[j] = 0 stop, else q[j+1]: u[j] / b[j]
         IF (bb .EQ. 0) THEN
           fault = 1
           IF (ktot > 0) fault = -ktot
           PRINT *, 'Lanczos algorithm halts'
           RETURN
         END IF
         DO i = 1, n
           Q(i,j+1) = U(i) / bb
         END DO
*       Lanczos step done         

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
*           full orthogonalization against prior Q
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
         DO h = 1, j
           p = 0
           DO i = 1, n
             p = p + Q(i,j+1) * Q(i,h)
           END DO
           DO i = 1, n
             Q(i,j+1) = Q(i,j+1) - p * Q(i,h)
           END DO
         END DO 

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
*          orthogonalize against good Ritz vectors
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
*            q[j]' = q[j] - ((q[j] . y[i]) / (y[i] . y[i])) @ y[i]
         DO i = 1, ky
           p = 0
           DO o = 1, n 
             p = p + Q(o,j+1) * Y(o,i)
           END DO
           IF (p > n * tolgx) THEN
             PRINT *, 'bla: orthogonality lost, proj=', p
             fault = 3
             RETURN
           END IF
           DO o = 1, n
             Q(o,j+1) = Q(o,j+1) - p * Y(o,i)
           END DO
         END DO
         
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*            orthogonalize against saved vectors
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ 
         DO h = 1, ks             ! saved converged vectors
           p = 0
           DO o = 1, n
             p = p + Q(o,j+1) * Z(o,h)
           END DO
           DO o = 1, n
             Q(o,j+1) = Q(o,j+1) - p * Z(o,h)
           END DO
         END DO
         
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*            check the scalar error bound
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*             find extreme values
         IF (j .EQ. 1) THEN
           abig = aa ; asmall = aa
           bbig = bb ; bsmall = bb
         ELSE          
           abig = MAX(abig, aa) ; asmall = MIN(asmall, aa) ! extreme A
           bbig = MAX(bbig, bb) ; bsmall = MIN(bsmall, bb) ! extreme B
         END IF

*            arcane formulae -- see LASO2
         xi = MAX(abig - wsmall, wbig - asmall)
         theta = (eta * xi + zeta * bbig + tolex) / bsmall
         zeta = eta ; eta = theta          
         IF (theta < tolg .AND. j < jn) CYCLE

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
*           solve the [j,j] sized tridiagonal eigenproblem
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
         DO i = 1, j               ! back up arguments
           W(i) = A(i)
           E(i) = B(i)
         END DO

         DO h = 1, j ; DO i = 1, j ; S(i,h) = 0. ; END DO ; END DO
         DO i = 1, j ; S(i,i) = 1. ; END DO                ! identity matrix

         CALL IMTQL2(jlim, j, W, E, S, fault)
         IF (fault .NE. 0) THEN 
           PRINT *, 'imtql2: error#', fault
           fault = 6
           RETURN
         END IF
           
         wbig = W(j) ; wsmall = W(1) 
         wlim = j       ! don't forget size of eigenproblem
         
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
*           compare residuals to tolerances
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
         kg = 0                   ! good
         consec = .TRUE.

         DO i = 1, j

*             r[j,i]  = b[j] @ |s[j,i]|
           r = B(j) * ABS(S(j,i))
           
           PRINT 35, i, r, W(i)
           IF (r < tolgx) PRINT 36
           PRINT 39
  35   FORMAT ('R(', I3, ')=', G12.4, ' W=', F14.6, $)
  36   FORMAT (' good', $)
  39   FORMAT ()
           
           IF (r < tolgx) THEN
             G(i) = .TRUE.
             kg = kg + 1
           ELSE
             G(i) = .FALSE.
             IF (consec) k = kg
             consec = .FALSE.
           END IF

         END DO         
         IF (kg .EQ. j) k = j        ! everything converges

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*    check if finished
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
         IF (test) THEN
           IF (k > 0 .AND. W(1) > wlast) RETURN ! success
           IF (k > 1 .AND. W(MAX(1,k)) > wlast) EXIT ! restart multiplicity test
         ELSE IF (k + ktot .GE. kreqd) THEN 
           EXIT                                        ! enough vectors
         ELSE 
           CONTINUE  ! for another iteration
         END IF

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*       compute good Ritz vectors
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
         i = 0         
         DO h = 1, j
           IF (G(h)) THEN 
             i = i + 1
             IF (i > v) EXIT

*                 Y[i] = Q S[i]
             DO ii = 1, n ; Y(ii,i) = 0 ; END DO
             DO jj = 1, j
               hold = S(jj,h)
               DO ii = 1, n
                 Y(ii,i) = Y(ii,i) + Q(ii,jj) * hold
               END DO
             END DO
             
             norm = 0
             DO o = 1, n
               hold = Y(o,i)
               norm = norm + hold * hold
             END DO
             norm = SQRT(norm)

             DO o = 1, n
               Y(o,i) = Y(o,i) / norm
             END DO
           END IF
         END DO                   
         ky = MIN(i, v)   ! remember how full Y is
         
*              orthogonalize Q against Y
         DO i = 1, ky
           p = 0
           DO o = 1, n 
             p = p + Q(o,j) * Y(o,i)
           END DO
           DO o = 1, n
             Q(o,j) = Q(o,j) - p * Y(o,i)
           END DO
           p = 0
           DO o = 1, n 
             p = p + Q(o,j+1) * Y(o,i)
           END DO
           DO o = 1, n
             Q(o,j+1) = Q(o,j+1) - p * Y(o,i)
           END DO
         END DO

*                reset recurrence
         zeta = epsilon ; eta = epsilon ; theta = epsilon
       END DO  ! next j
*        end of main loop

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*    compute new vectors and merge into saved vectors
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
       ksold = ks
       ks = MIN(ks+kg, v)      ! save all converged vectors if enough storage
       jj = 1                 ! # of L values consumed + 1
       o = 1                  ! index in new list       

       DO WHILE (o .LE. wlim)
         IF (G(o)) EXIT
         o = o + 1            ! only point to good eigenvalues
       END DO
       DO h = 1, ks
         IF (o > wlim) THEN
           jj = jj + 1
           CYCLE
         ELSE IF (jj > ksold) THEN
           GO TO 50
         ELSE IF (W(o) < L(h)) THEN
           DO ii = ks-1, h, -1              ! scoot down data
             L(ii+1) = L(ii)
             DO i = 1, n
               Z(i,ii+1) = Z(i,ii)
             END DO
           END DO
         ELSE 
           jj = jj + 1
           CYCLE
         END IF
         
*         next element must be computed from S and Q
  50     L(h) = W(o) 

*              Z[i] = Q S[i]         
         DO i = 1, n ; Z(i,h) = 0 ; END DO
         DO j = 1, wlim
           hold = S(j,o)
           DO i = 1, n
             Z(i,h) = Z(i,h) + Q(i,j) * hold
           END DO
         END DO

*         orthonormalize against prior Z
         DO j = 1, h-1
           p = 0
           DO i = 1, n
             p = p + Z(i,h) * Z(i,j)
           END DO
           DO i = 1, n
             Z(i,h) = Z(i,h) - p * Z(i,j)
           END DO
         END DO 
         norm = 0
         DO i = 1, n
           hold = Z(i,h)
           norm = norm + hold * hold
         END DO
         norm = SQRT(norm)
         DO i = 1, n
           Z(i,h) = Z(i,h) / norm
         END DO

         o = o + 1
         DO WHILE (o .LE. wlim)
           IF (G(o)) EXIT
           o = o + 1  
         END DO         
       END DO  ! next h

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*          nonzero returns
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*              find all eigenvalues on small problems 
       IF (ks .EQ. n) THEN
         fault = -n
         RETURN   !  actually happens
       END IF
         
       IF (iter .EQ. maxiter) THEN
         fault = 2
         RETURN
       END IF
       
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*        enough vectors?
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
       ktot = ktot + k
       IF (ktot .GE. kreqd .AND. .NOT. test) THEN
         test = .TRUE.
         wlast = L(kreqd)
         PRINT *, 'beginning multiplicity test!'
         GO TO 10
       END IF

*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@       
*       new starting vector is combination of Y yet uncomputed
*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ 
       o = 0
       DO i = 1, n ; U(i) = 0. ; END DO
       DO h = k, wlim
         IF (G(h)) CYCLE
         o = o + 1
         DO j = 1, wlim
           hold = S(j,h)
           DO i = 1, n
             U(i) = U(i) + Q(i,j) * hold
           END DO
         END DO
         IF (o .GE. kreqd-ktot) EXIT       ! take the lowest ungood vectors
       END DO
       
       DO i = 1, n                ! add in some randomness
         Q(i,1) = U(i) + (RAND(0) - 0.5) / n  
       END DO

       PRINT *, 'restarting!'           
       GO TO 20
       
      END   ! of bla

*++++++++++++++++++++++++++End of file eigen.f++++++++++++++++++++++++++
