*-----------------------------------------------------------------------
* chord.f :  chordal music analysis by the Pardo-Birmingham algorithm
*      by Andy Allinger, 2011
*      andy_a@users.sourceforge.net
*
*    This program is free software, you may use or redistribute it 
* according to the terms of the Allegro Giftware License.  If you did 
* not receive a copy of this license, visit:  alleg.sf.net
*
*-----------------------------------------------------------------------
*
* based on the algorithm published in:
*     Bryan Pardo & William P. Birmingham, 
*     "The Chordal Analysis of Tonal Music" 
*     University of Michigan technical report, 28 March 2001
*
* and:
*     Bryan Pardo & William P. Birmingham
*     "Algorithms for Chordal Analysis"
*     Computer Music Journal, v. 26 n. 2 pp. 27-49, Summer 2002
*
*-----------------------------------------------------------------------
*    Public domain routines QSORTR, IORDER, and RORDER by Robert Renka
* of Oak Ridge National Laboratory can be used to sort the input, and
* are available as part of the Naval Surface Warfare Center FORTRAN
* library.  This lately has been available from the University of Alberta.
*
*-----------------------------------------------------------------------
*
*                                 I/O LIST
*  NAME              TYPE      In/Out       DESCRIPTION
* note_time[n]      REAL        IN        times of note on/off events
*                                           (assumed already sorted)
* pitch[n]          INTEGER     IN        MIDI note #
* offon[n]          INTEGER     IN        0=OFF, 1=ON
* n                 INTEGER     IN        number of note events
* key_time[k]       REAL        IN        times of key change events
*                                           (assumed already sorted)
* sharps[k]         INTEGER     IN        # sharps in the key signature {-7...7}
*                                           (matches MIDI representation)
* majmin[k]         INTEGER     IN        0=major key, 1=minor key
* k                 INTEGER     IN        number of key change events
* c                 INTEGER     OUT       # of chord segments detected
* partit[n]         REAL        OUT       times of chord changes
* q[n]              INTEGER     OUT       "qualities" of chords
*                                           (as in the templates, see below)
* r[n]              INTEGER     OUT       roots of chords {0...11}
*                                           as pitch % 12
* p                 REAL        OUT       probability of chord sequence
* change[0:12*num_q-1,
*        0:12*num_q-1,
*        0:1]       REAL        IN        probability of chord given a chord
*                                          in Roman numeral notation.  The 
*                                          table is given twice, once
*                                          for major keys, and again for 
*                                          minor keys.  num_q is the number 
*                                          of chord qualities.  A separate 
*                                          routine is needed to define change[].
* x                 LOGICAL     IN        input is training data.  extract
*                                            chord change probabilities
* fault             INTEGER     OUT       error code, 0=success
*
*-----------------------------------------------------------------------
      SUBROUTINE chordal (note_time, pitch, offon, n, key_time, sharps,
     &              majmin, k, c, partit, q, r, p, change, x, fault)
       
       IMPLICIT NONE
       INTEGER num_q ; PARAMETER (num_q = 25)        ! how many chord qualities
       
       INTEGER pitch, offon, n, sharps, majmin, k, c, q, r, fault
       REAL note_time, key_time, partit, p, change
       LOGICAL x
       DIMENSION note_time(n), pitch(n), offon(n), key_time(k), 
     &     sharps(k), majmin(k), partit(n), q(n), r(n)
       DIMENSION change(0:12*num_q-1, 0:12*num_q-1, 0:1)
        
*          local variables
       INTEGER a,                              ! counter
     &        chprev, chnext,                  ! successive chords
     &        cumnew(0:11), cumprev(0:11),     ! cumulative pitch vectors
     &        cumseg(0:11, 0:n),               ! "           "
     &        evals,
     &        hsteps(0:23),                    ! shortcut for % 12
     &        h, i, j, L, m,                   ! counters
     &        label,                           ! GO TO destination
     &        mode,                            ! major or minor key
     &        npc,                             ! pitch % 12
     &        pall,                            ! # possible segmentation points
     &        prev,                            ! pointer to previous segment
     &        qbest, qdel, qkeep, qold, qnew,  ! qualities
     &        rbest, rdel, rkeep, rold, rnew,  ! roots of chords
     &        seg(0:11, 0:n),                  ! pitch vectors, each minimal 
                                               !    segment
     &        skip, 
     &        state(0:11),                     ! pitch vector
     &        template(0:11, 0:num_q-1),       ! chord templates
     &        ties,
     &        tonic,                           ! what key signature
     &        tm12,                            ! tonic transposed to -11...0
     &        weight(0:11),                    ! pitch vector to analyze
     &        wsum                             ! sum of a weight vector
     
       REAL  BIG,                               ! very large number
     &       ctPP, ctPA, ctAP, ctAA,            ! contingency table cells
     &       div,                               ! a divisor
     &       EPS,                               ! machine precision
     &       f, g,                              ! nominal probability
     &       fw, gw,                            ! expected frequency
     &       machine,                           ! external function
     &       sbest, score, sdel, skeep, sprev,  ! possible segment scores
     &       sqroot(32),                        ! square roots of small integers
     &       t, tprev                           ! event times
       
       LOGICAL plist(n),               ! whether a partition is kept
     &         shifted                 ! whether templates have been transposed
       
       SAVE shifted, template, sqroot
       
*-----------------------------------------------------------------------
*       chord templates:  bits 0...11 of an integer represent pitches as
*         MIDI note number % 12
*      If a note is present in the chord, the corresponding bit is set
*      in the template.  Template data is included for a root note of C
*      and transposed to the other keys.  For example, C Major has the notes
*      C, E, G and the template {0, 4, 7}, which as an integer is:
*        2^0 + 2^4 + 2^7 = 145
* 
*      The order of the array is significant, lower numbered qualities
*      have precedence in case of ties during template-matching.
*    
*     You are free to redefine which templates are included.  There is no 
* authority as to what is or is not a real chord.  This list, while not too
* ambitious, should be adequate for most purposes.
*-----------------------------------------------------------------------
*                          integer  name                     template   symbol

*        triads and tetrads      
      DATA template(0, 0) / 145/  ! major                    {0,4,7}    C
      DATA template(0, 1) / 137/  ! minor                    {0,3,7}    Cm
      DATA template(0, 2) /1169/  ! seventh                  {0,4,7,10} C7 
      DATA template(0, 3) /2193/  ! major seventh            {0,4,7,11} Cmaj7
      DATA template(0, 4) /1161/  ! minor seventh            {0,3,7,10} Cm7
      DATA template(0, 5) / 657/  ! sixth                    {0,4,7,9}  C6
      DATA template(0, 6) / 649/  ! minor sixth              {0,3,7,9}  Cm6
      DATA template(0, 7) /  73/  ! diminished               {0,3,6}    Cdim
      DATA template(0, 8) / 273/  ! augmented                {0,4,8}    Caug
      DATA template(0, 9) / 161/  ! suspended fourth         {0,5,7}    Csus4
      DATA template(0,10) /  81/  ! flat fifth               {0,4,6}    C-5
      DATA template(0,11) / 585/  ! diminished seventh       {0,3,6,9}  Cdim7
      DATA template(0,12) /1297/  ! augmented seventh        {0,4,8,10} Caug7
      DATA template(0,13) /1185/  ! seventh suspended fourth {0,5,7,10} C7sus4
      DATA template(0,14) /1105/  ! seventh flat fifth       {0,4,6,10} C7-5 
      DATA template(0,15) / 149/  ! add nine                 {0,2,4,7}  Cadd9
 
*       pentads, hexads
      DATA template(0,16) /1173/  ! ninth                  {0,2,4,7,10}    C9
      DATA template(0,17) /2197/  ! major ninth            {0,2,4,7,11}    Cmaj9
      DATA template(0,18) /1165/  ! minor ninth            {0,2,3,7,10}    Cm9
      DATA template(0,19) /1205/  ! eleventh               {0,2,4,5,7,10}  C11
      DATA template(0,20) /1685/  ! thirteenth             {0,2,4,7,9,10}  C13
      DATA template(0,21) / 661/  ! six nine               {0,2,4,7,9}     C69
      DATA template(0,22) / 653/  ! minor six nine         {0,2,3,7,9}     Cm69
      DATA template(0,23) /1171/  ! seventh flat ninth     {0,1,4,7,10}    C7b9
      DATA template(0,24) /1177/  ! seventh sharp ninth    {0,3,4,7,10}    C7#9

      DATA shifted /.FALSE./
      DATA hsteps / 0,1,2,3,4,5,6,7,8,9,10,11,0,1,2,3,4,5,6,7,8,9,10,11/
      DATA f, g / .20, .02/       ! probabilities of pitches in the template
                                  ! and of pitches not in the template
                               
*-----------------------------------------------------------------------
*         initialize, validate
*-----------------------------------------------------------------------
       fault = 0                 ! clear error codes
       EPS = machine(1)
       BIG = machine(3)
!       lf = LOG(f) ; lg = LOG(g)
       ties = 0 ; evals = 0
       
       IF (n < 2) THEN
         fault = 1               ! at least one NoteOff and NoteOn
         RETURN
       END IF
       
       IF (.NOT. shifted) THEN
       
*        transpose chord templates into all keys
         DO m = 0, num_q - 1
           DO L = 1, 11
             template(L,m) = ISHFTC(template(L-1,m), 1, 12)
           END DO
         END DO
         
*         hoard square roots in memory - 10% time savings on Mozart K159
         DO i = 1, 32 ; sqroot(i) = SQRT(FLOAT(i)) ; END DO
                
         shifted = .TRUE.
       END IF

*-----------------------------------------------------------------------
*      Find minimal segments -initialize states, times, and partition times
*-----------------------------------------------------------------------
       DO L = 0, 11 ; state(L) = 0 ; END DO
       tprev = 0
       partit(1) = 0
       i = 1
       
*-----------------------------------------------------------------------
*          consider each NoteOff or NoteOn as a possible partition point
*-----------------------------------------------------------------------
       DO j = 1, n
         npc = MOD(pitch(j), 12)
         t = note_time(j)
         IF (t > tprev) THEN                         ! different time?
           DO L = 0, 11 ; seg(L, i) = state(L) ; END DO
           tprev = t
           i = i + 1
           partit(i) = t
         END IF
         IF (offon(j) .EQ. 1) THEN
           state(npc) = state(npc) + 1               ! increment pitch vector
         ELSE IF (offon(j) .EQ. 0) THEN
           state(npc) = state(npc) - 1               ! decrement pitch vector
         ELSE 
           fault = 2                  ! bad input
           RETURN
         END IF
       END DO

*-----------------------------------------------------------------------
*       Accumulate the pitch vectors.  This differs from the published 
*       algorithm.  No recursion is required to find the sum of pitch
*       vectors over a range, just a single subtraction.
*-----------------------------------------------------------------------
       pall = i                        ! initialize
       DO j = 0, 11
         seg(j, 0) = 0
         cumprev(j) = 0
         cumseg(j, 0) = 0
       END DO
       
       qbest = 0 ; rbest = 0       ! default initial chord is C
       
       DO i = 1, pall              ! over all minimal segments
         DO j = 0, 11
           cumnew(j) = seg(j,i) + cumprev(j)
           cumseg(j,i) = cumnew(j)
           cumprev(j) = cumnew(j)
         END DO
       END DO
       
*-----------------------------------------------------------------------
*          HarmAn search.  Consider each partition point in order, as 
*       a possible point for a chord change, based on whether keeping the
*       partition or deleting it yields a better match to a chord template.
*-----------------------------------------------------------------------
       a = 1 ; prev = 0
       wsum = 0
       DO j = 0, 11 
         weight(j) = seg(j, 1) 
         wsum = wsum + weight(j)
       END DO
       wsum = MAX(wsum, 1)
       ASSIGN 10 TO label
       GO TO 50
       
  10   q(1) = qbest
       r(1) = rbest
       sprev = sbest
       
*-----------------------------------------------------------------------
*       Begin main loop               
*-----------------------------------------------------------------------
       prev = 1
       DO a = 2, pall-1
         wsum = 0
         DO j = 0, 11  
           weight(j) = seg(j,a) 
           wsum = wsum + weight(j) 
         END DO
         wsum = MAX(wsum, 1)
         ASSIGN 20 TO label
         GO TO 50
         
  20     qkeep = qbest
         rkeep = rbest
         skeep = sbest
         
         wsum = 0
         DO j = 0, 11  
           weight(j) = cumseg(j,a) - cumseg(j,prev-1) 
           wsum = wsum + weight(j)
         END DO
         wsum = MAX(wsum, 1)
         ASSIGN 30 TO label
         GO TO 50
         
  30     qdel = qbest
         rdel = rbest
         sdel = sbest
         
         IF (skeep + sprev > sdel) THEN
           plist(a) = .TRUE.
           q(a) = qkeep
           r(a) = rkeep
           sprev = skeep
           prev = a
         ELSE 
           plist(a) = .FALSE.
           q(prev) = qdel
           r(prev) = rdel
           sprev = sdel
         END IF
       END DO

*-----------------------------------------------------------------------
*         if called with the -X option, count the chord transition probabilities
*        instead of returning chords and the overall probability
*-----------------------------------------------------------------------
       IF (x) GO TO 70

*-----------------------------------------------------------------------
*        remove values of q[] and r[] at deleted partition points, find 
*       probability of the sequence of chord changes, figured as a Markov
*       chain of events E[1]...E[n] :        
*         P(E) = P(E[n]) * PROD ( P(E[i-1] | E[i]) )           2...i...n
*-----------------------------------------------------------------------
       p = 0
       h = 1
       c = 1
       qold = q(1) ; rold = r(1)
       tonic = MOD(7 * sharps(1), 12)
       IF (tonic < 0) THEN
         tm12 = tonic
         tonic = tonic + 12
       ELSE
         tm12 = tonic - 12
       END IF
       mode = majmin(1)
       chprev = hsteps(r(1) - tm12) + 12 * q(1)
       
       DO a = 2, pall - 1
         IF (h < k) THEN                       ! change key
           IF (partit(a) .GE. key_time(h+1)) THEN
             h = h + 1
             tonic = MOD(7 * sharps(h), 12)
             IF (tonic < 0) THEN
               tm12 = tonic
               tonic = tonic + 12
             ELSE
               tm12 = tonic - 12
             END IF
             mode = majmin(h)
           END IF
         END IF
       
         IF ( .NOT. plist(a)) CYCLE           ! not a partition point
         
         qnew = q(a)
         rnew = r(a)
         IF ( qold .EQ. qnew .AND. rold .EQ. rnew ) CYCLE ! skip duplicates
         
         c = c + 1                         ! copy data to lower position
         partit(c) = partit(a)
         q(c) = qnew
         r(c) = rnew
         chnext = hsteps(rnew - tm12) + 12 * qnew
         
         p = p + change(chprev, chnext, mode)  ! Markov chain of log probability
         chprev = chnext
         qold = qnew
         rold = rnew
       END DO
       
*         penalty for not ending on the tonic
       IF (chnext .NE. tonic) p = p + LOG(1.0 / 1000)
       PRINT *, 'evaluated ', evals, ' segments with ', ties, ' ties'
       
*-----------------------------------------------------------------------
*            End of main routine
*-------------------------------------------------------------------- ---
       RETURN       
              
*-----------------------------------------------------------------------
  50   CONTINUE !            score a segment
*                         the original scoring rules are commented out
*-----------------------------------------------------------------------
       sbest = -BIG                           ! real-valued scores
       evals = evals + 1
       fw = f * wsum ; gw = g * wsum

       DO m = 0, num_q - 1                     ! for each quality
         DO L = 0, 11                           ! for each root
            score = 0
           ctPP = 0; ctPA = 0 ; ctAP = 0 ; ctAA = 0
           DO j = 0, 11                            ! for each pitch class
             IF ( BTEST(template(L,m), j) ) THEN
               IF (weight(j) > 0) THEN                 
*                 score = score +  weight(j)   ! add weight for present pitches
                 ctPP = ctPP + weight(j)
               ELSE  
*                 score = score - 1        ! subtract one for absent pitches
                  ctPA = ctPA + fw
               END IF

             ELSE             
               IF (weight(j) > 0) THEN
*                 score = score -  weight(j)  ! subtract for wrong pitches
                 ctAP = ctAP + weight(j)
               ELSE
*                 CONTINUE
                 ctAA = ctAA + gw
               END IF

             END IF
           END DO
           
*             empirical contingency table formula 
*            No theory behind the square root, but it gives OK answers
           IF (wsum < 33) THEN
             div = sqroot(wsum)
           ELSE
             div = SQRT(FLOAT(wsum))
           END IF
!           PRINT *, 'Using div=', div

           score = (ctPP*ctAA - ctPA*ctAP) / div
                      
*                 tiebreakers
           IF (ABS(score-sbest) .LE. EPS) THEN 
             ties = ties + 1

*                prefer clearer harmonies
             IF  (m < qbest) THEN                 
               qbest = m
               rbest = L
               sbest = score
               CYCLE
             ELSE IF (m > qbest) THEN
               CYCLE
             END IF

*                    prefer more root pitch
             IF (weight(L) > weight(rbest)) THEN    
               qbest = m
               rbest = L
               sbest = score
               CYCLE
             ELSE IF (weight(L) < weight(rbest)) THEN
               CYCLE
             END IF

*                resolve aug, dim7, 7b5 based on prior chord
             IF ( (m .EQ. 6) .OR. (m .EQ. 7) .OR. (m .EQ. 9)
     &                      .AND. (prev > 1) ) THEN
               IF ( MOD(L - r(prev) + 12, 12) 
     &                  < MOD(rbest - r(prev) + 12, 12) ) THEN
                 qbest = m
                 rbest = L
                 sbest = score
                 CYCLE
               ELSE IF ( MOD(L - r(prev) + 12, 12) 
     &                  >  MOD(rbest - r(prev) + 12, 12) ) THEN
                 CYCLE
               END IF
             END IF  ! dim 7 resolution

*                     nearest root on the circle of fifths
             IF (prev > 1) THEN
               IF ( ABS(MOD(77 * (L - r(prev)) +6, 12) -6)
     &            < ABS(MOD(77 * (rbest - r(prev)) +6, 12) -6) ) THEN
                 qbest = m 
                 rbest = L
                 sbest = score
                 CYCLE
               ELSE IF ( ABS(MOD(77 * (L - r(prev)) +6, 12) -6) 
     &            > ABS(MOD(77 * (rbest - r(prev)) +6, 12) -6) ) THEN
                 CYCLE
               END IF
             END IF
                          
             PRINT *, 'chordal: warning - unbreakable tie'
           ELSE IF (score > sbest)  THEN        ! new preferred chord
             qbest = m
             rbest = L
             sbest = score
           END IF  ! tiebreaking
         END DO  ! next L
       END DO  ! next m
       GO TO label
       
*-----------------------------------------------------------------------       
  70   CONTINUE  ! count chord transitions
*-----------------------------------------------------------------------
       h = 0
       skip = 1
       DO i = 1, pall - 1
         IF (h < k) THEN                       ! change key
           IF (partit(i) .GE. key_time(h+1)) THEN
             h = h + 1
             tonic = MOD(7 * sharps(h), 12)
             IF (tonic < 0) THEN
               tm12 = tonic
               tonic = tonic + 12
             ELSE
               tm12 = tonic - 12
             END IF
             mode = majmin(h)
             IF ( plist(i) .AND. (partit(i) .EQ. key_time(h)) ) THEN
               skip = 1          ! skip one transition on a neat key change
             ELSE
               skip = 2        ! skip two transitions on a mid-chord key change
             END IF
           END IF
         END IF
         
         IF (.NOT. plist(i)) CYCLE
         
         chnext = hsteps(r(i) - tm12) + 12 * q(i)

*          put data into matrix
         IF (skip .LE. 0) change(chprev,chnext,mode) 
     &                        = change(chprev,chnext,mode) + 1
         chprev = chnext
         skip = skip - 1
       END DO
        
       END      ! of chordal
       
       
*-----------------------------------------------------------------------
*        Chord Transitition Log Probabilities
*      after calling chordal() for all input files, take the 
*      probability of a chord, given a following chord.  That is,
*
*         P(X[i] | X[i+1]) = P(X[i] & X[i+1]) / P(X[i+1])
*             ctlp = ln(P)
* 
*      change[chprev, chnext, mode]   REAL    In    Count of transitions
*                                             Out   Log probabilities
*-----------------------------------------------------------------------
       SUBROUTINE ctlp(change)
       
       IMPLICIT NONE
       INTEGER num_q ; PARAMETER (num_q = 25)
       REAL change ; DIMENSION change(0:12*num_q-1, 0:12*num_q-1, 0:1)
       INTEGER i, j, L
       REAL r, sum, default
       default = LOG( 0.001)              ! probability for unobserved changes
       
       DO L = 0, 1                 ! major and minor modes
         DO j = 0, 12 * num_q - 1
           sum = 0.0
           DO i = 0, 12 * num_q - 1              ! add
             sum = sum + change(i, j, L)
           END DO
           IF (sum .EQ. 0.0) sum = 1.0           ! allow a minimum of 1
           DO i = 0, 12 * num_q - 1
             r = change(i, j, L)
             IF (r .EQ. 0.0) THEN
               change(i, j, L) = default
             ELSE
               change(i, j, L) = LOG(r / sum)
             END IF
           END DO  
         END DO  ! next j
       END DO   ! next L
       END    ! of ctlp
       
       
*-----------------------------------------------------------------------
*  Simple key-finding based on the Krumhansl-Schmuckler algorithm.
*  Cross-entropy is used instead of the correlation coefficient.       
*-----------------------------------------------------------------------
*     Refer to:
*         David Temperley, 
*         Music and Probability, ch. 6:  "A Polyphonic Key-Finding Model"
*         MIT Press, 2007        
*----------------------------------------------------------------------- 
*    Name               Type      In/Out   Description
*   pitch[n]           INTEGER     In       MIDI note numbers
*   dur[n]             REAL        In       durations in seconds
*   n                  INTEGER     In       # of note events
*   sharps             INTEGER     Out      most likely key signature (# sharps)
*   mode               INTEGER     Out      0=major, 1=minor
*   fault              INTEGER     Out      error code, 0=OK
*-----------------------------------------------------------------------
       SUBROUTINE skey(pitch, dur, n, sharps, mode, fault)
       
       IMPLICIT NONE
       INTEGER n, pitch, sharps, mode, fault
       REAL dur
       DIMENSION pitch(n), dur(n)
       
*                          local variables
       INTEGER i, j, k,                      ! counters
     &         npc,                          ! neutral pitch class
     &         root                          ! tonic key
     
       REAL h, hbest,                       ! cross-entropy
     &      machine,                        ! external function
     &      obs(0:11),                      ! observed pitch-class frequencies
     &      profile(0:11,0:11,0:1),         ! Key profiles as log probability
     &      sum
     
       LOGICAL shifted
       SAVE profile, shifted
       
*-----------------------------------------------------------------------
*        The Kostka-Payne key profiles, as log probabilities
*          profile[npc, root, mode]
*-----------------------------------------------------------------------
       DATA profile( 0,0,0)  / -1.736 /  
       DATA profile( 1,0,0)  / -4.259 /
       DATA profile( 2,0,0)  / -2.163 /
       DATA profile( 3,0,0)  / -3.947 /
       DATA profile( 4,0,0)  / -1.846 /
       DATA profile( 5,0,0)  / -2.223 /
       DATA profile( 6,0,0)  / -3.789 /
       DATA profile( 7,0,0)  / -1.781 /
       DATA profile( 8,0,0)  / -3.709 /
       DATA profile( 9,0,0)  / -2.451 /
       DATA profile(10,0,0)  / -4.311 /
       DATA profile(11,0,0)  / -2.362 /

       DATA profile( 0,0,1)  / -1.771 /
       DATA profile( 1,0,1)  / -3.908 /
       DATA profile( 2,0,1)  / -2.178 /
       DATA profile( 3,0,1)  / -1.912 /
       DATA profile( 4,0,1)  / -4.447 /
       DATA profile( 5,0,1)  / -2.208 /
       DATA profile( 6,0,1)  / -3.685 /
       DATA profile( 7,0,1)  / -1.723 /
       DATA profile( 8,0,1)  / -2.337 /
       DATA profile( 9,0,1)  / -4.134 /
       DATA profile(10,0,1)  / -3.448 /
       DATA profile(11,0,1)  / -2.540 /
       DATA shifted  /.FALSE./
      
*-----------------------------------------------------------------------
*         transpose profiles
*-----------------------------------------------------------------------
       fault = 0
       IF (.NOT. shifted) THEN
         DO k = 0, 1             ! major, minor
           DO j = 1, 11          ! roots 1...11
             DO i = 0, 11        ! each pitch class
               profile(i,j,k) = profile(MOD(i+11,12), j-1, k)
             END DO
           END DO
         END DO
       END IF
  
*-----------------------------------------------------------------------
*       get an observation vector
*-----------------------------------------------------------------------
       DO i = 0, 11 ; obs(i) = 0.0 ; END DO     ! initialize observation
      
       DO i = 1, n                      ! add durations of each pitch class
         npc = MOD(pitch(i), 12)
         obs(npc) = obs(npc) + dur(i)
       END DO      
       sum = 0.0
       DO j = 0, 11 ; sum = sum + obs(j) ; END DO     ! add
       IF (sum .EQ. 0.0) THEN
         fault = 1
         RETURN
       END IF
       DO j = 0, 11 ; obs(j) = obs(j) / sum ; END DO  ! divide
      
*-----------------------------------------------------------------------
*     measure cross-entropy      H = -SUM (Po[i] @ ln(Pm[i])
*          where     Po   observed frequency
*                    Pm   model probability
*-----------------------------------------------------------------------
       hbest = machine(3)        ! very large number
       DO k = 0, 1                   ! major, minor
         DO j = 0, 11                  ! each root
           h = 0             
           DO i = 0, 11                   ! each pitch class
             h = h - obs(i) * profile(i,j,k)
           END DO
           IF (h < hbest) THEN
             hbest = h
             root = j
             mode = k
           END IF
         END DO
       END DO
      
*-----------------------------------------------------------------------
*      convert root to number of sharps in key signature
*      magic formula:     s = (-77 @ r - 6) % 12 + 6
*-----------------------------------------------------------------------
       sharps = MOD(-77 * root -6, 12) + 6
       END   ! of skey
       
*+++++++++++++++++++++++ End of file harmony.f +++++++++++++++++++++++++      
              
