*-----------------------------------------------------------------------
*        kasc - K-means Approximate Spectral Clustering.  
*
*               Analyze data by grouping like items together.  
*                       by Andy Allinger, 2011
*-----------------------------------------------------------------------
* based on the algorithm discovered in:
*     Donghui Yan, Ling Huang, and Michael I. Jordan, 
*     "Fast Approximate Spectral Clustering" 
*     University of California Berkeley technical report, 2009.
*-----------------------------------------------------------------------
* and the spectral method discovered in:
*     Francis R. Bach and Michael I. Jordan, 
*     "Learning Spectral Clustering," 
*     University of California Berkeley technical report, 2003.
************************************************************************
* Everyone may freely use, modify, and copy this program for any purpose
* according to the terms of the Gnu General Public License, version 2, 
* or at your option, any later version.
*
* Subroutines which do not derive from Gnu software may, at your option,
* be used according to the terms of the Allegro Giftware License.
*
* If you did not receive a copy of these licenses, refer to:
*      www.fsf.org
*      alleg.sourceforge.net
************************************************************************
*                             TRANSFER LIST
*
* ____Variable_________Type______I/O___________Description______________
*    X [m,n]          REAL      In            Data points in R^n
*    m                INTEGER   In            Number of points
*    n                INTEGER   In            Number of measurements per point
*    r                INTEGER   In            Number of distinct observers
*                                              if none of the variables are 
*                                              contra-categorical, r is ignored
*                                              if in doubt, use 1
*    C [klim,n]       REAL      Out           Cluster centroids
*    klim             INTEGER   In            Row dimension of C
*    kmin             INTEGER   In            Fewest clusters to make
*    k                INTEGER   In/Out        Default/actual clusters to make
*    kmax             INTEGER   In            Most clusters to make
*    Q [m]            INTEGER   Out           What cluster each point is in
*    P [klim]         INTEGER   Out           Number of points in each cluster
*    T [n]            REAL      In            When T < 0, period of modulo 
*                                              arithmetic on distances
*                                              T = 0, common math
*                                              T > 0, data is a categorical
*                                              variable represented by 
*                                              non-negative numbers, or data
*                                              is a counter-categorical variable
*                                              represented by negative 
*                                              approximations of integers.
*                                              The value of T in this case is
*                                              the weight of the variable.
*                                              if in doubt use /n * 0/
*    y                REAL      In            Minkowski's parameter
*                                              affects how distance is measured
*                                              if in doubt use 2.0
*    f                REAL      In            Nearness parameter
*                                              If in doubt use 1.0
*    opt              INTEGER   In            Options: 
*                                              bit 0:  median instead of mean
*                                              bit 1:  Cauchy instead of Gauss
*                                              bit 2:  repeat clustering steps
*                                              bit 3:  cluster slow and careful
*                                              bit 4:  user supplied centers
*                                              bit 5:  standardize the input
*                                              bit 6:  use a sparsified matrix
*                                              If in doubt, use 0 
*    u                REAL      Out           Residual error
*    fault            INTEGER   Out           Error code: 
*                                              -3  one cluster
*                                              -2  m = 2
*                                              -1  all singleton clusters
*                                               0  normal return
*                                               1  k out of bounds
*                                               2  no data
*                                               3  more clusters than points
*                                               4  r out of bounds
*                                               5  no weights assigned
*                                               6  eigensolver failed
*                                               7  file i/o error
*                                               8  file i/o error
*                                               Negative errors are not fatal
*
*                              GLOBALS
*
*    I/O  Units #7, #8                         scratch files
*-----------------------------------------------------------------------
      SUBROUTINE kasc(X,m,n,r,C,klim,kmin,k,kmax,Q,P,T,y,f,opt,u,fault)
      
      IMPLICIT NONE
      INTEGER m, n, r, klim, kmin, k, kmax, Q, P, opt, fault 
      REAL X, C, T, y, f, u
      DIMENSION X(m,n), C(klim,n), Q(m), P(kmax), T(n)
      
*-----------------------------------------------------------------------
*            local variables
*-----------------------------------------------------------------------
      INTEGER E                                ! maximum eigenproblem size
      PARAMETER (E = 256)                      ! edit to match your budget
      
      INTEGER A(r,kmax),    ! count of how often an observer ID is in a cluster
     &        A0(r,E),      ! observer frequencies in the data reduction step
     &        h, i, i0,     ! count trials, count points, random point
     &        ind(m),       ! permutation vector
     &        iopt,         ! choice of k-means initialization method
     &        iter,         ! maximum iterations of the k-means steps
     &        j,            ! count dimensions
     &        ktry,         ! trial value of k
     &        L,            ! count clusters
     &        m1,           ! either E or m, depending on if data is reduced
     &        o, pos,       ! position to select in array of length o
     &        P0(E),        ! cluster populations in data reduction step
     &        Q1(E),        ! cluster membership from spectral step
     &        Qbest(m),     ! saved cluster membership
     &        sample,       ! number of points in subsample
     &        swap          
     
      REAL alpha,           ! parameter in Cauchy and Gauss distributions
     &     ave(n),          ! average of input data in a column
     &     BIG,             ! very large number
     &     biggap,          ! largest eigengap
     &     C1(kmax,kmax),   ! cluster centers in eigenspace
     &     Cbest(kmax,n),   ! saved cluster centers
     &     D(E),            ! sum of rows of W
     &     DISR(E),         ! inverse of square root of D
     &     DSR(E),          ! square root of D
     &     dev(n),          ! standard deviation or mean deviation in a column
     &     dif,
     &     dist0,           ! external function
     &     dist2,           ! external function
     &     G(E,n),          ! representative points 
     &     Gbest(E,n)       ! saved representative points
      REAL gap,             ! eigengap
     &     glim(n), llim(n),! range of input data
     &     lambda(E),       ! eigenvalues
     &     MACHEP,          ! machine precision
     &     machine,         ! external function:  find machine constants
     &     median,          ! external function
     &     mode,            ! external function
     &     rank,            ! external function
     &     s,               ! typical distance
     &     sigma,           ! deviation across dimensions
     &     sum,
     &     ubest,           ! residual from clustering routines
     &     V(E,E),          ! eigenvectors
     &     W(E,E),          ! similarity matrix
     &     Xsub(m/10,n),    ! subsample of data points
     &     Z(E*(E-1)/2),    ! all distances in single array
     &     zero(n),         ! n @ 0
     &     zero1(E)         ! E @ 0
     
      LOGICAL brief,      ! skip clustering restarts
     &        cauchy,     ! Cauchy distribution instead of Gauss's bell curve
     &        HW,         ! use the Hartigan-Wong algorithm when appropriate
     &        init,       ! use the input centers as initial values
     &        neg,        ! data in a column is negative
     &        robust,     ! use the median instead of mean average
     &        sparse,     ! demand the sparse clustering method
     &        stand,      ! standardize the input
     &        thuro       ! keep points "live" longer during clustering

      EXTERNAL dist2      
      SAVE iter, BIG, MACHEP
      DATA iter, BIG, MACHEP / 1000, -1.0, -1.0/

*-----------------------------------------------------------------------
*       begin executable statements
*-----------------------------------------------------------------------
      IF (MACHEP < 0) MACHEP = machine(1)                ! estimate precision
      IF (BIG < 0) BIG = machine(3)
        
      fault = 0                                 ! clear error codes
      DO j = 1, n ; zero(j) = 0.0 ; END DO     ! zero arrays
      DO j = 1, E ; zero1(j) = 0.0 ; END DO
      
*       Use median and average deviation instead of mean and standard deviation
      robust = BTEST(opt, 0)
      
*       decide if routine KMNS by Hartigan & Wong may be used
      IF (robust) THEN
        HW = .FALSE.
      ELSE
        HW = .TRUE.             ! use regular KMNS
        DO j = 1, n ; IF (ABS(T(j)) > MACHEP) HW = .FALSE. ; END DO
      END IF

      cauchy = BTEST(opt, 1)           !  use the Cauchy distribution?
      brief = .NOT. BTEST(opt, 2)      !  skip clustering restarts? 
      thuro = BTEST(opt, 3)            !  be thorough in clustering?
      init = BTEST(opt, 4)             !  use input centers?
      stand = BTEST(opt, 5)            !  standardize the input?
      sparse= BTEST(opt, 6)            !  demand the sparse matrix method
 
*-----------------------------------------------------------------------
*          validate input
*-----------------------------------------------------------------------
      IF (k < 1 .OR. k < kmin .OR. k > kmax .OR. klim < kmax) THEN
        fault = 1
        RETURN
      END IF
      
      IF (m<1 .OR. n<1 .OR. r<0 .OR. y<0 .OR. f<0 .OR. u<0) THEN
        fault = 2
        RETURN
      END IF
      
      IF (k > m) THEN
        fault = 3
        RETURN
      END IF
      
*       Ensure that data is either all categorical or all contra-categorical
      DO j = 1, n
        IF (T(j) < MACHEP) CYCLE
        neg = .FALSE.
        IF (X(1,j) < 0) neg = .TRUE.
        DO i = 2, m
          IF (neg .NEQV. ( X(i,j) < 0)) THEN 
            fault = 4
            RETURN
          END IF
        END DO
      END DO
        
      IF (ABS(y) < MACHEP) THEN              ! improper usage
        DO j = 1, n                          ! y=0 is the same as all variables
          IF (ABS(T(j)) > MACHEP) GO TO 10   ! categorical, so at least one 
        END DO                               ! element of T[] must be nonzero
        fault = 5
        RETURN
      END IF
  10  CONTINUE
  
*-----------------------------------------------------------------------
*       a useless, but permissible case      
*-----------------------------------------------------------------------
      IF (m .EQ. k) THEN
        DO j = 1, n
          DO i = 1, m
            C(i,j) = X(i,j)
          END DO
        END DO
        DO i = 1, m
          Q(i) = i
        END DO
        DO L = 1, k
          P(L) = 1
        END DO
        u = 0
*               Don't like nonzero exits?  Don't pass weird arguments.
        fault = -1 
        RETURN
      END IF

*-----------------------------------------------------------------------
*       find average, deviation, and range
*-----------------------------------------------------------------------
      IF (robust) THEN
        DO j = 1, n
          IF (T(j) > MACHEP) THEN
            ave(j) = mode( X(1,j), m, fault)
            IF (fault .NE. 0) PRINT *, 'mode:  error#', fault
          ELSE 
            ave(j) = median( X(1,j), m, T(j), fault)
            IF (fault > 0) STOP 'error in function median()'
          END IF
!            PRINT *, 'found ave(', j, ')=', ave(j)
          llim(j) = X(1,j) ; glim(j) = X(1,j)
          sum = 0.0 
          DO i = 1, m
            llim(j) = MIN( llim(j), X(i,j) )
            glim(j) = MAX( glim(j), X(i,j) )
            dif = dist0(X(i,j), ave(j), T(j), MACHEP)
            sum = sum + ABS(dif)
          END DO
          dev(j) = sum / m             ! mean absolute deviation
          PRINT *, 'dimension ', j, ' average=',ave(j),' dev=',dev(j)
        END DO         ! next j
  
      ELSE           ! not robust
        DO j = 1, n
          IF (T(j) > MACHEP) THEN
            ave(j) = mode( X(1,j), m, fault)
            IF (fault .NE. 0) PRINT *, 'mode:  error#', fault
          ELSE IF (T(j) < 0) THEN
            ave(j) = median( X(1,j), m, T(j), fault)
            IF (fault > 0) STOP 'error in function median()'
          ELSE
            sum = 0.0
            DO i = 1, m ; sum = sum + X(i,j) ; END DO    ! add column
            ave(j) = sum / m
          END IF
          llim(j) = X(1,j) ; glim(j) = X(1,j)
          sum = 0.0 
          DO i = 1, m
            llim(j) = MIN( llim(j), X(i,j) )
            glim(j) = MAX( glim(j), X(i,j) )
            dif = dist0(X(i,j), ave(j), T(j), MACHEP)
            sum = sum + dif * dif
          END DO
          dev(j) = SQRT(sum / m)     ! standard deviation
          PRINT *, 'dimension ',j, ' average=',ave(j),' dev=',dev(j)
        END DO
      END IF

*          deviation over all dimensions
      sigma = dist2(dev, zero, n, T, y)

*-----------------------------------------------------------------------
*      special case m = 2
*-----------------------------------------------------------------------
      IF (m .EQ. 2) THEN
        DO j = 1, n
          C(1,j) = ave(j)
        END DO
        Q(1) = 1 ; Q(2) = 1 
        P(1) = 2
        u = sigma * 2
        fault = -2
        RETURN
      END IF

*-----------------------------------------------------------------------
*         special case k = 1
*-----------------------------------------------------------------------
      IF (k .EQ. 1) THEN
        DO j = 1, n                  ! cluster center is the average
          C(1,j) = ave(j)
        END DO
        DO i = 1, m                  ! all memberships the same
          Q(i) = 1
        END DO
        P(1) = m
        u = sigma * m
        fault = -3
        RETURN
      END IF

*-----------------------------------------------------------------------
*              Standardize input if desired
*       Don't alter modular or categorical ; don't divide by zero.
*-----------------------------------------------------------------------
      IF (stand) THEN
        OPEN (UNIT=7,FORM='UNFORMATTED',STATUS='SCRATCH',IOSTAT=fault)
        REWIND (UNIT=7)         ! Clive Page says to do this
        WRITE (UNIT=7, IOSTAT=fault) X
        IF (fault .NE. 0) THEN 
          PRINT *, 'Trouble opening scratch file'
          fault = 7
        END IF
        DO j = 1, n
          IF (ABS(T(n)) > MACHEP .OR. ABS(dev(j)) < MACHEP ) CYCLE 
          DO i = 1, m
            X(i,j) = (X(i,j) - ave(j)) / dev(j)
          END DO
          llim(j) = (llim(j) - ave(j)) / dev(j)
          glim(j) = (glim(j) - ave(j)) / dev(j)
        END DO
      END IF

*-----------------------------------------------------------------------
*       choose what algorithms are proper for the combination of k and m
*-----------------------------------------------------------------------
*       sparse spectral, based on assumption that Lanczos is O(K @ N^2)
*         and that QL is O(N^3)
      IF ((m > E .AND. k * FLOAT(m) **2 < FLOAT(E) **3) .OR. sparse)
     &      GO TO 20
     
      IF (k .GE. E) GO TO 30   ! conventional k-means or k-medians
      
*       decide if a data reduction step is needed      
      IF (m > E) THEN
        m1 = E
        GO TO 50               ! clustering as a data reduction step
      ELSE
        m1 = m
        GO TO 70               ! advance to spectral step
      END IF

*----------------------------------------------------------------------
  20  CONTINUE  ! use a sparse matrix (instead of Yan data reduction)
*----------------------------------------------------------------------
       CALL sclust(X, m, n, C1, kmin, k, kmax, Q, Qbest, P, T, y, f, 
     &   cauchy, brief, iter, dist2, u, fault)
       GO TO 90
       
*----------------------------------------------------------------------
  30   CONTINUE  !   k is too large for spectral clustering:  
*                prepare k-means or k-medians routines as the solution 
*----------------------------------------------------------------------
      ubest = BIG      ! cluster several times and use the best result
      DO h = 1, 10                  ! 10 is editable
        IF (init) GO TO 40
        iopt = 5 - MOD(h, 5) 
        IF (iopt .EQ. 5 .AND. (m < 20 * k) ) CYCLE  ! subsample large sets only

*        clear frequencies matrix
        IF (iopt .EQ. 1 .OR. iopt .EQ. 2) THEN
          DO L = 1, k
            DO j = 1, r
              A(j,L) = 0
            END DO
          END DO
        END IF
            
*        random cluster centers
        IF (iopt .EQ. 1) THEN
          DO j = 1, n          
            DO L = 1, k
              IF (T(j) < MACHEP) THEN
                C(L, j) = RAND(0)*(glim(j)-llim(j)) + llim(j)
              ELSE
                C(L,j) = MIN(AINT(RAND(0)*(glim(j)-llim(j)+1) +llim(j)),
     &                         AINT(glim(j)) )
              END IF 
            END DO
          END DO

*       random points as centers
        ELSE IF (iopt .EQ. 2) THEN
          DO L = 1, k
            i0 = MIN(INT(RAND(0) * m) + 1, m)
            DO j = 1, n
              C(L,j) = X(i0,j) 
            END DO
          END DO
 
*       random cluster memberships
        ELSE IF (iopt .EQ. 3 .OR. iopt .EQ. 5) THEN
          DO i = 1, m
            Q(i) = MIN(INT(RAND(0) * k) + 1, k)
          END DO

*         Hoare partitioning
        ELSE IF (iopt .EQ. 4) THEN
          CALL hclust(X, m, m, n, k, Q, P, T, y, fault)
          IF (fault > 0) STOP 'hclust: Trouble partitioning'
        END IF

*          Let centers be the average of points contained within them
        IF (iopt .EQ. 3 .OR. iopt .EQ. 4 .OR. iopt .EQ. 5)
     &       CALL center(X,m,m,n,C,klim,k,A,r,Q,P,T,robust)

*         multi-stage with subsampling
        IF (iopt .EQ. 5) THEN
          sample = m / 10
          DO i = 1, m ; ind(i) = i ; END DO ! make a random permutation vector
          DO i = m, m-sample+1, -1   ! partial Durstenfeld shuffle
            i0 = MIN(INT(RAND(0)*i) + 1, i)
            swap = ind(i0) ; ind(i0) = ind(i) ; ind(i) = swap
          END DO
          
          DO j = 1, n                     ! transfer data to temp. array
            DO i = 1, sample     
              Xsub(i,j) = X(ind(m-i+1),j)
            END DO
          END DO
        
          IF (robust) THEN 
            CALL kmedians(Xsub, sample, n, C, klim, k, A, r,
     &            Q, P, T, thuro, iter, u, fault)
            IF (fault > 0) PRINT *,'kmedians:  error #', fault
          ELSE IF (HW) THEN
            CALL KMNS(Xsub,sample,n,C,klim,k,Q,P,iter,u,fault)
            IF (fault > 0) PRINT *, 'KMNS:  error #', fault
          ELSE
            CALL kmeans(Xsub, sample, n, C, klim, k, A, r,
     &             Q, P, T, thuro, iter, u, fault)
            IF (fault > 0) PRINT *, 'kmeans:  error #', fault
          END IF  ! robust?
        END IF  ! subsample?

*-----------------------------------------------------------------------
  40  CONTINUE               ! Cluster
*-----------------------------------------------------------------------
        IF (robust) THEN
          CALL kmedians(X, m, n, C, klim, k, A, r,
     &            Q, P, T, thuro, iter, u, fault)
          IF (fault > 0) PRINT *, 'kmedians:  error #', fault
        ELSE IF (HW) THEN
          CALL KMNS(X, m, n, C, klim, k, Q, P, iter, u,fault)
          IF (fault > 0) PRINT *, 'KMNS:  error #', fault
        ELSE
          CALL kmeans(X, m, n, C, klim, k, A, r, 
     &              Q, P, T, thuro, iter, u, fault)
          IF (fault > 0) PRINT *, 'kmeans:  error #', fault
        END IF
        
        fault = 0         ! ignore clustering routine errors
        IF (brief) GO TO 90       ! reduce computation
        
        IF (u < ubest) THEN
          IF (init) THEN
            PRINT *, "Found a better clustering, custom initial values"
          ELSE
            PRINT *, "Found a better clustering with option", iopt
          END IF
          ubest = u
          DO i = 1, m                   ! save best answer so far
            Qbest(i) = Q(i)       
          END DO
          DO j = 1, n
            DO L = 1, k
              Cbest(L,j) = C(L,j)
            END DO
          END DO
        ELSE 
          PRINT *, 'u=', u, ' no improvement'
        END IF      
      END DO  ! next h
 
      DO i = 1, m               ! restore best result
        Q(i) = Qbest(i)
      END DO
      DO j = 1, n
        DO L = 1, k
          C(L,j) = Cbest(L,j)
        END DO
      u = ubest
      END DO
     
*       skip to end phase
      GO TO 90

*----------------------------------------------------------------------
  50  CONTINUE    ! data reduction step:
*       cluster to reduce X[m,n] to representative points G[E,n]
*----------------------------------------------------------------------
      ubest = BIG      ! cluster several times and use the best result
      DO h = 1, 10
        IF (init) GO TO 60
        iopt = 5 - MOD(h, 5)
        IF (iopt .EQ. 5 .AND. (m < 20 * E) ) THEN
          PRINT *, 'Skipping subsampling step'
          CYCLE  ! subsample large sets only
        END IF
        
*          clear frequency matrix
        IF (iopt .EQ. 1 .OR. iopt .EQ. 2) THEN
          DO L = 1, E
            DO j = 1, r
              A0(j,L) = 0
            END DO
          END DO
        END IF

*       random cluster centers         
        IF (iopt .EQ. 1) THEN
          DO j = 1, n
            DO L = 1, E
              IF (T(j) < MACHEP) THEN
                G(L, j) = RAND(0)*(glim(j)-llim(j)) + llim(j)
              ELSE 
                G(L,j) = MIN(AINT(RAND(0)*(glim(j)-llim(j)+1)+llim(j)),
     &                        AINT(glim(j)) ) ! integer categories
              END IF
            END DO
          END DO

*       random points as centers
        ELSE IF (iopt .EQ. 2) THEN
          DO L = 1, E
            i0 = MIN(INT(RAND(0) * m) + 1, m)
            DO j = 1, n
              G(L,j) = X(i0,j) 
            END DO
          END DO
 
*       random cluster memberships
        ELSE IF (iopt .EQ. 3 .OR. iopt .EQ. 5) THEN
          DO i = 1, m
            Q(i) = MIN(INT(RAND(0) * E) + 1, E)
          END DO

*         Hoare partitioning
        ELSE IF (iopt .EQ. 4) THEN
          CALL hclust(X, m, m, n, E, Q, P0, T, y, fault)
          IF (fault > 0) STOP 'hclust: Trouble partitioning'
        END IF
        
*         Let centers be the average of points contained within them
        IF (iopt .EQ. 3 .OR. iopt .EQ. 4 .OR. iopt .EQ. 5)
     &       CALL center(X,m,m,n,G,E,E,A0,r,Q,P0,T,robust)

*         multi-stage with subsampling
        IF (iopt .EQ. 5) THEN
          sample = m / 10
!          PRINT *, 'Using a subsample of ', sample
          
          DO i = 1, m ; ind(i) = i ; END DO ! make a random permutation vector
          DO i = m, m-sample+1, -1   ! partial Durstenfeld shuffle
            i0 = MIN(INT(RAND(0)*i) + 1, i)
            swap = ind(i0) ; ind(i0) = ind(i) ; ind(i) = swap
          END DO
          
          DO j = 1, n                     ! transfer data to temp. array
            DO i = 1, sample     
              Xsub(i,j) = X(ind(m-i+1),j)
            END DO
          END DO
        
          IF (robust) THEN
            CALL kmedians(Xsub, sample, n, G, E, E, A0, r, 
     &               Q, P0, T, thuro, iter, u, fault)
            IF (fault > 0) PRINT *, 'kmedians:  Error #', fault
          ELSE IF (HW) THEN
            CALL KMNS(Xsub,sample,n,G,E,E,Q,P0,iter,u,fault)
            IF (fault > 0) PRINT *, 'KMNS:  Error #', fault
          ELSE
            CALL kmeans(Xsub, sample, n, G, E, E, A0, r, 
     &               Q, P0, T, thuro, iter, U, fault)
            IF (fault > 0) PRINT *, 'kmeans:  Error #', fault
          END IF
        END IF   ! option 5 ?

*-----------------------------------------------------------------------
  60  CONTINUE                   !  cluster
*-----------------------------------------------------------------------
        IF (robust) THEN
          CALL kmedians(X, m, n, G, E, E, A0, r, 
     &            Q, P0, T, thuro, iter, u, fault)
          IF (fault > 0) PRINT *, 'kmedians:  error #', fault
        ELSE IF (HW) THEN
          CALL KMNS(X, m, n, G, E, E, Q, P0, iter, u, fault)
          IF (fault > 0) PRINT *, 'KMNS:  error #', fault
        ELSE
          CALL kmeans(X, m, n, G, E, E, A0, r, 
     &             Q, P0, T, thuro, iter, u, fault)
          IF (fault > 0) PRINT *, 'kmeans:  error #', fault
        END IF

        fault = 0             ! ignore clustering errors
        IF (brief) GO TO 70   ! cluster only once

        IF (u < ubest) THEN
          IF (init) THEN
            PRINT *, 'Found a better clustering, custom option, u=', u
          ELSE
            PRINT *, "Found better clustering: option", iopt, ' u=',u
          END IF
          ubest = u
          DO i = 1, m                   ! save best answer so far
            Qbest(i) = Q(i)       
            IF (Q(i) > E .OR. Q(i) < 1) stop 'Q value out of range!'
          END DO
          DO j = 1, n
            DO L = 1, E
              Gbest(L,j) = G(L,j)
            END DO
          END DO
        ELSE 
          PRINT *, 'u=', u, ' no improvement vs. ubest=', ubest
        END IF      
      END DO  ! next h
      
*          keep best values
      DO i = 1, m
        Q(i) = Qbest(i)
      END DO
      DO j = 1, n
        DO L = 1, E
          G(L,j) = Gbest(L,j)
        END DO
      END DO    

*-----------------------------------------------------------------------
  70  CONTINUE                            !   spectral clustering 
*-----------------------------------------------------------------------
      IF (m > E) THEN       
        CALL distance(G, E, n, W, E, T, y)    ! fill lower triangle of matrix
      ELSE                                 
        CALL distance(X, m, n, W, E, T, y)
      END IF

*          find a typical distance -- based on the assumption that distances
*       are of two classes:  intra-cluster and inter-cluster.  Further assume 
*       that clusters are of equal size (m / k).  Then the rank position 
*       of the median intra-cluster distance should be m @ (m-k) / (4 @ k)
      o = 0
      DO j = 1, m1-1
        DO i = j+1, m1
          o = o + 1
          Z(o) = W(i,j)
        END DO
      END DO
      pos = NINT( m1*(m1-k) / FLOAT(4*k) )
      pos = MAX(pos, 1) ; pos = MIN(pos, o)
      s = rank(Z, o, pos, fault)        
      PRINT *, 'found s=', s,  ' SIGMA=', sigma
      IF (fault .NE. 0) STOP 'error in function rank()'

*-----------------------------------------------------------------------
*      the parameter alpha ~ gamma of the Cauchy distribution
*                          ~ sigma of the Gauss distribution
*-----------------------------------------------------------------------
      IF (.NOT. stand) s = MIN(s, sigma) ! deviation already taken into account
      s = MAX(s, (MACHEP*m1))  ! don't divide by zero

*          f is an arbitrary scalar that depends on what kind of answer you want
*      small f causes only very near points to have significant values in the
*      affinity matrix.  Results from large f resemble more a conventional
*      k-means clustering.  So if you're trying to separate intertwined spirals,
*      by all means, use f < 1.  More common statistical processing may do well
*      to leave f = 1.  At least, that's my guess.
      alpha = f * (s **2) / 2

*         convert distances to similarities        
      IF (cauchy) THEN     ! Cauchy distribution on distance
        DO j = 1, m1-1
          DO i = j+1, m1
            W(i,j) = 1 / (W(i,j) **2 + alpha)
          END DO
        END DO
         
      ELSE                  ! Gauss distribution
        DO j = 1, m1-1
          DO i = j+1, m1
            W(i,j) = EXP( -(W(i,j) **2) / alpha)
          END DO
        END DO
      END IF

*-----------------------------------------------------------------------
*               normalize the matrix
*-----------------------------------------------------------------------
*         vectors required for the Bach-Jordan method
      DO i = 1, m1
        D(i) = 0.0
        DSR(i) = 0.0
        DISR(i) = 0.0
      END DO
        
*          sum along the row up to the diagonal, then down the column        
      DO i = 1, m1
        DO j = 1, i-1
          D(i) = D(i) + W(i,j)
        END DO
        DO j = i+1, m1
          D(i) = D(i) + W(j,i)
        END DO
      END DO
        
      DO i = 1, m1
        DSR(i) = SQRT(D(i))
        DISR(i) = 1 / DSR(i)
      END DO
        
*              L = W' = I - D^(-1/2) @ W @ D^(-1/2)
      DO j = 1, m1-1
        DO i = j+1, m1
          W(i,j) = -1 * DISR(i) * W(i,j) * DISR(j)
        END DO
      END DO
         
      DO i = 1, m1
        W(i,i) = 1.0
      END DO

*-----------------------------------------------------------------------
*         get the eigenvalues and eigenvectors by the QL factorization 
*-----------------------------------------------------------------------
!        Print *, 'Call SEIGV with m1=', m1
      CALL SEIGV(W, E, m1, lambda, V, fault)
      IF (fault .NE. 0) THEN 
        PRINT *, 'Eigensolver returned #', fault, ' unrecoverable error'
        fault = 6
        GO TO 30
      END IF
        
*          look for widest eigengap
      biggap = 0
      ktry = -1

      DO j = 2, m1-1
        gap = ABS(lambda(j+1) - lambda(j) )
        IF (gap > biggap) THEN
          biggap = gap
          ktry = j
        END IF
      END DO
 
      IF (biggap .EQ. 0) PRINT *, 'No eigengap found!'
      PRINT *, 'found ', ktry, ' as a natural number of clusters '
      IF ( (ktry < kmin) .OR. (ktry > kmax) ) ktry = k
      k = ktry       ! return value of k

*----------------------------------------------------------------------- 
*         initialize for weighted k-means clustering
*-----------------------------------------------------------------------
      ubest = BIG
      DO h = 1, 6                    ! 6 is editable 
*        on even steps, random membership initialization
        IF (MOD(h,2) .EQ. 0) THEN
          DO i = 1, m1
            Q1(i) = MIN(INT(RAND(0) * k) + 1, k)
          END DO
        ELSE             !    odd steps, initialize by partitioning 
          CALL hclust(V, E, m1, k, k, Q1, P, zero1, 2.0, fault)
          IF (fault > 0) STOP 'hclust: failed at spectral step'
        END IF
        CALL center(V,E,m1,k,C1,kmax,k,A,r,Q1,P,zero1,.FALSE.)

*-----------------------------------------------------------------------
*           cluster eigenvectors by weighted k-means
*-----------------------------------------------------------------------
        CALL wkm(V, E, m1, kmax,k, C1, D, DSR, DISR, Q1, iter, u, fault)
        IF (fault > 0) PRINT *, 'wkm: error#', fault
        fault = 0                       ! error #2 is not a big deal, ignore
        IF (brief) GO TO 80
                 
        IF (u < ubest) THEN
          IF (h > 1) THEN
            PRINT *, 'better result alpha=', alpha, ' s=', s,
     &        ' init: ', MOD(h,2), ' u=', u, ' h=', h
          END IF
          ubest = u
          DO i = 1, m1
            Qbest(i) = Q1(i)
          END DO
        END IF
      END DO      ! next h
      
      DO i = 1, m1                   ! dimensions of Q1 and Qbest both .GE. m1
        Q1(i) = Qbest(i)
      END DO
      u = ubest
      
*-----------------------------------------------------------------------
  80  CONTINUE
*        lookup memberships from representative points
*-----------------------------------------------------------------------
      IF (m > E) THEN
        DO i = 1, m
          Q(i) = Q1(Q(i))
        END DO
      ELSE
        DO i = 1, m
          Q(i) = Q1(i)
        END DO
      END IF  
      
*----------------------------------------------------------------------- 
  90  CONTINUE
*         count cluster populations, return centers
*-----------------------------------------------------------------------      
*     Restore original data
      IF (stand) THEN
        REWIND (UNIT=7)
        READ (UNIT=7, IOSTAT=fault) X
        IF (fault .NE. 0) THEN
          PRINT *, 'Trouble reading scratch file'
          fault = 7
        END IF
        CLOSE (UNIT=7)
      END IF

      PRINT *, 'Before final centering: ', P
      CALL center(X, m, m, n, C, klim, k, A, r, Q, P, T, robust)
      PRINT *, 'After final centering:', P

      RETURN
      END          ! of kasc


*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
*                       -----Giftware-----
*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
*    Let each center be the average of its points
*        X [mlim,n]         data points                 Real
*        mlim               row dimension of X          Integer
*        m                  rows                        Integer
*        n                  columns                     Integer
*        C [klim,n]         centers                     Real
*        klim               row dimension of C[]        Integer
*        k                  number of clusters          Integer
*        A [r,klim]         frequencies of observers    Integer
*        r                  number of observers         Integer
*        Q [m]              cluster membership          Integer
*        P [k]              cluster population          Integer
*        T [n]              periods/weights             Real
*        robust             median or mean              Logical
*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
      SUBROUTINE center(X,mlim,m,n,C,klim,k,A,r,Q,P,T,robust)
      
      IMPLICIT NONE
      INTEGER m, mlim, n, klim, k, A, r, Q, P
      REAL X, C, T
      LOGICAL robust 
      DIMENSION X(mlim,n), C(klim,n), A(r,k), Q(m), P(k), T(n)

      INTEGER fault, h, h1, i, j, L, L1, o
      REAL Z, median, mode, machine, MACHEP
      DIMENSION Z(m)
      
      fault = 0
      MACHEP = machine(1)

*       count cluster populations      
      DO L = 1, k ; P(L) = 0 ; END DO 
      DO i = 1, m
        L1 = Q(i)
        P(L1) = P(L1) + 1
      END DO
      
      PRINT *, 'k=', k, 'm=', m, ' P=', P

*       averages 
      IF (robust) THEN
        DO L = 1, k
          IF (P(L) .EQ. 0) CYCLE       ! skip empty clusters
          DO j = 1, n
            IF (T(j) > MACHEP .AND. X(1,j) < 0) CYCLE   ! skip contra-cat.   
            o = 0
            DO i = 1, m
              IF (Q(i) .NE. L) CYCLE        ! count only points in the cluster
              o = o + 1
              Z(o) = X(i,j)             
            END DO
            IF (T(j) > MACHEP) THEN
              C(L,j) = mode(Z, o, fault)
            ELSE
              C(L,j) = median(Z, o, T(j), fault)
            END IF
            IF (fault > 0) STOP 'center: failed to find average'
          END DO     ! next j
        END DO    ! next L
      
      ELSE     ! not robust
        DO j = 1, n                          ! zero
          IF (ABS(T(j)) > MACHEP) CYCLE
          DO L = 1, k
            IF (P(L) .EQ. 0) CYCLE
            C(L,j) = 0.0 
          END DO  
        END DO     
      
        DO i = 1, m                           ! add
          L1 = Q(i)
          DO j = 1, n 
            IF (ABS(T(j)) > MACHEP) CYCLE
            C(L1,j) = C(L1,j) + X(i,j)
          END DO
        END DO
       
        DO j = 1, n                    ! divide
          IF (ABS(T(j)) > MACHEP) CYCLE
          DO L = 1, k
            IF (P(L) .EQ. 0) CYCLE
            C(L,j) = C(L,j) / P(L)
          END DO
        END DO
        
*           use the mode for categorical variables and median for modulo
        DO L = 1, k
          IF (P(L) .EQ. 0) CYCLE
          DO j = 1, n
            IF (ABS(T(j))<MACHEP .OR. (T(j)>MACHEP.AND.X(1,j)< 0)) CYCLE
            o = 0
            DO i = 1, m
              IF (Q(i) .NE. L) CYCLE
              o = o + 1
              Z(o) = X(i, j)             
            END DO
            IF (T(j) > MACHEP) THEN
              C(L,j) = mode(Z, o, fault)
              IF (fault .NE. 0) PRINT *, 'mode:  Error#', fault
            ELSE
              C(L,j) = median(Z, o, T(j), fault)
              IF (fault > 0) PRINT *, 'median:  Error#', fault
            END IF
          END DO     ! next j
        END DO    ! next k
      END IF    ! robust?

*           contra-categorical frequencies
      DO j = 1, n
        IF (T(j) < MACHEP .OR. X(1,j) .GE. 0) CYCLE  ! skip others        
        DO L = 1, k                              ! clear
          DO h = 1, r ; A(h,L) = 0 ; END DO
        END DO
        DO i = 1, m                              ! count
          L1 = Q(i)
          h1 = -NINT(X(i,j))
          A(h1,L1) = A(h1,L1) + 1
        END DO
      END DO
          
      END   ! of center


*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*                  -----------Giftware------------ 
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*     dist0 - subtraction made complicated
*
*            A           value                         Real
*            B           value                         Real
*            T           period                        Real
*            MACHEP      computer precision            Real
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      FUNCTION dist0(A, B, T, MACHEP)
      
      IMPLICIT NONE
      REAL A, B, T, MACHEP, dist0
       
      IF ( ABS(T) < MACHEP) THEN                  ! common math
        dist0 = ABS(A - B)
      ELSE IF (T < 0) THEN                         ! modulo arithmetic
        dist0 = MOD( ABS(A - B), T )
        dist0 = MIN(dist0, -T - dist0) 
!        PRINT *, 'Taking a remainder in dist0'
      ELSE IF (A .GE. 0) THEN                       ! categorical difference
          IF (ABS(A - B) > MACHEP) THEN                 
            dist0 = T
          ELSE
            dist0 = 0
          END IF
      ELSE                                           ! contra-categorical
        IF (ABS(A - B) > MACHEP) THEN
          dist0 = 0
        ELSE
          dist0 = T
        END IF
      END IF

      END


*||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| 
*                  -----------Giftware------------
*||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| 
*      dist2 - the Minkowski distance between two points 
*         D = (SUM ((A[i]-B[i]) % T[i]) ^P )^(1/P) 
*
*              P=2 is the usual Pythagorean/Euclidean distance 
*              P=1 is the city-block distance, which is more appropriate 
*                when measurements are in various units 
*              and P may be in general a nonnegative real number 
*          A [n]           data point                    Real 
*          B [n]           data point                    Real 
*          n               dimension of space            Integer 
*          T [n]           modulo arithmetic divisors    Real
*          p               Minkowski's parameter         Real .GE. 0 
*||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| 
      FUNCTION dist2(A, B, n, T, p)

      IMPLICIT NONE
      INTEGER n
      REAL A, B, T, p 
      DIMENSION A(n), B(n), T(n)

      INTEGER j
      REAL dif, sum, machine, MACHEP, dist0, dist2
      SAVE MACHEP ; DATA MACHEP / -1 /       
      IF (MACHEP < 0) MACHEP = machine(1)

      sum = 0.0
      
*       city block/Manhattan
      IF ( ABS(p-1) < MACHEP ) THEN       ! must not use EQ operator on REAL
        DO j = 1, n
          dif = dist0(A(j), B(j), T(j), MACHEP)
          sum = sum + dif
        END DO
        dist2 = sum
      
*       Pythagorean/Euclidean
      ELSE IF ( ABS(p-2) < MACHEP ) THEN
        DO j = 1, n
          dif = dist0(A(j), B(j), T(j), MACHEP)
          sum = sum + dif * dif
        END DO
        dist2 = SQRT(sum)

*       interpret zeroth order as Hamming distance
      ELSE IF ( ABS(p) < MACHEP ) THEN
        DO j = 1, n
          dif = dist0(A(j), B(j), T(j), MACHEP)        
          IF (dif > MACHEP) sum = sum + ABS(T(j))
        END DO
        dist2 = sum

*       infinite order / maximum.  Exponents > 7 are nearly infinite.
      ELSE IF (p > 7) THEN
        DO j = 1, n
          dif = dist0(A(j), B(j), T(j), MACHEP)
          sum = MAX(sum, dif)
        END DO
        dist2 = sum
          
*       general case      
      ELSE IF (p > 0) THEN
        DO j = 1, n
          dif = dist0(A(j), B(j), T(j), MACHEP)
!          PRINT *, 'dif evaluates to:', dif
          sum = sum + dif ** p
        END DO
        dist2 = sum ** (1 / p)
        
      ELSE                         !      bad input values
        STOP 'dist2: Minkowski parameter must not be negative'
      END IF
      
      END
      

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*                  -----------Giftware------------
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
* calculate the Minkowski distance between all points of an array:
*   D[i,j] = (SUMk ((X[i,k]-X[j,k]) %O) ^P )^(1/P)
*  which for P=2 is the usual Pythagorean/Euclidean distance
*        and for P=1 is the city-block distance, which is more appropriate 
*          when measurements are in various units
*        and P may be in general a nonnegative real number
*         
*        X [m, n]      data points                     Real
*        m             points in array                 Integer
*        n             dimension of space              Integer
*        D [E,m]       array of distances              Real
*                       only lower triangle is filled
*        E             row dimension of A
*        T [n]         modulo arithmetic divisors      Real
*                       essential feature for music theory
*                       set zero for common distances
*        p             Minkowski's parameter           Real .GE. 0
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      SUBROUTINE distance(X, m, n, D, E, T, p)

      IMPLICIT NONE
      REAL X, D, T, p, dif, sum, MACHEP, machine, dist0
      INTEGER m, n, E, i, ii, j
      DIMENSION X(m,n), D(E,m), T(n)
      SAVE MACHEP ; DATA MACHEP / -1 /  
      IF (MACHEP < 0) MACHEP = machine(1)
      
*       city block/Manhattan
      IF ( ABS(p-1) < MACHEP) THEN       ! must not use EQ operator on REAL
        DO ii = 1, m-1
          DO i = ii+1, m
            sum = 0.0
            DO j = 1, n
              dif = dist0(X(i,j), X(ii,j), T(j), MACHEP)
              sum = sum + dif
            END DO
            D(i,ii) = sum          ! put result in the lower triangle of D
          END DO
        END DO 
      
*       Egyptian/Pythagorean/Euclidean  (probably somebody else before that)
      ELSE IF ( ABS(p-2) < MACHEP ) THEN
        DO ii = 1, m-1
          DO i = ii+1, m 
            sum = 0.0
            DO j = 1, n
              dif = dist0(X(i,j), X(ii,j), T(j), MACHEP)
              sum = sum + dif * dif
            END DO
            D(i,ii) = SQRT(sum)
          END DO
        END DO

*       interpret zeroth order as Hamming distance
      ELSE IF ( ABS(p) < MACHEP) THEN
        DO ii = 1, m-1
          DO i = ii+1, m
            sum = 0.0
            DO j = 1, n
              dif = dist0(X(i,j), X(ii,j), T(j), MACHEP)
              IF (dif > MACHEP) sum = sum + ABS(T(j))
            END DO
            D(i,ii) = sum
          END DO
        END DO

*       infinite order / maximum.  Exponents > 7 are nearly infinite.
      ELSE IF (p > 7) THEN
        DO ii = 1, m-1
          DO i = ii+1, m
            sum = 0.0
            DO j = 1, n
              dif = dist0(X(i,j), X(ii,j), T(j), MACHEP)
              sum = MAX(sum, dif)
            END DO
            D(i,ii) = sum
          END DO
        END DO
          
*       general case      
      ELSE IF (p > 0) THEN
        DO ii = 1, m-1
          DO i = ii+1, m
            sum = 0.0
            DO j = 1, n
              dif = dist0(X(i,j), X(ii,j), T(j), MACHEP)
              sum = sum + dif ** p
            END DO
            D(i,ii) = sum ** (1 / p)
          END DO
        END DO
        
      ELSE                      ! bad input for p
        STOP 'distance: Minkowski parameter must not be negative'
      END IF
      
      END


*=======================================================================
*                           -----Giftware-----
*=======================================================================
*   Divisive clustering based on partial sorting, a reduced form of 
*  Hoare's quick sort algorithm.
* 
*          X [mlim,n]                   data array
*          mlim                         row dimension of X
*          m                            number of data points
*          n                            dimension of data
*          k                            desired number of clusters
*          Q [m]                        what cluster each point is in
*          P [k]                        cluster populations
*          T [n]                        modulo value for distance
*          y                            Minkowski's parameter
*          fault                        exit code
*=======================================================================
      SUBROUTINE hclust(X, mlim, m, n, k, Q, P, T, y, fault)

      IMPLICIT NONE
      INTEGER m, mlim, n, k, Q, P, fault
      REAL X, T, y 
      DIMENSION X(mlim,n), Q(m), P(k), T(n)

*          Local variables
      INTEGER hi, lo, ihi, ilo, i, j, L, LL, ind, bound, iprev, randL
      INTEGER  bigsize, bigpart, partsize, iswap, mbits, randH, randM
      REAL Xtop, Xbot, Xtry, Xswap, dist2, dhl, dhm, dml
      DIMENSION Xtop(n), Xbot(n), Xtry(n), bound(0:k), ind(m)
      
*        copy data to scratch file.  Fill array ind[], initialize bound[]
      fault = 0
      OPEN (UNIT=8,FORM='UNFORMATTED',STATUS='SCRATCH',IOSTAT=fault)
      REWIND (UNIT=8)
      WRITE (UNIT=8, IOSTAT=fault) X
      IF (fault .NE. 0) THEN
        PRINT *, 'Trouble writing to scratch file.'
        fault = 8
      END IF     
      DO i = 1, m ; ind(i) = i ; END DO
      bound(0) = 0
      bound(1) = m

*        Get number of bits required to represent m
      i = BIT_SIZE(m)
 10   i = i - 1
      IF (.NOT. BTEST(m, i)) GO TO 10
      mbits = i + 1


*=======================================================================
*        Partition a section of the array such that elements closer to the first
*        element are moved to the front, and elements closer to the last element
*        are moved to the rear.  Continue until enough clusters have been
*        made.
*=======================================================================
      DO L = 1, k-1
 
        partsize = bound(1)         ! find largest partition
        bigsize = partsize
        bigpart = 1
        DO LL = 2, L
           partsize = bound(LL) - bound(LL-1)
          IF (partsize > bigsize) THEN
            bigsize = partsize
            bigpart = LL
          END IF
        END DO
          
        hi = bound(bigpart)               ! find limits of partition
        lo = bound(bigpart-1) +1

*           swap random pivot values to ends of partition
        randH = MIN( INT(RAND(0)*(hi-lo+1)), (hi-lo) ) + lo
 20     randL = MIN( INT(RAND(0)*(hi-lo+1)), (hi-lo) ) + lo
        IF (randL .EQ. randH) GO TO 20
        
*           a third pivot value in case first two are near each other
        IF (partsize > 2) THEN
 30       randM = MIN( INT(RAND(0)*(hi-lo+1)), (hi-lo) ) + lo
          IF (randM .EQ. randH .OR. randM .EQ. randL) GO TO 30
          DO j = 1, n
            Xtop(j) = X(randH,j)
            Xtry(j) = X(randM,j)
            Xbot(j) = X(randL,j)
          END DO
          dhl = dist2(Xtop, Xbot, n, T, y)
          dhm = dist2(Xtop, Xtry, n, T, y)
          dml = dist2(Xtry, Xbot, n, T, y)
          IF (dhm > dhl .OR. dml > dhl) THEN
            IF (dhm > dml) THEN
!              PRINT *, 'Replacing randL with randM'
              randL = randM
            ELSE
!              PRINT *, 'Replacing randH with randM'
              randH = randM
            END IF
          END IF
        END IF
        
        DO j = 1, n                            ! swap
          Xtop(j) = X(randH,j)
          X(randH,j) = X(hi,j)
          X(hi,j) = Xtop(j)
          Xbot(j) = X(randL,j)
          X(randL,j) = X(lo,j)
          X(lo,j) = Xbot(j)
        END DO
        
*=======================================================================
*            sort
*=======================================================================
       ihi = hi - 1
       ilo = lo + 1

  50   IF (ilo .LE. ihi) THEN 
         DO j = 1, n ; Xtry(j) = X(ilo,j) ; END DO
         IF (dist2(Xtry,Xbot,n, T, y).LE.dist2(Xtry,Xtop, n, T, y)) THEN
           ilo = ilo + 1
           GO TO 50
         END IF
       END IF
  
  60   IF (ihi .GE. ilo) THEN
        DO j = 1, n ; Xtry(j) = X(ihi,j) ; END DO
        IF (dist2(Xtry,Xtop, n, T, y).LE.dist2(Xtry,Xbot, n, T, y)) THEN
          ihi = ihi - 1
          GO TO 60
        END IF
       END IF
 
       IF (ilo < ihi) THEN
         DO j = 1, n
           Xswap = X(ilo,j) ; X(ilo,j) = X(ihi,j) ; X(ihi,j) = Xswap
         END DO
         iswap = ind(ilo) ; ind(ilo) = ind(ihi) ; ind(ihi) = iswap
         ilo = ilo + 1
         ihi = ihi - 1
         IF (ilo .LE. ihi) GO TO 50
       END IF
       
*=======================================================================
*         Add this partition boundary to the bounds array       
*=======================================================================
       IF (ilo .EQ. hi) THEN                 ! all distances were equal
         bound(L+1) = (lo + hi) / 2            ! default to equal split
       ELSE
         bound(L+1) = ihi        
         CALL rsort(bound(1), L+1, mbits)  
       END IF
      END DO       ! next partition
      
*=======================================================================
*        Recover cluster memberships and populations from the index array
*=======================================================================
      L = 1
      iprev = 0
      DO i = 1, m
        IF (i .EQ. bound(L)) THEN
          P(L) = i - iprev
          iprev = i
        END IF
        IF (i > bound(L)) L = L + 1
        Q(ind(i)) = L
      END DO

*        restore data from scratch file
      REWIND (UNIT=8)
      READ (UNIT=8, IOSTAT=fault) X
      IF (fault .NE. 0) THEN
        PRINT *, 'Trouble reading scratch file'
        fault = 8
      END IF
      CLOSE (UNIT=8)
      
      END  ! of hclust


************************************************************************
*                         *** Giftware ***
************************************************************************
* sclust - spectral clustering with a sparse affinity matrix
* 
* _Name___________Type_______In/Out_______Description__________________
*  X[m,n]         Real       in           data points in R^n
*  m              Integer    in           number of points
*  n              Integer    in           dimension of data
*  C1[kmax,kmax]  Real       work array   cluster centers in eigenspace
*  kmin           Integer    in           fewest clusters to make
*  k              Integer    in/out       default/actual clusters made
*  kmax           Integer    in           most clusters to make
*  Q[m]           Integer    out          cluster memberships of each point
*  P[kmax]        Integer    work array   population in seed clusters ONLY
*  T[n]           Real       in           period of modulo arithmetic
*  y              Real       in           Minkowski's exponent
*  cauchy         Logical    in           whether to use Cauchy distribution
*                                          (or else Gauss)
*  brief          Logical    in           whether to skip clustering retries
*  iter           Integer    in           maximum weighted k-means iterations 
*  dfunc          Function   in           Real valued function of prototype
*                                          dfunc(A[n], B[n], n, T[n], y)
*                                            REAL A, B, T, y; INTEGER n
*  u              Real       out          clustering residual 
*  fault          Integer    out          error code
************************************************************************
      SUBROUTINE sclust(X, m, n, C1, kmin, k, kmax,
     &      Q, Qbest, P, T, y, f, cauchy, brief, iter, dfunc, u, fault)

*          transfer list
       IMPLICIT NONE
       INTEGER m, n, kmin, k, kmax, Q, Qbest, P, iter, fault
       REAL X, C1, T, y, f, dfunc, u
       LOGICAL cauchy, brief
       DIMENSION X(m,n), C1(kmax, kmax), Q(m), Qbest(m), P(kmax), T(n)

*        constants
       INTEGER h, luck           ! values to keep in each column, extra vectors
       PARAMETER (h = 20, luck=7) 

*          local variables
       INTEGER 
     &         g, i, j, o, r,                 ! counters
     &         IA(2*m*h+m), JA(2*m*h+m),      ! indices of matrix A
     &         itop(h),                ! positions of small distances in column
     &         iswap,
     &         ktry,                   ! possible number of clusters
     &         mbits                   ! bits to represent m
       
       REAL 
     &      D(m), DSR(m), DISR(m),       ! sum of rows of A, D^(-1/2)
     &      gap, biggap,                 ! eigengaps
     &      machine,                     ! external function
     &      tmax,                        ! largest of smallest distances
     &      ubest,                       ! lowest residual
     &      W(m, kmax),                  ! eigenvectors passed to wkm
     &      xi(n), xj(n),                ! points in R^n
     &      zero1(kmax)                  ! vector of zeros

       DOUBLE PRECISION  
     &                   alpha,             ! similarity parameter
     &                   A(2*m*h+m),        ! matrix entries
     &                   BIG,               ! very large number
     &                   dd,                ! a distance
     &                   dmachine,          ! external function
     &                   EPS,               ! machine double precision
     &                   L((kmax+1)*luck),  ! eigenvalues
     &                   s,                 ! typical distance
     &                   swap,  
     &                   top(h),            ! shortest distances in a column
     &                   V(m,(kmax+1)*luck) ! eigenvectors returned by bla

       EXTERNAL dfunc
       
************************************************************************
*       begin
************************************************************************
       fault = 0
       DO j = 1, kmax ; zero1(j) = 0. ; END DO
       EPS = dmachine(1)
       BIG = dmachine(3)

*        build a sparse matrix, first get distance
       r = 0
       DO j = 1, m
         DO i = 1, h ; top(i) = BIG ; END DO
         DO i = 1, m
           IF (i .EQ. j) CYCLE
           DO o = 1, n ; xi(o) = X(i,o) ; xj(o) = X(j,o) ; END DO
           dd = DBLE(dfunc(xi, xj, n, T, y))  ! upsample

*            add jitter (so that order in X should not be important)
           dd = dd * (1. + h * EPS * (RAND(0) -.5) )
                  
*           find least values in each column      
           IF (dd < top(h)) THEN
             itop(h) = i
             top(h) = dd
             tmax = 0.               ! find biggest small distance
             DO o = 1, h
               IF (top(o) > tmax) THEN
                 tmax = top(o)
                 g = o
               END IF      
             END DO             

*                swap distances
             iswap = itop(g) ; itop(g) = itop(h) ; itop(h) = iswap
             swap = top(g) ; top(g) = top(h) ; top(h) = swap
           END IF
         END DO
            
*           add to array
         DO o = 1, h
           r = r + 1
           A(r) = top(o)                ! insert the point
           IA(r) = itop(o)
           JA(r) = j

           r = r + 1
           A(r) = top(o)               ! insert its transpose
           IA(r) = j
           JA(r) = itop(o)
         END DO
       END DO
  
*       remove duplicate entries
       mbits = BIT_SIZE(m)
       DO WHILE (.NOT. BTEST(m, mbits-1)) 
         mbits = mbits - 1
       END DO

       CALL riid(IA, JA, A, r, mbits)             ! sort on i
       CALL riid(JA, IA, A, r, mbits)             ! sort on j
       
!!       debug:
!       DO o = 1, r
!         PRINT 37, o, IA(o), JA(o), A(o)
!       END DO
!  37   FORMAT ('pos', I6, ' i=', I3,' j=',I3,' A=', G12.2)
              
       g = 1
       DO o = 2, r
         IF (IA(o) .EQ. IA(g) .AND. JA(o) .EQ. JA(g)) CYCLE
         g = g + 1
         A(g) = A(o) ; IA(g) = IA(o) ; JA(g) = JA(o)
       END DO
       r = g

************************************************************************
*           make graph Laplacian
************************************************************************
*        take typical distance to be the longest distance
       s = 0.
       DO o = 1, r
         s = MAX(s, A(o))
       END DO
       s = MAX(s, m * EPS)             ! don't divide by zero
         
*         convert distances to affinities
       alpha = f * (s **2) / 2
       IF (cauchy) THEN
         DO i = 1, r
           A(i) = 1 / (A(i) **2 + alpha)
         END DO
       ELSE
         DO i = 1, R
           A(i) = EXP( -(A(i) **2) / alpha)
         END DO
       END IF
         
*          normalize the matrix
       DO i = 1, m
         D(i) = 0.
         DSR(i) = 0.
         DISR(i) = 0.
       END DO         
         
       DO o = 1, r                   ! row sums
         D(IA(o)) = D(IA(o)) + SNGL(A(o))
         D(JA(o)) = D(JA(o)) + SNGL(A(o))
       END DO
  
       DO i = 1, m
         DSR(i) = SQRT(D(i))
         DISR(i) = 1 / DSR(i)
       END DO
           
*         A' = I - D^(-1/2) @ A @ D^(-1/2)
       DO o = 1, r
         A(o) = -1 * DISR(IA(o)) * A(o) * DISR(JA(o))
       END DO
       DO i = 1, m          ! ones on the diagonal
         A(r+i) = 1.
         IA(r+i) = i
         JA(r+i) = i
       END DO
       r = r + m
       
* DEBUG:  dump the graph Laplacian
!      DO i = 1, M
!        DO J = 1, M
!          PRINT 530, A(i,j)
!        END DO
!        PRINT 502
!      END DO 
! 530  FORMAT (F6.3, $)
!      PAUSE

************************************************************************
*       pull the eigenvectors by the Lanczos algorithm
************************************************************************
       CALL bla(A, IA, JA, r, m, m, kmax+1, L, V, fault) ! not the fastest
       PRINT *, 'bla returns #', fault
       IF (fault > 0) RETURN                  ! give up
       
*      dump the output
!      PRINT *, 'First ten eigenvalues:'
!      DO i = 1, 10
!        PRINT 501, V(i)
!      END DO
!      print 502
!      
!      PRINT *, 'Beginning of first eigenvector:'
!      DO i = 1, 100
!        PRINT 501, X(i)
!      END DO
!      PRINT 502
!      
! 501   FORMAT (F10.3, ' ', $)
! 502   FORMAT ()

*           look for widest eigengap
       biggap = 0
       ktry = -1
       
       DO j = 2, kmax
         gap = L(j+1) - L(j)
         IF (gap > biggap) THEN
           biggap = gap
           ktry = j
         END IF
       END DO
              
       IF (biggap .EQ. 0) PRINT *, 'sclust: no eigengap found'
       PRINT *, 'found ', ktry, ' as a natural number of clusters'
       IF (ktry < kmin .OR. ktry > kmax) ktry = k
       k = ktry     ! return value of k

*       reduce precision
       DO j = 1, k
         DO i = 1, m
           W(i,j) = SNGL(V(i,j))
         END DO
       END DO

************************************************************************
*                    cluster
************************************************************************
       ubest = machine(3)
       DO o = 1, 6
         IF (BTEST(o, 0)) THEN
           CALL hclust(W, m, m, k, k, Q, P, zero1, 2.0, fault)
           IF (fault > 0) PRINT *, 'hclust: failed at sparse step'       
         ELSE
           DO i = 1, m
             Q(i) = MIN(INT(RAND(0) * k) + 1, k)
           END DO
         END IF

         DO g = 1, k ; P(g) = 0 ; END DO
         DO g = 1, k ; DO j = 1, k ; C1(j,g) = 0. ; END DO ; END DO
         DO i = 1, m 
           g = Q(i)
           P(g) = P(g) + 1                     ! count
           DO j = 1, k
             C1(g,j) = C1(g,j) + W(i,j)          ! add
           END DO
         END DO
         DO g = 1, k
           i = P(g)
           DO j = 1, k
             C1(g,j) = C1(g,j) / FLOAT(i)        ! divide
           END DO
         END DO
       
         CALL wkm(W, m, m, kmax, k, C1, D, DSR, DISR, Q, iter, u, fault)
         IF (fault > 0) PRINT *, 'wkm: error#', fault
         fault = 0   ! ignore clustering errors
         IF (brief) RETURN
         
         IF (u < ubest) THEN
           ubest = u
           DO i = 1, m
             Qbest(i) = Q(i)
           END DO
         END IF
       END DO ! next o
         
       DO i = 1, m
         Q(i) = Qbest(i)
       END DO
       
      END ! of sclust       


************************************************************************
*  R : A Computer Language for Statistical Data Analysis
*  Copyright (C) 2004   The R Development Core Team.
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, a copy is available at
*  http://www.r-project.org/Licenses/
************************************************************************
*      this subroutine is derived from the R function kmeans_Lloyd
*
*_________variable__I/O_______Description___________________type________
*         x [m,n]   in        data points                   Real
*         m         in        number of points              Integer
*         n         in        dimension of the data         Integer
*         c [k,n]   both      center points of clusters     Real
*         klim      in        row dimension of c            Integer
*         k         in        number of clusters to make    Integer
*         a [r,k]   in        observer frequencies          Integer
*         r         in        number of observers           Integer
*         q [m]     out       what cluster a point is in    Integer
*         t [n]     in        period of modulo arithmetic   Real
*         p [k]     out       number of points in cluster   Integer
*         thuro     in        examine points more times     Logical
*         iter      in        maximum iterations            Integer
*         u         out       total distance                Real
*         fault     out       error code                    Integer
************************************************************************
      SUBROUTINE kmeans(x,m,n,c,klim,k,a,r,q,p,t,thuro,iter,u,fault)
       
      IMPLICIT NONE
      INTEGER m, n, klim, k, a, r, q, p, iter, fault
      REAL x, c, t, u
      LOGICAL thuro
      DIMENSION x(m,n), c(klim,n), a(r,k), q(m), p(k), t(n)

      INTEGER h, i, j, h1, L, L1, livex, Lnew, o, qprev1, qprev2, s, ttl
      REAL best, sum, dif, z
      REAL BIG, MACHEP, machine, median, mode
      LOGICAL updated, livec
      INTEGER pcount        ! debug: how many live points
      DIMENSION livec(k), livex(m), qprev1(m), qprev2(m), s(n), z(m)
      
      DATA BIG, Lnew /1.0E30, 1 /
      MACHEP = machine(1)
      BIG = machine(3)
      
      fault = 0                          ! validate
      IF (k .LE. 1 .OR. k .GE. m) THEN
       fault = 3
       RETURN
      END IF
      
      DO j = 1, n                     ! do comparisons once per kmeans() call
       IF (ABS(t(j)) < MACHEP) THEN
        s(j) = 0                              ! common
       ELSE IF (t(j) < 0) THEN
        s(j) = -1                             ! modulo
       ELSE IF (x(1,j) .GE. 0) THEN
        s(j) = 1                              ! categorical
       ELSE 
        s(j) = 2                              ! contra-categorical
       END IF
      END DO

      ttl = 2              ! this can be increased for slightly better accuracy
      IF (thuro) ttl = 3
      DO i = 1, m 
       livex(i) = ttl       
       qprev1(i) = -1
       q(i) = 0
      END DO
      DO L = 1, k ; livec(L) = .TRUE. ; END DO

*          main loop 
      DO o = 1, iter 

       DO i = 1, m                  ! watch for repetition
        qprev2(i) = qprev1(i) ; qprev1(i) = q(i)
       END DO
       updated = .FALSE.

************************************************************************
*             find nearest center for each point 
************************************************************************
       DO i = 1, m
        IF (o > 1) THEN
         L1 = q(i)
         IF (livex(i) < 1) CYCLE           ! skip dead points
         IF (p(L1) .EQ. 1) CYCLE    ! don't empty out a cluster
        END IF
        best = BIG
        DO L = 1, k
         sum = 0.0
         DO j = 1, n          
          IF (s(j) .EQ. 0) THEN                  ! common math
           dif = x(i,j) - c(L,j)
          ELSE IF (s(j) .EQ. -1) THEN               ! modulo arithmetic
           dif = MOD(ABS(x(i,j) - c(L,j)), t(j))
           dif = MIN(dif, -t(j) - dif)
          ELSE IF (s(j) .EQ. 1) THEN         ! categorical difference
           IF (ABS(x(i,j) - c(L,j)) > MACHEP) THEN ; dif = t(j) 
           ELSE ; dif = 0 ; END IF
          ELSE                            ! contra-categorical difference
           dif = a(-NINT(X(i,j)),L)
           IF (o > 1) THEN               ! attempt to stabilize algorithm
            IF (L .EQ. L1) THEN
             dif = dif - 1
            END IF
           END IF
           dif = dif * t(j) / p(L) ! weight by t, and per-capita of cluster
          END IF                      ! intended to not equalize all clusters
          sum = sum + dif * dif
         END DO
           
         IF (sum < best) THEN
          best = sum
          Lnew = L
         END IF	   
        END DO        ! next L
        
        IF (q(i) .NE. Lnew) THEN
         IF (o > 1) THEN 
          p(L1) = p(L1) - 1
          p(Lnew) = p(Lnew) + 1
          livec(L1) = .TRUE.              ! recenter old cluster
          livec(Lnew) = .TRUE.           ! recenter new cluster
          livex(i) = ttl                 ! reset ttl for point
         END IF
         q(i) = Lnew
         updated = .TRUE.
        ELSE
         livex(i) = livex(i) - 1     ! decrement ttl for point
        END IF          
       END DO         ! next i

       IF (.NOT. updated) GO TO 90

*        all points live in live clusters
       pcount = 0
       DO i = 1, m
        IF (livex(i) > 0) pcount = pcount + 1
        IF (thuro) THEN
         L1 = q(i)                           ! can cause slowdowns
         IF (livec(L1)) livex(i) = ttl
        END IF
       END DO
       
*        fill in empty clusters
       IF (o .EQ. 1) THEN
        DO L = 1, k ; p(L) = 0 ; END DO
        DO i = 1, m                        ! count
         L1 = q(i)
         p(L1) = p(L1) + 1
        END DO

        DO 40 L = 1, k                       ! redistribute 
         IF (p(L) .EQ. 0) THEN
          DO i = 1, m
           L1 = q(i)
           IF (p(L1) > 1) THEN
            q(i) = L
            p(L) = 1
            p(L1) = p(L1) - 1
            GO TO 40
           END IF   ! non-singleton cluster ?
          END DO 
          fault = 1       ! remedial transfer fails
         END IF
  40    CONTINUE  
       END IF ! first iteration ?
     
************************************************************************
*        update each center 
************************************************************************
       DO j = 1, n                          ! zero
        IF (s(j) .NE. 0) CYCLE
        DO L = 1, k
         IF ( (.NOT. livec(L)) .OR. (p(L) .EQ. 0) ) CYCLE ! skip dead / empty
          c(L,j) = 0.0 
         END DO  
       END DO     
      
       DO i = 1, m                           ! add
        L1 = q(i)
        IF (.NOT. livec(L1)) CYCLE
        DO j = 1, n 
         IF (s(j) .NE. 0) CYCLE
         c(L1,j) = c(L1,j) + x(i,j)
        END DO
       END DO
       
       DO j = 1, n                    ! divide
        IF (s(j) .NE. 0) CYCLE
        DO L = 1, k
         IF ( (.NOT. livec(L)) .OR. (p(L) .EQ. 0) ) CYCLE
         c(L,j) = c(L,j) / p(L)
        END DO
       END DO
        
*           use the mode for categorical variables   
       DO L = 1, k
        IF ( (.NOT. livec(L)) .OR. (p(L) .EQ. 0)) CYCLE
        DO j = 1, n
         IF (s(j) .NE. 1 .AND. s(j) .NE. -1) CYCLE
         h = 0
         DO i = 1, m
          IF (q(i) .NE. L) CYCLE
          h = h + 1
          z(h) = x(i, j)             
         END DO
         IF (s(j) .EQ. 1) THEN
           C(L,j) = mode(z, h, fault)
           IF (fault .NE. 0) PRINT *, 'mode:  Error#', fault
         ELSE 
           C(L,j) = median(z, h, t(j), fault)
           IF (fault > 0) PRINT *, 'median:  Error#', fault
         END IF
        END DO     ! next j
       END DO    ! next k

*           contra-categorical frequencies
       DO j = 1, n
        IF (s(j) .NE. 2) CYCLE  ! skip others        
        DO L = 1, k                              ! clear
         DO h = 1, r ; A(h,L) = 0 ; END DO
        END DO
        DO i = 1, m                              ! count
         L1 = q(i)
         h1 = -NINT(x(i,j))
!         PRINT *, 'L1=',L1,' h1=',h1,' x=',X(i,j),' a=',a(h1,L1)
         a(h1,L1) = a(h1,L1) + 1
        END DO
       END DO
       
!        DEBUG::
       PRINT *, 'iter=', o, ' with ', pcount, ' live points.'
*          look for looping
       DO i = 1, m
        IF (q(i) .NE. qprev2(i)) GO TO 80
       END DO
       PRINT *, 'Quitting on the loop criterion'
       GO TO 90
      
 80    IF (MOD(o,50) .EQ. 0) THEN                 ! occasional optimal transfer
        DO L = 1, k ; livec(L) = .TRUE. ; END DO
        DO i = 1, m ; livex(i) = ttl ; END DO
       ELSE
        DO L = 1, k ; livec(L) = .FALSE. ; END DO
       END IF
      
      END DO  ! next o
      fault = 2      ! too many iterations

************************************************************************
  90  CONTINUE      !      total distance
************************************************************************
      PRINT *, 'done after ', o, ' iterations. '
!      PRINT *, A
      u = 0.0
      DO i = 1, m
       L1 = q(i)
       DO j = 1, n       
        IF (s(j) .EQ. 0) THEN                  ! common math
         dif = x(i,j) - c(L1,j)
        ELSE IF (s(j) .EQ. -1) THEN                    ! modulo arithmetic
         dif = MOD(ABS(x(i,j) - c(L1,j)), t(j))
         dif = MIN(dif, -t(j) - dif)
        ELSE IF (s(j) .EQ. 1) THEN                ! categorical difference
         IF (ABS(x(i,j) - c(L1,j)) > MACHEP) THEN ; dif = t(j)
         ELSE ; dif = 0 ; END IF
        ELSE                                 ! counter-categorical difference
         dif = a(-NINT(x(i,j)),L1) - 1
         IF (dif < 0) PRINT *, 'bad value of A!'
         dif = dif * t(j) / p(L1)
        END IF
        u = u + dif * dif 
       END DO
      END DO
      
      RETURN
      END

************************************************************************
*  R : A Computer Language for Statistical Data Analysis
*  Copyright (C) 2004   The R Development Core Team.
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, a copy is available at
*  http://www.r-project.org/Licenses/
************************************************************************
*  a k-medians clustering routine based on the Gnu R routine kmeans_Lloyd
*
*______variable_____I/O________Description___________________type_______
*       x [m,n]     in        data points                   Real
*       m           in        number of points              Integer
*       n           in        dimension of the data         Integer
*       c [klim,n]  both      center points of clusters     Real
*       klim        in        row dimension of c            Integer
*       k           in        number of clusters to make    Integer
*       a [r,k]     in        observer frequencies          Integer
*       r           in        number of observers           Integer
*       q [m]       out       what cluster a point is in    Integer
*       p [k]       out       # of points in each cluster   Integer
*       t [n]       in        period of modulo arithmetic   Real
*       thuro       in        keep points live              Logical
*       iter        in        maximum iterations            Integer
*       u           out       residual                      Real
*       fault       out       error code                    Integer
************************************************************************
      SUBROUTINE kmedians(x,m,n,c,klim,k,a,r,q,p,t,thuro,iter,u,fault)
       
      IMPLICIT NONE
      INTEGER m, n, klim, k, a, r, q, p, iter, fault
      REAL x, c, t, u
      LOGICAL thuro
      DIMENSION x(m,n), c(klim,n), a(r,k), q(m), p(k), t(n)

      INTEGER h, h1, i, j, L, L1, livex, Lnew, o, qprev1, qprev2, s, ttl
      REAL best, sum, dif, BIG, z, MACHEP
      LOGICAL updated, livec
      DIMENSION livec(k), livex(m), qprev1(m), qprev2(m), s(n), z(m)
      REAL median, mode, machine        ! external functions
      INTEGER pcount                              ! how many live points
      
      SAVE MACHEP, BIG ; DATA MACHEP, BIG /-1, 1.0E30 /
      IF (MACHEP < 0) MACHEP = machine(1)
      BIG = machine(3)
      
      fault = 0                          ! validate, initialize
      IF (k .LE. 1 .OR. k .GE. m) THEN
       fault = 3
       RETURN
      END IF

*          do comparisons once per function call
      DO j = 1, n
       IF (ABS(t(j)) < MACHEP) THEN 
        s(j) = 0                              ! common
       ELSE IF (t(j) < 0) THEN
        s(j) = -1                             ! modulo
       ELSE IF (x(1,j) .GE. 0) THEN
        s(j) = 1                              ! categorical
       ELSE
        s(j) = 2                              ! contra-categorical
       END IF
      END DO

      ttl = 2
      IF (thuro) ttl = 3
      DO L = 1, k ; livec(L) = .TRUE. ; END DO   ! initialize time-to-live
      DO i = 1, m 
       livex(i) = ttl
       qprev1(i) = -1
       q(i) = 0
      END DO

*                      main loop
      DO o = 1, iter 
      
       DO i = 1, m    ! save previous values
        qprev2(i) = qprev1(i) ; qprev1(i) = q(i)
       END DO
       updated = .FALSE.

************************************************************************
*             find nearest center for each point 
************************************************************************
       DO i = 1, m
        IF (o > 1) THEN
         L1 = q(i) 
         IF (livex(i) < 1) CYCLE  ! skip dead points
         IF (p(L1) .EQ. 1) CYCLE  ! don't empty a cluster
        END IF

        best = BIG        
        DO L = 1, k
         sum = 0.0
         DO j = 1, n
          IF (s(j) .EQ. 0) THEN                         ! common arithmetic
           dif = ABS(x(i,j) - c(L,j))
          ELSE IF (s(j) .EQ. -1) THEN                   ! modulo arithmetic
           dif = MOD(ABS(x(i,j) - c(L,j)), t(j))
           dif = MIN(dif, -t(j) - dif)
          ELSE IF (s(j) .EQ. 1) THEN                    ! categorical difference
           IF (ABS(x(i,j) - c(L,j)) > MACHEP) THEN ; dif = t(j)
           ELSE ; dif = 0 ; END IF
          ELSE                                   ! contra-categorical difference
           dif = a(-NINT(x(i,j)),L)
           IF (o > 1) THEN               ! attempt to stabilize algorithm
            IF (L .EQ. L1) THEN
             dif = dif - 1
            END IF
           END IF
           dif = dif * t(j) / p(L)
          END IF
          
          sum = sum + dif
         END DO
           
         IF (sum < best) THEN
          best = sum
          Lnew = L
         END IF	   
        END DO        ! next L
        
        IF (q(i) .NE. Lnew) THEN
         IF (o > 1) THEN  
          p(L1) = p(L1) - 1
          p(Lnew) = p(Lnew) + 1
          livec(L1) = .TRUE.           ! recenter old cluster
          livec(Lnew) = .TRUE.        ! recenter new cluster
          livex(i) = ttl              ! reset ttl for point
         END IF
         q(i) = Lnew
         updated = .TRUE.
        ELSE
         livex(i) = livex(i) - 1      ! decrement ttl for point
        END IF          
       END DO         ! next i
       
       IF (.NOT. updated) GO TO 90

!          this has no benefit for the House Votes data set
       pcount = 0
       DO i = 1, m
        IF (thuro) THEN
         L1 = q(i)
         IF (livec(L1)) livex(i) = ttl     ! update all points in live clusters 
        END IF
        IF (livex(i) > 0) pcount = pcount + 1
       END DO
     
************************************************************************
*          fill empty clusters
************************************************************************
       IF (o .EQ. 1) THEN                    
        DO L = 1, k ; p(L) = 0 ; END DO          ! count
        DO i = 1, m
         L1 = q(i)
         p(L1) = p(L1) + 1
        END DO
        
        DO 40 L = 1, k                          ! redistribute
         IF (p(L) .EQ. 0) THEN
          DO i = 1, m
           L1 = q(i)
           IF (p(L1) > 1) THEN
            q(i) = L
            p(L) = 1
            p(L1) = p(L1) -1
            GO TO 40
           END IF
          END DO
          fault = 1        ! remedial transfer fails
         END IF
 40     CONTINUE
       END IF ! first iteration?
             
************************************************************************
*          medians or modes for each dimension of points in each cluster       
************************************************************************
       DO L = 1, k
        IF (.NOT. livec(L) .OR. p(L) .EQ. 0) CYCLE      ! skip dead clusters
        DO j = 1, n
         IF (s(j) .EQ. 2) CYCLE
         h = 0
         DO i = 1, m
          IF (q(i) .NE. L) CYCLE
          h = h + 1
          z(h) = x(i,j)
         END DO
         IF (s(j) < 1) THEN
          C(L,j) = median(z, h, t(j), fault)
          IF (fault > 0) PRINT *, 'median: Error#', fault
         ELSE 
          C(L,j) = mode(z, h, fault)
          IF (fault > 0) PRINT *, 'mode:  Error#', fault
         END IF
        END DO   ! next j
       END DO   ! next L
        
*           contra-categorical frequencies
       DO j = 1, n
        IF (s(j) .NE. 2) CYCLE               ! skip regular data
        DO L = 1, k                              ! clear
         DO h = 1, r ; a(h,L) = 0 ; END DO
        END DO
        DO i = 1, m                              ! count
         L1 = Q(i)
         h1 = -NINT(x(i,j))
         a(h1,L1) = a(h1,L1) + 1
        END DO
       END DO
       
!        debug: 
       PRINT *,'iter ', o, ' live pts=',pcount

*            check if anything is changing
       DO i = 1, m
        IF (q(i) .NE. qprev2(i)) GO TO 80
       END DO    
       PRINT *, 'Quitting on loop criterion after ', o, ' iterations'
       GO TO 90
       
 80    IF (MOD(o,50) .EQ. 0) THEN
        DO L = 1, k ; livec(L) = .TRUE. ; END DO
        DO i = 1, m ; livex(i) = ttl ; END DO
       ELSE
        DO L = 1, k ; livec(L) = .FALSE. ; END DO 
       END IF
      END DO                   ! next o
      fault = 2                ! too many iterations
  
************************************************************************
 90   CONTINUE               !  total distance
************************************************************************
      u = 0.0      
      DO i = 1, m
       L1 = q(i)
       DO j = 1, n
        IF (s(j) .EQ. 0) THEN
         dif = ABS(x(i,j) - c(L1,j))
        ELSE IF (s(j) .EQ. -1) THEN
         dif = MOD(ABS(x(i,j) - c(L1,j)), t(j))
         dif = MIN(dif, -t(j) - dif)
        ELSE IF (s(j) .EQ. 1) THEN
         IF (ABS(x(i,j) - c(L1,j)) > MACHEP) THEN ; dif = t(j)
         ELSE ; dif = 0 ; END IF
        ELSE
         dif = a(-NINT(x(i,j)),L1) - 1
         
         dif = dif * t(j) / p(L1)
        END IF
  
        u = u + dif
       END DO
      END DO
      
      END


************************************************************************
*  R : A Computer Language for Statistical Data Analysis
*  Copyright (C) 2004   The R Development Core Team.
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, a copy is available at
*  http://www.r-project.org/Licenses/
************************************************************************
*    a derivative work modified from the Gnu R function kmeans_Lloyd.  
* Any errors in this program are not the problem of the R maintainers.  
* Instead, send bug reports to:
*              andy_a@users.sourceforge.net
************************************************************************
*      weighted k-means step of the Bach-Jordan algorithm
* The distance from cluster center r to point i is replaced with 
*       |V[i,j] @ D[i]^(-1/2) - mu[r,j]|         over j
* where    mu[L,j] = SUM(D[i]^(1/2) @ V[i,j]) / SUM(D[i])
* and    V contains eigenvectors in its columns  
*    
*      refer to:  Francis R. Bach and Michael I. Jordan, "Learning Spectral
* Clustering," University of California Berkeley technical report, 2003.
*
*                          ARGUMENTS
*
*_________variable_____I/O_______Description_____________________type________
*          v [mlim,k]   in        first k eigenvectors            Real
*          mlim         in        row dimension of V              Integer
*          m            in        number of points                Integer
*          klim         in        row dimension of C              Integer
*          k            in        number of clusters to make      Integer
*          c [klim,k]   in        center points of clusters       Real
*          d [m]        in        rows sums of similarity matrix  Real
*          dsr [m]      in        d[m] ^ (1/2)                    Real
*          disr [m]     in        d[m] ^ (-1/2)                   Real
*          q [m]        out       what cluster a point is in      Integer
*          iter         in        maximum iterations              Integer
*          u            out       total distance                  Real
*          fault        out       error code                      Integer
************************************************************************
      SUBROUTINE wkm(v,mlim,m,klim,k,c,d,dsr,disr,q,iter,u,fault)
       
      IMPLICIT NONE
      INTEGER m, mlim, klim, k, q, iter, fault
      REAL v, c, d, dsr, disr, u, w
      DIMENSION v(mlim, k), c(klim, k), d(m), dsr(m), disr(m), q(m)

      INTEGER i, j, L, Lnew, n, o, p, L1
      REAL best, dd, tmp, big, sd, sdsrv, mu
      LOGICAL updated
      DIMENSION p(k), sd(k), sdsrv(k,k), mu(k,k)
      
      DATA big, Lnew /1.0E30, 1 /
      
      fault = 0                          ! validate
      n = k                           ! define n for clarity only
      IF (k .LE. 1 .OR. k .GE. m) THEN
        fault = 3
        RETURN
      END IF

      DO o = 1, iter            ! main loop
       updated = .FALSE.

*             find nearest center for each point 
       IF (o > 1) THEN
        DO i = 1, m
         IF (p(q(i)) .EQ. 1) CYCLE   ! don't empty a cluster
         best = big
         DO L = 1, k
          dd = 0.0
          
          DO j = 1, n
           tmp = v(i,j) * disr(i) - mu(L,j)
           dd = dd + tmp * tmp
          END DO
          
          IF (dd < best) THEN
           best = dd
           lnew = L
          END IF
         END DO
         
         IF (q(i) .NE. Lnew) THEN
          L1 = q(i)
          p(L1) = p(L1) -1
          p(Lnew) = p(Lnew) + 1
          updated = .TRUE.
          q(i) = Lnew
         END IF
        END DO
        
       ELSE
        DO i = 1, m
         best = big        
         DO L = 1, k
          dd = 0.0
 
          DO j = 1, n
           tmp = v(i,j) - c(L,j)
           dd = dd + tmp * tmp
          END DO
            
          IF (dd < best) THEN
           best = dd
           lnew = L
          END IF	   
         END DO        ! next L
         IF (q(i) .NE. Lnew) THEN 
           updated = .TRUE.
           q(i) = Lnew
         END IF
        END DO         ! next i

        DO L = 1, k ; p(L) = 0 ; END DO         ! count
        DO i = 1, m
         L1 = q(i)
         p(L1) = p(L1) + 1
        END DO
         
        DO 40 L = 1, k                    ! redistribute
         IF (p(L) .EQ. 0) THEN
          DO i = 1, m
           L1 = q(i)
           IF (p(L1) > 1) THEN
            q(i) = L
            p(L) = 1
            p(L1) = p(L1) - 1
            GO TO 40
           END IF
          END DO
          fault = 1            ! remedial transfer fails
         END IF
  40    CONTINUE
       END IF
       
       IF (.NOT. updated) GO TO 90
       
*        update each center 
       DO j = 1, n ; DO L = 1, k               ! initialize to zero
         sdsrv(L,j) = 0.0
         mu(L,j) = 0.0
       END DO ; END DO
       DO L =  1, k 
        sd(L) = 0.0 
       END DO  
       
       DO i = 1, m                  ! add
        L1 = q(i)
        sd(L1) = sd(L1) + d(i)          !   SUM d[i]        
        DO j = 1, n 
         sdsrv(L1,j) = sdsrv(L1,j) + dsr(i) * v(i,j)   ! SUM d[i]^(1/2) @ v[i,j]
        END DO
       END DO  ! next i

       DO j = 1, n                  ! divide
        DO L = 1, k
         mu(L,j) = sdsrv(L,j) / sd(L)
        END DO
       END DO
      
      END DO  ! next o
      fault = 2      ! too many iterations

  90  CONTINUE    !      total distance
      u = 0.0D0
      DO i = 1, m
        L1 = q(i)
        w = 0.0D0
        DO j = 1, n
          tmp = v(i,j) * disr(i) - mu(L1,j)
          w = w + tmp * tmp
        END DO
        u = u + d(i) * w
      END DO

      END  ! of wkm

*+++++++++++++++++++++++++End of file kasc.f++++++++++++++++++++++++++++
