*-----------------------------------------------------------------------
*   KMNS - by J.A. Hartigan & M.A. Wong
* Code in this file based on Applied Statistics algorithms
* (C) Royal Statistical Society 1979
*
*  according to Statlib (http://lib.stat.cmu.edu/apstat/): 
* "The Royal Statistical Society holds the copyright to these routines, but 
* has given its permission for their distribution provided that no fee is
* charged."
*          
*   Based on a version taken from the Gnu R project.
C
C all variables are now declared.                  B.D. Ripley 1998/06/17
C Declaration re-ordering to satisfy "f77 -ansi",  M.Maechler  2001/04/12
C new remedial transfer phase, fewer arguments     A. Allinger 2011/05/30
C
C
C     ALGORITHM AS 136  APPL. STATIST. (1979) VOL.28, NO.1
C
C     Divide M points in N-dimensional space into K clusters so that
C     the within cluster sum of squares is minimized.
C
*                             I/O LIST
*
*     _Name________________Description_______________I/O____Data_Type___
*     A [M, N]     the points                        In      Real
*     M            number of points                  In      Integer
*     N            dimension of the data R^N         In      Integer
*     C [KLIM, N]  the cluster centers               Both    Real   
*     KLIM         row dimension of C                In      Integer   
*     K            desired number of clusters        In      Integer
*     IC1 [M]      nearest center to a data point    Out     Integer
*     NC [K]       number of points in each cluster  Out     Integer
*     ITER         maximum iterations                In      Integer
*     U            total squared distance,           Out     Real
*                   points to centers
*     IFAULT       error code                        Out     Integer
*
*                         LOCAL VARIABLES
*
*     _Name________________Description____________________Data_Type___
*      AN1 [K]    ratio that corrects a square distance     Real
*                  to a center if a point were removed
*                  from the cluster
*      AN2 [K]    ratio that corrects a square distance     Real
*                  to a center if a point were added
*                  to the cluster
*      NCP [K]    step the cluster is no longer updated     Integer
*      D [M]      distance from point to where its          Real
*                  nearest center would be, if the point
*                  were not in the cluster
*      ITRAN[K]   relates to whether there is convergence   Integer
*      LIVE[K]         "                "
*      DT         square distance
*      ISTEP      steps in the quick-transfer phase         Integer
*      ICOUNT     steps in quick-transfer w/o a transfer    Integer
*-----------------------------------------------------------------------

      SUBROUTINE KMNS(A,M,N,C,KLIM,K,IC1,NC,ITER,U,IFAULT)

      IMPLICIT NONE
      INTEGER M,N,KLIM,K,ITER,IFAULT
      INTEGER IC1(M), IC2(M), NC(K), NCP(K), ITRAN(K), LIVE(K)
      REAL A(M,N), D(M), C(KLIM,N), AN1(K), AN2(K), U
 
      REAL DT1, DT2, ZERO, ONE
      INTEGER I,J,L,INDX,IJ,II,LL
      REAL BIG, DA, TEMP, DB, DC, AA

C     Define BIG to be a very large positive number
      DATA BIG /1.E30/, ZERO /0.0/, ONE /1.0/

*     Validate K, return error code 3 for K < 2 or K > M-1
      IFAULT = 3
      IF (K .LE. 1 .OR. K .GE. M) RETURN
      IFAULT = 0

*-----------------------------------------------------------------------
*     Initial values - note that C[] is assumed to have initial values
*       already assigned!
C     For each point I, find its two closest centres, IC1(I) and
C     IC2(I).     Assign it to IC1(I).

      DO 50 I = 1, M
        IC1(I) = 1
        IC2(I) = 2

        DT1 = ZERO    ! square distance from centroids 1 and 2 to point
        DT2 = ZERO
        DO 10 J = 1, N
          DA = A(I,J) - C(1,J)
          DB = A(I,J) - C(2,J)
          DT1 = DT1 + DA*DA
          DT2 = DT2 + DB*DB
   10   CONTINUE
        IF (DT1 > DT2) THEN  ! swap the square distances if appropriate
          IC1(I) = 2
          IC2(I) = 1
          TEMP = DT1
          DT1 = DT2
          DT2 = TEMP
        END IF
        DO 50 L = 3, K         ! compare the point to the centroids 3...K
          DB = ZERO
          DO 30 J = 1, N
            DC = A(I,J) - C(L,J)
            DB = DB + DC*DC
            IF (DB .GE. DT2) GO TO 50       
   30     CONTINUE
          IF (DB < DT1) GO TO 40         
          DT2 = DB                        ! better point than IC2
          IC2(I) = L
          GO TO 50
   40     DT2 = DT1                     ! better point than IC1
          IC2(I) = IC1(I)
          DT1 = DB
          IC1(I) = L
   50 CONTINUE

*-----------------------------------------------------------------------
C     Update cluster centres to be the average of points contained
C     within them.

      DO 70 L = 1, K            ! set previous values to zero
        NC(L) = 0
        DO 60 J = 1, N
   60   C(L,J) = ZERO
   70 CONTINUE

      DO 90 I = 1, M
        L = IC1(I)                   ! index of the cluster nearest point I
        NC(L) = NC(L) + 1            ! increment count of point in cluster
   90 CONTINUE

*------------------------------------------------------------------------
C     Check to see if there is any empty cluster at this stage
*          remedial transfer phase -- added 5/30/2011 AJA
*------------------------------------------------------------------------
      DO 100 L = 1, K
        IF (NC(L) .EQ. 0) THEN
          DO I = 1, M
            LL = IC1(I)
            IF (NC(LL) > 1) THEN
              IC1(I) = L
              IC2(I) = LL
              NC(L) = 1
              NC(LL) = NC(LL) - 1
              GO TO 100                   ! next L
            END IF
          END DO ! next point

          IFAULT = 1                     ! if remedial transfer fails
          RETURN
        END IF
 100  CONTINUE
*------------------------------------------end of remedial tranfer

      DO I = 1, M                 ! add
        L = IC1(I)
        DO J = 1, N
          C(L,J) = C(L,J) + A(I,J)
        END DO
      END DO

      DO 120 L = 1, K
        AA = NC(L) 
        DO 110 J = 1, N
  110   C(L,J) = C(L,J) / AA         !    divide by the number of points

C     Initialize AN1, AN2, ITRAN & NCP
C     AN1(L) = NC(L) / (NC(L) - 1)
C     AN2(L) = NC(L) / (NC(L) + 1)
C     ITRAN(L) = 1 if cluster L is updated in the quick-transfer stage,
C              = 0 otherwise
C     In the optimal-transfer stage, NCP(L) stores the step at which
C     cluster L is last updated.
C     In the quick-transfer stage, NCP(L) stores the step at which
C     cluster L is last updated plus M.

        AN2(L) = AA / (AA + ONE)
        AN1(L) = BIG
        IF (AA > ONE) AN1(L) = AA / (AA - ONE)
        ITRAN(L) = 1
        NCP(L) = -1
  120 CONTINUE
      INDX = 0

*-----------------------------------------------------------------------
*       begin main loop
*-----------------------------------------------------------------------
      DO 140 IJ = 1, ITER

C     In this stage, there is only one pass through the data.   Each
C     point is re-allocated, if necessary, to the cluster that will
C     induce the maximum reduction in within-cluster sum of squares.

        CALL OPTRA(A, M, N, C, KLIM, K, IC1, IC2, NC, AN1, AN2, 
     &    NCP, D, ITRAN, LIVE, INDX)

C     Stop if no transfer took place in the last M optimal transfer
C     steps.

        IF (INDX .EQ. M) GO TO 150

C     Each point is tested in turn to see if it should be re-allocated
C     to the cluster to which it is most likely to be transferred,
C     IC2(I), from its present cluster, IC1(I).   Loop through the
C     data until no further change is to take place.

        CALL QTRAN(A, M, N, C, KLIM, K, IC1, IC2, NC, AN1, AN2, 
     &    NCP, D, ITRAN, INDX)

C     If there are only two clusters, there is no need to re-enter the
C     optimal transfer stage.

        IF (K .EQ. 2) GO TO 150

C     NCP has to be set to 0 before entering OPTRA.

        DO 130 L = 1, K
  130   NCP(L) = 0
  140 CONTINUE   ! end of main loop

C     Since the specified number of iterations has been exceeded, set
C     IFAULT = 2.   This may indicate unforeseen looping.

      IFAULT = 2

*-----------------------------------------------------------------------
*     Compute final centers 
*-----------------------------------------------------------------------
  150 DO 160 L = 1, K
        DO 160 J = 1, N
          C(L,J) = ZERO
  160 CONTINUE
      DO 170 I = 1, M
        II = IC1(I)
        DO 170 J = 1, N
          C(II,J) = C(II,J) + A(I,J)
  170 CONTINUE
      DO 180 J = 1, N
        DO 180 L = 1, K
  180   C(L,J) = C(L,J) / FLOAT(NC(L))
  
*----------------------------------------------------------------------
*    compute overall error (was array WSS[])
*----------------------------------------------------------------------
      U = 0.0
      DO I = 1, M
        II = IC1(I)
        DO J = 1, N
          DA = A(I,J) - C(II,J)
          U = U + DA*DA
        END DO
      END DO
          
      RETURN
      END

*--------------------------------------------------------------------------
      SUBROUTINE OPTRA(A, M, N, C, KLIM, K, IC1, IC2, NC, AN1, 
     & AN2, NCP, D, ITRAN, LIVE, INDX)

C     ALGORITHM AS 136.1  APPL. STATIST. (1979) VOL.28, NO.1

C     This is the optimal transfer stage.
C     Each point is re-allocated, if necessary, to the cluster that
C     will induce a maximum reduction in the within-cluster sum of
C     squares.
*---------------------------------------------------------------------------
      IMPLICIT NONE
      INTEGER M,N,KLIM,K,INDX
      INTEGER IC1(M), IC2(M), NC(K), NCP(K), ITRAN(K), LIVE(K)
      REAL    A(M,N), D(M), C(KLIM,N), AN1(K), AN2(K)

      INTEGER L,I,L1,L2,LL,J
      REAL ZERO, ONE
      REAL BIG,DE,DF,DA,DB,R2,RR,DC,DD,AL1,ALW,AL2,ALT

C     Define BIG to be a very large positive number.

      DATA BIG /1.0E30/, ZERO /0.0/, ONE/1.0/

C     If cluster L is updated in the last quick-transfer stage, it
C     belongs to the live set throughout this stage.   Otherwise, at
C     each step, it is not in the live set if it has not been updated
C     in the last M optimal transfer steps.

      DO 10 L = 1, K
        IF (ITRAN(L) .EQ. 1) LIVE(L) = M + 1
   10 CONTINUE
      DO 100 I = 1, M
        INDX = INDX + 1
        L1 = IC1(I)                       ! index of nearest cluster
        L2 = IC2(I)                  ! index of second-nearest cluster
        LL = L2

C     If point I is the only member of cluster L1, no transfer.

        IF (NC(L1) .EQ. 1) GO TO 90

C     If L1 has not yet been updated in this stage, no need to
C     re-compute D(I).

        IF (NCP(L1) .EQ. 0) GO TO 30
        DE = ZERO
        DO 20 J = 1, N
          DF = A(I,J) - C(L1,J)
          DE = DE + DF*DF
   20   CONTINUE
        D(I) = DE * AN1(L1)      ! weighted sum of squares from point I to L1

C     Find the cluster with minimum R2.

   30   DA = ZERO
        DO 40 J = 1, N
          DB = A(I,J) - C(L2,J)
          DA = DA + DB*DB
   40   CONTINUE
        R2 = DA * AN2(L2)       ! weighted sum of squares from point I to L2
        DO 60 L = 1, K

C     If I >= LIVE(L1), then L1 is not in the live set.   If this is
C     true, we only need to consider clusters that are in the live set
C     for possible transfer of point I.   Otherwise, we need to consider
C     all possible clusters.

          IF (I .GE. LIVE(L1) .AND. I .GE. LIVE(L) .OR. L .EQ. L1 .OR.
     &        L .EQ. LL) GO TO 60
          RR = R2 / AN2(L)
          DC = ZERO
          DO 50 J = 1, N
            DD = A(I,J) - C(L,J)
            DC = DC + DD*DD
            IF (DC .GE. RR) GO TO 60
   50     CONTINUE
          R2 = DC * AN2(L)             !   a new second-nearest cluster
          L2 = L
   60     CONTINUE
          IF (R2 < D(I)) GO TO 70

C     If no transfer is necessary, L2 is the new IC2(I).

          IC2(I) = L2
          GO TO 90

C     Update cluster centres, LIVE, NCP, AN1 & AN2 for clusters L1 and
C     L2, and update IC1(I) & IC2(I).

   70     INDX = 0
          LIVE(L1) = M + I
          LIVE(L2) = M + I
          NCP(L1) = I
          NCP(L2) = I
          AL1 = NC(L1)
          ALW = AL1 - ONE
          AL2 = NC(L2)
          ALT = AL2 + ONE
          DO 80 J = 1, N
            C(L1,J) = (C(L1,J) * AL1 - A(I,J)) / ALW  ! change center of L1
            C(L2,J) = (C(L2,J) * AL2 + A(I,J)) / ALT  ! change center of L2
   80     CONTINUE
          NC(L1) = NC(L1) - 1    ! change number of points in L1 and L2
          NC(L2) = NC(L2) + 1
          AN2(L1) = ALW / AL1     !  change AN1 and AN2
          AN1(L1) = BIG          
          IF (ALW > ONE) AN1(L1) = ALW / (ALW - ONE)
          AN1(L2) = ALT / AL2
          AN2(L2) = ALT / (ALT + ONE)
          IC1(I) = L2             ! update the nearest clusters to point I
          IC2(I) = L1
   90   CONTINUE
        IF (INDX .EQ. M) RETURN
  100 CONTINUE
      DO 110 L = 1, K

C     ITRAN(L) = 0 before entering QTRAN.   Also, LIVE(L) has to be
C     decreased by M before re-entering OPTRA.

        ITRAN(L) = 0
        LIVE(L) = LIVE(L) - M
  110 CONTINUE

      RETURN
      END

*-------------------------------------------------------------------------
      SUBROUTINE QTRAN(A, M, N, C, KLIM, K, IC1, IC2, NC, AN1, 
     & AN2, NCP, D, ITRAN, INDX)
C
C     ALGORITHM AS 136.2  APPL. STATIST. (1979) VOL.28, NO.1
C
C     This is the quick transfer stage.
C     IC1(I) is the cluster which point I belongs to.
C     IC2(I) is the cluster which point I is most likely to be
C         transferred to.
C     For each point I, IC1(I) & IC2(I) are switched, if necessary, to
C     reduce within-cluster sum of squares.  The cluster centres are
C     updated after each step.
*-------------------------------------------------------------------------
      IMPLICIT NONE
      INTEGER M,N,KLIM,K,INDX
      INTEGER IC1(M), IC2(M), NC(K), NCP(K), ITRAN(K)
      REAL A(M,N), D(M), C(KLIM,N), AN1(K), AN2(K)

      REAL ZERO, ONE
      INTEGER ICOUN,ISTEP,I,L1,L2,J
      REAL BIG,DA,DB,DD,AL1,ALW,AL2,ALT,R2,DE

C     Define BIG to be a very large positive number

      DATA BIG /1.0E30/, ZERO /0.0/, ONE /1.0/

C     In the optimal transfer stage, NCP(L) indicates the step at which
C     cluster L is last updated.   In the quick transfer stage, NCP(L)
C     is equal to the step at which cluster L is last updated plus M.

      ICOUN = 0
      ISTEP = 0
   10 DO 70 I = 1, M
        ICOUN = ICOUN + 1
        ISTEP = ISTEP + 1
        L1 = IC1(I)
        L2 = IC2(I)

C     If point I is the only member of cluster L1, no transfer.

        IF (NC(L1) .EQ. 1) GO TO 60

C     If ISTEP > NCP(L1), no need to re-compute distance from point I to
C     cluster L1.   Note that if cluster L1 is last updated exactly M
C     steps ago, we still need to compute the distance from point I to
C     cluster L1.

        IF (ISTEP > NCP(L1)) GO TO 30
        DA = ZERO
        DO 20 J = 1, N
          DB = A(I,J) - C(L1,J)
          DA = DA + DB*DB
   20   CONTINUE
        D(I) = DA * AN1(L1)

C     If ISTEP >= both NCP(L1) & NCP(L2) there will be no transfer of
C     point I at this step.

   30   IF (ISTEP .GE. NCP(L1) .AND. ISTEP .GE. NCP(L2)) GO TO 60
        R2 = D(I) / AN2(L2)
        DD = ZERO
        DO 40 J = 1, N
          DE = A(I,J) - C(L2,J)
          DD = DD + DE*DE
          IF (DD .GE. R2) GO TO 60
   40   CONTINUE

C     Update cluster centres, NCP, NC, ITRAN, AN1 & AN2 for clusters
C     L1 & L2.   Also update IC1(I) & IC2(I).   Note that if any
C     updating occurs in this stage, INDX is set back to 0.

        ICOUN = 0
        INDX = 0
        ITRAN(L1) = 1
        ITRAN(L2) = 1
        NCP(L1) = ISTEP + M
        NCP(L2) = ISTEP + M
        AL1 = NC(L1)
        ALW = AL1 - ONE
        AL2 = NC(L2)
        ALT = AL2 + ONE
        DO 50 J = 1, N                                  ! new centers
          C(L1,J) = (C(L1,J) * AL1 - A(I,J)) / ALW
          C(L2,J) = (C(L2,J) * AL2 + A(I,J)) / ALT
   50   CONTINUE
        NC(L1) = NC(L1) - 1
        NC(L2) = NC(L2) + 1
        AN2(L1) = ALW / AL1                            ! update AN1 and AN2
        AN1(L1) = BIG
        IF (ALW > ONE) AN1(L1) = ALW / (ALW - ONE)
        AN1(L2) = ALT / AL2
        AN2(L2) = ALT / (ALT + ONE)
        IC1(I) = L2
        IC2(I) = L1

C     If no re-allocation took place in the last M steps, return.

   60   IF (ICOUN .EQ. M) RETURN
   70 CONTINUE
      GO TO 10
      END
