*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
* rhythm.f - Subroutines for rhythm detection in the time domain
* by Andy Allinger, 2011
*
* This program is free software; you may use, modify, and redistribute it
* according to the terms of the Allegro Giftware License.  If you did not 
* receive a copy of this license, visit:
*  alleg.sourceforge.net
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
* A bank of oscillatory filters responds to a sequence of events in time.
* Each filter has the undriven form of a decaying logarithmic spiral.
* Filters respond to events by adjusting their amplitude, phase, and frequency.
* For general reference, see:
*
*   Theodosius Pavlidis
*   Biological Oscillators:  their Mathematical Analysis
*   Academic Press, 1973.
*
*   Edward W. Large and John F. Kolen
*   "Resonance and the Perception of Musical Meter"
*   Connection Science v.6 n.1 pp.177-208, 1994
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*                              I/O LIST
*   Name      Type     In/Out      Description
*   T[m]      REAL     In          Onset times of events
*   S[m]      REAL     In          Saliences of events
*   m         INTEGER  In          Number of events
*   n         INTEGER  In          Number of evalutation timepoints
*   o         INTEGER  In          Number of oscillators
*   pip       REAL     In          Step size in time
*   fdeep     REAL     In          Center frequency of deepest oscillator
*   fspac     REAL     In          Ratio of frequencies of adjacent oscillators
*   frang     REAL     In          Ratio of min to max frequency of a single
*                                  oscillator
*   A[o,0:n]  REAL     Out         Amplitude 
*   P[o,0:n]  REAL     Out         Phase
*   F[o,0:n]  REAL     Out         Frequency
*   fault     INTEGER  Out         nonzero on error
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       SUBROUTINE oscill(T,S,m,n,o,pip,fdeep,fspac,frang,A,P,F,fault)

       IMPLICIT NONE
       INTEGER m, n, o, fault
       REAL T, S, pip, fdeep, fspac, frang, A, P, F
       DIMENSION T(m), S(m), A(o,0:n), P(o,0:n), F(o,0:n)
       
*         constants
!       INTEGER difpts
       REAL eta, L, pi, rmin, rmax, twopi, v
       PARAMETER ( 
!       difpts = 10,         ! max. points for approximating nu
     &             eta = 0.3,           ! coupling strength parameter
     &             L = 3.0,             ! half-life of amplitude decay, seconds
     &             pi = 3.141592653589793,
     &             rmin = .0001,         ! distance to stay away from origin
     &             rmax = .9999,         ! closest approach to unit circle
     &             twopi  = 6.283185307179586,
     &             v = 1.0 )             

*        local variables
       INTEGER    i, j, k            ! counters
     
       REAL  ARTNQ,                    ! external arctangent function 
     &       ave,                      ! average of S
     &       b,                        ! amplitude decay rate
     &       b1, b2,                   ! amplitude decay for whole pip, interval
     &       c,                        ! salience decay rate
     &       crc(5), src(4), !erc(10), ! RAF coefficients for cos, sin, e^
     &       dt, dx, dy,               ! changes in t, x, and y
     &       EPS,                      ! very small number
     &       fmin, f0, fmax,           ! min, central, max freq. each oscillator
     &       machine,                  ! external function to find precision
     &       nu                        ! frequency, in Hertz
     
       REAL  
     &       phi,                 ! phase, measured in beats elapsed
     &       pprev,               ! phase at previous event
     &       rho, r2,             ! amplitude, amplitude squared
     &       s0, s1, ss, sbig,    ! present salience, salience accumulation
     &       sa,                  ! salience adjustment factor
!     &       sw,                  ! weight accumulation
     &       SNAP,                ! replacement for EXP
     &       theta, th2,          ! phase % 1, in radians, theta^2
     &       t0,                  ! time of prior event
     &       tnext,               ! time of next timepoint
     &       tprev,               ! time last processed event or pip
     &       tt,                  ! same as T(i)
     &       u,                   ! sharpening coefficient given rho, theta
     &       w,                   ! weight coefficient of an event 
     &       x, y                 ! x = rho @ cos(theta) ; y = rho @ sin(theta)
     
       DOUBLE PRECISION 
     &     st0, st1, st2, sp, stp, det             ! least-squares sums

       LOGICAL e               ! whether exponentiation is required
       
*  coefficients for rational algebraic function approximations of sin, cos
       DATA src / -0.1011064007,   0.9974632118, 
     &             0.003690355291, 0.06145460818 /
       DATA crc /  0.01919787449, -0.4526488730, 0.9999274175, 
     &             0.001379739651, 0.04693012443 /
       SAVE src, crc
       
       INTEGER processed, ignored

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*             begin 
       fault = 0         ! clear error codes
       EPS = machine(1)
       b = LOG(1./2) / L
       b1 = EXP(b * pip)

*       validate
       IF (m < 1 .OR. n < 1 .OR. o < 1) THEN
         fault = 1
         RETURN
       END IF

*       check that enough timepoints have been allocated for the given events
       IF (n < T(m) / pip) THEN
         fault = 2
         RETURN
       END IF
       
*         get f0, take root of frang
       frang = SQRT(frang)   ! range is now half-range
       f0 = fdeep / fspac    ! OK to multiply by fspac on iteration k = 1 

*         normalize salience
       ave = 0
       DO i = 1, m ; ave = ave + S(i) ; END DO
       ave = ave / m 
       sa = eta / ave
       DO i = 1, m ; S(i) = S(i) * sa ; END DO
       
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*         for each oscillator       
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       DO k = 1, o
         f0 = f0 * fspac          ! central frequency
         fmin = f0 / frang        ! minimum frequency
         fmax = f0 * frang        ! maximum frequency
         e = .FALSE.              ! no stimulus event intervening between pips
         c = 2 * LOG(1./2) * f0   ! longer memory for slow oscillators

         rho = rmin          ! minimal amplitude
         phi = 0             ! initial phase
         nu = f0             ! default frequency
         A(k,0) = rho ; P(k,0) = phi ; F(k,0) = f0        ! pass thru origin
         
         i = 1
         t0 = 0 ; tnext = 0 ; tprev = 0
         ss = 0 ; sbig = 0
         pprev = 0 
         st0 = 0 ; st1 = 0 ; st2 = 0 ; sp = 0 ; stp = 0
         processed = 0 ; ignored = 0

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*          for each evaluation point
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         DO j = 1, n
           tnext = tnext + pip
  10       IF (i > m) GO TO 50       ! no more events ?
           tt = T(i)
           IF (tt > tnext) GO TO 50   ! no event before tnext ?

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*   apply the stimulus - first advance r, phi to present
*    then make an adjustment in cartesian coordinates
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*                phi' = phi + nu @ dt
           dt = tt - tprev           
           phi = phi + nu * dt

*               rho' = rho @ e^(b @ dt)
           rho = rho * SNAP(b * dt)           !  approximate exponential decay 
           rho = MIN(MAX(rho, rmin), rmax)
                                                                      
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*             ignore stimuli that occur too close together
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            s1 = S(i)
            s0 = SNAP(c * (tt-t0)) * ss
            IF (s1 < s0 .AND. phi-pprev < 0.618034 .AND. s1 < sbig) THEN
              i = i + 1
              tprev = tt 
              e = .TRUE.
              ignored = ignored + 1
              GO TO 10
            ELSE
              processed = processed + 1
              ss = s0 + s1
              sbig = s1
            END IF

*             x = rho @ COS(theta) 
           theta = twopi * (phi - ANINT(phi))      !  -pi .LE. theta < pi
           th2 = theta * theta                             !  approximate cosine
           x = rho * ( (crc(1) * th2 + crc(2)) * th2 + crc(3) ) / 
     &         ( (crc(4) * th2 + crc(5)) * th2 + 1 )

*            y = rho @ SIN(theta)  
           y = rho * ( (src(1) * th2 + src(2)) * theta ) /   ! approximate sine
     &             ( (src(3) * th2 + src(4)) * th2 + 1 )
           
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*  Use activation sharpening.  Refer to:
*   J. Devin McAuley
*   Perception of Time as Phase:  Toward an Adaptive Oscillator Model of 
*   Rhythmic Pattern Processing
*   University of Indiana Ph.D. Thesis, July 1995, p. 58
*           u = ((1 + COS(theta)) / 2)^r
* a better choice might be the wrapped Cauchy distribution:
*           u = (1 - r^2) / ( 2 @ pi @ (1 + r^2 - 2 @ r @ COS(theta)) )
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
           r2 = rho * rho
           u = (1 - r2) / (twopi * (1 + r2 - 2 * x))
           
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*  Consider a chemical oscillator with reactant concentrations x and y,
* subject to a stimulus of adding a substance with y = 0, and x = v,
* the maximum concentration of x.  Then the phase transition will fulfill:
*                dy/dx = -y / (v - x)
* Refer to:
*            Arthur T. Winfree
*            The Geometry of Biological Time
*            Springer-Verlag, 1980, p. 139
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
           w = (s1 * u) / SQRT((v-x)*(v-x) + y*y)
           dx = w * (v-x) 
           dy = w * (-y) 

           x = MIN(x + dx, v)
           IF (ABS(dy) .GE. ABS(y)) THEN   ! don't overcorrect
             IF (y < 0) phi = phi + 1    ! increment phase if x-axis is reached
             y = 0
           ELSE 
             y = y + dy
           END IF

           rho = SQRT(x*x + y*y)
           IF (rho < rmin) THEN   ! don't fall on zero (the "phase singularity")
             x = rmin
             rho = rmin
           END IF
           IF (rho > rmax) rho = rmax         ! needless?
           
           theta = ARTNQ(y, x)
           phi = AINT(phi) + theta / twopi
           
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*   Measure frequency as the slope of a line of phase as a function of time
*                 nu = dphi / dt
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*           decay the running sums for the least squares line 
           b2 = SNAP(b * (tt-t0))              ! decay factor
           st0 = b2 * st0 + 1                  ! SUM t^0
           st1 = b2 * st1 + tt                 ! SUM t^1
           st2 = b2 * st2 + tt * tt            ! SUM t^2
           sp = b2 * sp + phi                  ! SUM phi
           stp = b2 * stp + tt * phi           ! SUM t @ phi
           det = st0 * st2 - st1 * st1        ! determinant of normal equations
           IF (det > EPS) nu = (st0 * stp - sp * st1) / det ! Cramer's rule
           nu = MIN(MAX(nu, fmin), fmax)   ! impose limits
           
           i = i + 1
           t0 = tt ; tprev = tt 
           pprev = phi
           e = .TRUE.
           GO TO 10
                      
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  50       CONTINUE   ! advance time to an evaluation point for A, P, F
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
           IF (e) THEN
             dt = tnext - tprev
             rho = rho * SNAP(b * dt)     !  approximate exponential function
             phi = phi + nu * dt
           ELSE 
             rho = rho * b1       ! avoid an exponentiation when possible
             phi = phi + nu * pip
           END IF
           
           A(k,j) = rho
           P(k,j) = phi
           F(k,j) = nu
           tprev = tnext
           e = .FALSE.
         END DO ! next j
         PRINT *, 'oscillator ', k, ', ',processed, ' events procesed ',
     &         ignored, ' events ignored.'
         
       END DO ! next k

!            debug dump
       PRINT *, 'amplitude'
       DO j = 1, n ; 
         DO k = 1, o; 
           PRINT 85, A(k,j) 
         END DO 
         PRINT 86	
       END DO
       PRINT *, 'phase'
       DO j = 1, n 
         do k = 1, o
           PRINT 85, P(k,j) 
         END DO
         print 86
       END DO
       PRINT *, 'frequency'
       DO J = 1, n 
         DO k = 1, o
           PRINT 85, F(k,j) 
         END DO
         PRINT 86
       END DO
  85   FORMAT (F13.5, $)
  86   FORMAT ()

      END ! of oscill
       


*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
* strata -  The best rhythic patterns are chosen from the output of the 
* oscillators with the Viterbi algorithm.  Dropped beats are filled in.
* The phase and time estimates are approximated with a smoothing a spline 
* yielding a function for phase given time, for each of several rhythmic 
* strata.  Refer to :
*  
*   Anssi P. Klapuri, Antti J. Eronen, and Jaakko T. Astola,
*   "Analysis of the Meter of Acoustic Musical Signals"
*   IEEE Transactions on Speech and Audio Processing, 2004
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*                               I/O LIST
*    Name      Type     In/Out      Description
*     T[m]      REAL     in          times of events
*     S[m]      REAL     in          saliences of events
*     L         INTEGER  in          number of levels ("strata") to make
*     LT        INTEGER  in          index of the tactus level, the first 
*                                     level made.  Strata with higher indices
*                                     have higher frequencies, and lower 
*                                     indices have lower frequencies.
*                                     Aside from this, the strata are in 
*                                     no particular order and may even cross.
*     m         INTEGER  in          number of events
*     n         INTEGER  in          number of evaluation timepoints ("pips")
*     q         INTEGER  in          number of spline interpolation points
*     pip       REAL     in          duration between oscillator evaluations
*     prob      REAL     out         total log probability 
*     CX[q+1]   REAL     out         spline coefficients
*     CY[q,L]   REAL     out          "      "
*     CA[q,L]   REAL     out          "      "
*     CB[q,L]   REAL     out          "      "
*     CC[q,L]   REAL     out          "      "
*     fault     INTEGER  out          error code
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      SUBROUTINE strata(T,S,L,LT,m,n,q,pip,prob,CX,CY,CA,CB,CC,fault)
      
       IMPLICIT NONE
       INTEGER L, LT, m, n, q, fault
       REAL T, S, pip, prob, CX, CY, CA, CB, CC
       DIMENSION T(m), S(m), CX(q+1), CY(q,L), CA(q,L), CB(q,L), CC(q,L)
      
*       local variables
       INTEGER o ; PARAMETER (o = 25)    ! number of oscillators
       INTEGER gn ; PARAMETER (gn = 1024)! evaluation points of spline
       INTEGER B(o,n), back,             ! back pointers
     &         C(0:n,2),                 ! best paths thru the Viterbi lattice
     &         h, hh, i, j, k, k1,       ! counters
     &         h0,                       ! 1: tactus level,  2: other level
     &         jnew, jold,               ! even or odd time step
     &         kmin, kmax, imin, imax    ! which oscillators to inspect
     
       REAL  
     &  A(o,0:n),              ! oscillator amplitude
     &  BIG,                   ! maximum number 
     &  ax, axb, axb2, fx, fxg, fxg2,   ! temporary storage
     &  crc(10), lcrc(10),     ! rational function coefficients
     &  CP(4),                 ! cubic polynomial coefficients
     &  D(0:n),                ! frequency of tactus level
     &  dp, dt,                ! change of phase, change of time
     &  e,                     ! phase error
     &  F(o,0:n),              ! oscillator output frequency
     &  fdeep, frang, fspac,   ! deepest natural frequency, range, spacing
     &  g,                     ! the Klapuri g function
     &  GX(gn), GY(gn), GA(gn), GB(gn), GC(gn),  ! g function parameters
     &  machine,               ! external function
     &  mu, nu,                ! average frequency, present frequency
     &  sigma2                 ! parameter for frequency distribution
     
      REAL 
     &  P(o,0:n),              ! oscillator phase
     &  pobs, ppre,            ! observed, predicted phase
     &  r,                     ! ratio of frequencies
     &  rmin,                  ! minimum amplitude
     &  tcum,                  ! cumulative time
     &  tf,                    ! time of final evaluation point
     &  TLN,                   ! Thrifty (Tightwad?) Logarithm Natural
     &  U(o,0:1),              ! cumulative log probabilities at even/odd steps
     &  ubest, utry,           ! log-prob. of likliest, other transitions
     &  W(0:n),                ! weights for least-squares spline
     &  X(0:n),                ! time input for spline
     &  Y(0:n),                ! phase input for spline
     &  Z(7*q+25)              ! work array for spline
     
       LOGICAL gready

      SAVE crc, lcrc, gready, GX, GY, GA, GB, GC
            
*///////////////////////////////////////////////////////////////////////
*  Quartic rational algebraic functions are used to replace function calls.
*  Given:  y = c0 + (c1 @ x) + (c2 @ x^2) + (c3 @ x^3) + (c4 @ x^4)
*  Let:  P = c4^(1/4)
*        Q = (c3 - P^3) / (4 @ P^3)
*        S = 3 @ Q^2 + 8 @ Q^3 + (c1 @ P - 2 @ c2 @ Q) / P^2
*        R = c2 / P^2 - (2 @ Q) - (6 @ Q^2) - S
*        T = c0 - Q^4 - (Q^2) @ (R + S) - (R @ S)
*  Thus:  y = ((P@x + Q)^2 + P@x + R)((P@x + Q)^2 + S) + T
*
*    attributed to Donald Knuth, published in:
*///////////////////////////////////////////////////////////////////////
* William H. Press, Brian P. Flannery, Saul A. Teukolsky, & Willam T. Vetterling
* Numerical Recipes in C:  the Art of Scientific Computing
* Cambridge University Press, 1988
*///////////////////////////////////////////////////////////////////////
*       approximation for log(cauchy(x)), sigma = 0.1 :
!       DATA crc / 4.651702062, 2.281753929, -123.0997148, 91.50121973,
!     & 11401.18150, 2.576522240, 2.649704134, -133.8329729, 100.1248745,
!     & 13588.37839 /
*  multiply the result by negative one.

*       approximation for log(cauchy(x)), sigma = 0.05 :
       DATA crc / 6.755572583, 3.201928019, -271.5329082, 215.6823040,
     &  59032.34381, 3.650193722, 3.279782120, -238.4334814, 
     &  187.4005801, 45116.81973 /
*  multiply the result by negative one.

*      approximation for log(log-Cauchy(x)), sigma = 0.2 :
!       DATA lcrc  / 2.502303089, 2.154313078, -556.9274635, 459.544565,
!     &   256370.3084, 2.454823733, -1.053187319, -11.32164174, 
!     &   2.799875171, 40.92122467 /
*  multiply the result by negative one.

*      approximation for log(log-Cauchy(x)), sigma = 0.083 :
       DATA lcrc / 3.027997741, 0.7103774745, -209.7279030, 156.3308326,
     &  32822.24571, 2.317047485, -0.9651216132, -10.06844877,
     &  0.6655135270, 15.59152713 /
*  multiply the result by negative one.

!*      approximation for log(log-Cauchy(x)), sigma = 0.3 :
!       DATA tcrc / 2.520268401, 1.968629105, -490.9258943, 401.2660553,
!     &  197330.4777, 2.696092463, -1.153445168, -13.41561413, 
!     &  6.308499916, 93.31790018 /
!*  multiply the result by negative one.

      DATA gready / .FALSE. / ! calculate the g function on first run
      
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*      begin
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*         define constants
       BIG = machine(3)
       fdeep = 0.1              ! slowest frequency
       frang = 4.0 / 3.0        ! range of each oscillator, following Large
       fspac = 2. ** (1./3.)    ! spacing between oscillators 
       rmin = 0.0001
       IF (.NOT. gready) THEN
         CALL gfunc(GX, GY, GA, GB, GC, gn)
         gready = .TRUE.
       END IF

*          compare number of strata against number of oscillators
       IF (L > o) THEN
         PRINT *, 'Too many rhythmic strata for oscillators allocated'
         fault = 1
         RETURN
       END IF
        
*           set up evaluation points
       tf = (n + 0.5) * pip      ! last evaluation point is beyond range of data
       dt = tf / q
       tcum = 0
       PRINT *, 'using tf=', tf
       DO j = 1, q+1
         CX(j) = tcum
         tcum = tcum + dt
       END DO
       
*          run the data thru the filterbank
       CALL oscill(T,S,m,n,o,pip,fdeep,fspac,frang,A,P,F,fault)
!       PRINT *, A
!       PAUSE
!       PRINT *, P
!       PAUSE
!       PRINT *, F
!       PAUSE
       
       prob = 0
       DO hh = 0, L-1          ! for each stratum
         h = MOD(LT+hh-1, L) +1     ! do the tactus level first
         
*|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
*            Viterbi algorithm:
*    The basic equation for conditional probability is:
*                     P(A&B) = P(A|B) @ P(B)
*   Suppose each event e in a sequence of n events is conditional only on 
* the event before it.  This yields the equation for a Markov chain:
*     P(E) = P(e[1]) @ PRODUCT P(e[i+1]|e[i])              where i = 1...n-1
*  where P(E) is the probability of the entire sequence.
*  Suppose also that knowledge of each event is supplemented by an observation
*  at each step.  Applying Bayes's rule:
*               P(E&O) = P(E|O) @ P(O) = P(O|E) @ P(E)  
*  and take the observation as certain, P(O) = 1, then
*                 P(E|O) = P(O|E) @ P(E)
* Bayesian reasoning is less elegant when the quantities involved must
* be measured.  For the present problem, transition probability P(e[i+1]|e[i])
* is based on the assumption that frequency remains steady and therefore phase
* advances steadily.  The observation likelihood, P(O|E), is taken to be 
* proportional to the oscillator amplitude, and to be higher if a frequency
* is proportional to the tactus (first detected, (strongest?)) frequency.
* The Viterbi algorithm asks at each step in the sequence, for each possible 
* event, what event was most likely at the prior step?  The answers for each
* step are kept, and at the end of the sequence the most likely chain of
* events may be traced back.
*
*|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
*          choose parameters for initial frequency distribution - from Klapuri
         IF (h .EQ. LT) THEN           ! tactus
           mu = 1. / 0.55
           sigma2 = 0.28 **2
         ELSE IF (h .EQ. LT-1) THEN    ! tatum
           mu = 1. / 0.18
           sigma2 = 0.39 **2
         ELSE IF (h .EQ. LT+1) THEN    ! measure
           mu = 1. / 2.1
           sigma2 = 0.26 **2
         ELSE                          ! uninformative distribution
           mu = 2.0                    ! arbitrary typical frequency
           sigma2 = 999.               ! arbitrary large number
         END IF
         
*|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
* initial distribution:  Log-Cauchy distribution on frequency
!lognormal distribution on frequency
!*         f = (1/(SQRT(2@pi)) ) @ e^( (LOG(mu/nu))^2 / (2 @ sigma^2))
!*  But since the log probability is being accumulated anyway, there's 
!*  no reason to exponentiate.  The constant factor 1 / SQRT(2@pi) may be 
!*  likewise disregarded.
*|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
         DO k = 1, o
!           utry = LOG(mu / F(k,0)) / sigma2
!           U(k,0) = -utry * utry / 2
           r = mu / F(k,0)
           IF (r > 1.) r = 1. / r
           utry = TLN(r)
           U(k,0) = TLN(sigma2 / (sigma2 + utry * utry))
         END DO
         
*         inspect only nearby oscillators
*         remember that higher metrical levels are actually lower frequencies
         IF (h .EQ. LT) THEN       ! don't allow tactus to pinch off all paths
           imin = h
           imax = o - L + h       ! prevents crashes, output can still be junk
         ELSE IF (h < LT) THEN     ! tatum ... high frequency
           imin = C(0,1) + 1
           imax = o
         ELSE                      ! strophe ... low frequency
           imin = 1
           imax = C(0,1) - 1
         END IF
         
*          forwards phase, for each time step
         DO j = 1, n
           IF (h .EQ. LT) THEN
             kmin = h
             kmax = o - L + h
           ELSE IF (h < LT) THEN
             kmin = C(j,1) + 1
             kmax = o
           ELSE 
             kmin = 1
             kmax = C(j,1) - 1
           END IF
           
*           alternate storage in array U[o,0:1] on even and odd j
           IF (BTEST(j, 0)) THEN
             jnew = 1  
             jold = 0
           ELSE
             jnew = 0
             jold = 1             
           END IF
           
*           for each oscillator at present time step
           DO k = kmin, kmax
             IF (A(k,j) < rmin) THEN      ! this cell has been marked 
!               PRINT *, 'zero encountered at i=', i, ' j=', j, ' k=', k
               U(k,jnew) = -BIG
               CYCLE
             END IF
!            PRINT *, 'A(', k,',',j,') =', A(k,j), ' .GE. rmin=', rmin
           
             ubest = -BIG
             nu = F(k,j)
             ppre = P(k,j) - nu * pip    ! predicted value for prior phase
             
*            for each oscillator at previous time step
             DO i = imin, imax
               utry = U(i,jold)              ! prior probability               
               IF (utry < ubest) CYCLE          ! bail out
               
*                 penalize change in frequency 
               r = nu / F(i,j-1)    
               IF (r > 1) r = 1 / r       ! restrict to 0 < r .LE. 1
               
*           approximate the log of the log-Cauchy probability distribution
*           as a function of ratio of old and new frequencies
               ax = lcrc(1) * r
               axb = ax + lcrc(2)
               axb2 = axb * axb
               fx = lcrc(6) * r
               fxg = fx + lcrc(7)
               fxg2 = fxg * fxg
               utry = utry -                  ! transition probability #1
     &    ( (axb2 + ax + lcrc(3)) * (axb2 + lcrc(4)) + lcrc(5)  ) 
     &  / ( (fxg2 + fx + lcrc(8)) * (fxg2 + lcrc(9)) + lcrc(10) )
               IF (utry < ubest) CYCLE      ! bail out 
               
*                penalize phase mismatch
               dp = ppre - P(i,j-1)
               e = ABS(dp - ANINT(dp))
               
*            approximate the log of the Cauchy probability distribution
               ax = crc(1) * e
               axb = ax + crc(2)
               axb2 = axb * axb
               fx = crc(6) * e
               fxg = fx + crc(7)
               fxg2 = fxg * fxg
               utry = utry -                ! transition probability #2
     &     ( (axb2 + ax + crc(3)) * (axb2 + crc(4)) + crc(5)  )
     &   / ( (fxg2 + fx + crc(8)) * (fxg2 + crc(9)) + crc(10) )

               IF (utry > ubest) THEN 
                 ubest = utry
                 back = i
               END IF
             END DO  ! next i
             
*              save pointer to most likely prior state
             B(k,j) = back
               
*               observation likelihood
*              is frequency a Farey relation to tactus?
             IF (h .NE. LT) THEN    
               r = D(j) / nu
               IF (r > 1.) r = 1. / r           ! evaluate g function
               CALL SCOMP(GX, GY, GA, GB, GC, gn, r, g, 1, fault)
               IF (fault .NE. 0) PRINT *, 'scomp: error#', fault
               ubest = ubest + TLN(g)                     ! likelihood #1
             END IF
               
*               oscillator amplitude 
             ubest = ubest + TLN(A(k,j))                  ! likelihood #2

*              is frequency within expected range?
             r = mu / nu
             IF (r > 1.) r = 1. / r
             utry = TLN(r)                                ! likelihood #3
             ubest = ubest + TLN(1. / (1. + utry * utry))

             U(k,jnew) = ubest
           END DO  ! next k

           imin = kmin
           imax = kmax
         END DO ! next j
         
*         find the highest scoring path
         ubest = -BIG
         IF (BTEST(n, 0)) THEN  ! is data in even or odd area?
           jnew = 1
         ELSE
           jnew = 0
         END IF
         DO k = kmin, kmax
           IF (U(k,jnew) > ubest) THEN
             back = k
             ubest = U(k,jnew)
           END IF
         END DO
         
*          cumulative probability 
         prob = prob + ubest
         
*          trace back pointers
!        DEBUG:
        PRINT *, 'B[]=', B

         IF (h .EQ. LT) THEN          
           h0 = 1                   ! one data set for tactus level 
         ELSE
           h0 = 2                   ! and one that gets overwritten
         END IF
         DO j = n, 1, -1
!           PRINT *, 'back is', back
           C(j,h0) = back
           back = B(back,j)
         END DO
         C(0,h0) = back                   ! End of Viterbi         
         PRINT *, 'C[]=', C         
!         PAUSE
         
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*           replace dropped beats
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         W(0) = A(back,0) ; X(0) = 0 ; Y(0) = P(back,0)
         tcum = 0 
         DO j = 1, n
           k = C(j-1,h0)   ! index of previous
           k1 = C(j,h0)    ! index of new
           ppre = P(k,j-1) + pip * (F(k,j-1) + F(k1,j)) / 2  ! midpoint integral
           pobs = P(k1,j)
           pobs = pobs + ANINT(ppre - pobs)   ! adjust by an integer 
           P(k1,j) = pobs
           
           W(j) = A(k1,j)
           A(k1,j) = 0.        ! prevent this cell from being used again
!           PRINT *, 'zeroing out cell A(', k1, ',', j, ')'
           tcum = tcum + pip
           X(j) = tcum
           Y(j) = pobs
         END DO  ! next j
         
!              more debugging         
         PRINT *, 'X=', X
         PRINT *, 'Y=', Y
         
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*          fit a smoothing spline
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         IF (n < 7) THEN
           CALL cfit(W, X, Y, n+1, CP, fault)
           IF (fault .NE. 0) PRINT *, 'cfit: error#', fault
           CY(1,h) = CP(1)
           CA(1,h) = CP(2)
           CB(1,h) = CP(3)
           CC(1,h) = CP(4)
         ELSE
           CALL SPFIT(X, Y, W, n+1, CX, q+1, 
     &       CY(1,h), CA(1,h), CB(1,h), CC(1,h), Z, fault)
           IF (fault .NE. 0) PRINT *, 'spfit: error #', fault
         END IF
         
!   DEBUG:        dump spline output
!       DO j = 0, 1000
!         CALL SCOMP(CX, CY(1,h), CA(1,h), CB(1,h), CC(1,h), 
!     &     q, (tcum * FLOAT(j)/1000.), Z(1), 1, fault)
!         IF (fault .ne. 0) PRINT *, 'scomp:  error #', fault
!         PRINT *, 'At time ', tcum*FLOAT(j)/1000., ' phase=', Z(1)
!       END DO
         
         
*          smooth estimate of tactus frequency
         IF (h .EQ. LT) THEN 
           CALL SEVAL(CX, CY(1,h), CA(1,h), CB(1,h), CC(1,h), q,
     &                    X, Y, D, W, n+1, fault)   ! re-use some variables
           IF (fault .NE. 0) PRINT *, 'seval: error#', fault
         END IF
       END DO ! next hh
       
      END  ! of strata
           

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*  dlsp - get the derivative of an unevenly spaced time series,
*          using the least-squares parabola
*            W[n]    DOUBLE PRECISION          importance weight of data points
*            X[n]    DOUBLE PRECISION          independent variable
*            Y[n]    DOUBLE PRECISION          dependent variable
*            n       INTEGER                   # points in approximation
*            h       INTEGER                   index to evaluate slope at
*            dydx    DOUBLE PRECISION          output
*            fault   INTEGER                   error code
*     it is assumed that arrays X and Y are ring buffers;
*   the newest point is X[h], and the oldest is X[h+1]
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       SUBROUTINE dlsp(W, X, Y, n, h, dydx, fault)
!              
!       IMPLICIT NONE
!       INTEGER n, h, fault
!       DOUBLE PRECISION W(n), X(n), Y(n), dydx
!       
!       INTEGER i, k
!       DOUBLE PRECISION  a1, a2, det, sub1, sub2, sub3,  ! sol'n to normal eqn's
!     &                   dmachine, EPS,             ! very small number
!     &                   Z(n),                      ! X rescaled to 0...zf
!     &                   zf                         ! small number (5?)
!       DOUBLE PRECISION xo, w1, x1, y1, z1
!       DOUBLE PRECISION zy, z2, dzdx
!       DOUBLE PRECISION sw,sx,sy,sz, sx2, sxy, sz2, sz3, sz4, szy, sz2y
!       DATA  zf / 5.0 /        ! adjust as needed to make matrix tractable
! 
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*           begin        
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       fault = 0
!       EPS = 3 * dmachine(1)
!       IF (h .EQ. n) THEN         ! find index of oldest point
!         k = 1  
!       ELSE
!         k = h + 1
!       END IF
!          
!*          small n
!       IF (n < 2) THEN 
!         PRINT *, 'dlsp:  error - not enough data'
!         fault = 1
!         RETURN
!       ELSE IF (n .EQ. 2) THEN
!         IF (ABS(X(1) - X(2)) < EPS) THEN
!           PRINT *, 'dlsp: division by zero'
!           fault= 2
!         ELSE
!           dydx = (Y(h) - Y(k)) / (X(h) - X(k))
!         END IF
!         RETURN
!       END IF
!
*       rescale  z = zf @ (x - xo) / (xf - xo)
!       xo = X(k) 
!       dzdx = zf / (X(h) - xo)      ! dz/dx 
!       DO i = 1, n
!         Z(i) = (X(i) - xo) * dzdx
!       END DO
!
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*      Normal equations for a parabola.  Refer to any statistics text, eg
*   Murray Spiegel, Schaum's Outline of Statistics, 1988 [1961]
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       sw = 0
!       sz = 0
!       sy = 0 
!       sz2 = 0
!       sz3 = 0
!       sz4 = 0
!       szy = 0
!       sz2y = 0
!       DO i = 1, n
!         w1 = W(i)
!         z1 = Z(i)
!         y1 = Y(i)
!         sw = sw + w1                   ! SUM w
!         sz = sz + z1 * w1              ! SUM z
!         sy = sy + y1 * w1              ! SUM y
!         z2 = z1 * z1               
!         sz2 = sz2 + z2 * w1            ! SUM z^2
!         sz3 = sz3 + z2 * z1 * w1       ! SUM z^3
!         sz4 = sz4 + z2 * z2 * w1       ! SUM z^4
!         zy = z1 * y1
!         szy = szy + zy * w1            ! SUM z @ y
!         sz2y = sz2y + zy * z1 * w1     ! SUM z^2 @ y
!       END DO
!       
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!*          solve the system by Cramer's rule
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       sub1 = sz * sz4 - sz2 * sz3
!       sub2 = sz * sz3 - sz2 * sz2
!       sub3 = sz * sz2y - sz2 * szy
!       det = sw*(sz2*sz4 - sz3*sz3) -sz*sub1 + sz2*sub2
!       IF (det < EPS) GO TO 90
!       a1 = ( sw*(szy*sz4 - sz2y*sz3) - sy*sub1 + sz2*sub3 ) / det
!       a2 = ( sw*(sz2*sz2y - sz3*szy) - sz*sub3 + sy*sub2 ) / det
!       
!*        evaluate the derivative dy/dz at zf
!       dydx = a1 + 2 * a2 * zf                    ! dy/dz
!       dydx = dydx * dzdx                    ! dy/dx = dy/dz @ dz/dx
!      RETURN  ! success!
!
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  90   CONTINUE    ! backup plan:  least-squares line
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!!       PRINT *, 'dlsp error#', fault
!       sx = 0
!       sx2 = 0
!       sxy = 0
!       DO i = 1, n
!         w1 = W(i)
!         x1 = X(i)
!         sw = sw + w1
!         sx = sx + x1 * w1
!         sx2 = sx2 + x1 * x1 * w1
!         sxy = sxy + x1 * Y(i) * w1
!       END DO
!       dydx = (sw * sxy - sx * sy) / (sw * sx2 - sx * sx)
!       fault = -1
!
!      END  ! of dlsp
!
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!*  slsl - Slope of the Least-Squares Line
!*            W[n]    DOUBLE PRECISION          importance weight of data points
!*            X[n]    DOUBLE PRECISION          independent variable
!*            Y[n]    DOUBLE PRECISION          dependent variable
!*            n       INTEGER                   # points in approximation
!*            h       INTEGER                   index to evaluate slope at
!*            dydx    DOUBLE PRECISION          output
!*            fault   INTEGER                   error code
!*     it is assumed that arrays X and Y are ring buffers;
!*   the newest point is X[h], and the oldest is X[h+1]
!*   Z[] is statically allocated.  If you want n > 12, edit the declaration.
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       SUBROUTINE slsl(W, X, Y, n, h, dydx, fault)
!              
!       IMPLICIT NONE
!       INTEGER n, h, fault
!       DOUBLE PRECISION W(n), X(n), Y(n), dydx
!       
!       INTEGER i, k
!       DOUBLE PRECISION  dmachine, EPS,             ! very small number
!     &                   Z(12),                     ! X rescaled to 0...zf
!     &                   zf                         ! small number (5?)
!       DOUBLE PRECISION xo, w1, y1, z1, det, dzdx
!       DOUBLE PRECISION sw, sy, sz, sz2, szy
!       DATA  zf / 5.0 /        ! adjust as needed to make matrix tractable
! 
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!*           begin        
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       fault = 0
!       EPS = n * dmachine(1)
!       IF (h .EQ. n) THEN         ! find index of oldest point
!         k = 1  
!       ELSE
!         k = h + 1
!       END IF
!          
!*          small n
!       IF (n < 2) THEN 
!         PRINT *, 'dlsp:  error - not enough data'
!         fault = 1
!         RETURN
!       ELSE IF (n .EQ. 2) THEN
!         IF (ABS(X(1) - X(2)) < EPS) THEN
!!           PRINT *, 'slsl: division by zero'
!           fault= -1
!         ELSE
!           dydx = (Y(h) - Y(k)) / (X(h) - X(k))
!         END IF
!         RETURN
!       END IF!
!
!*       rescale  z = zf @ (x - xo) / (xf - xo)
!       xo = X(k) 
!       dzdx = zf / (X(h) - xo)      ! dz/dx 
!       DO i = 1, n
!         Z(i) = (X(i) - xo) * dzdx
!       END DO!
!
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!*      Normal equations for a line.  Refer to any statistics text, eg
!*   Murray Spiegel, Schaum's Outline of Statistics, 1988 [1961]
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       sw = 0
!       sy = 0 
!       sz = 0
!       sz2 = 0
!       szy = 0
!       DO i = 1, n
!         w1 = W(i)
!         z1 = Z(i)
!         y1 = Y(i)
!         sw = sw + w1                   ! SUM w
!         sz = sz + z1 * w1              ! SUM z
!         sy = sy + y1 * w1              ! SUM y
!         sz2 = sz2 + z1 * z1 * w1       ! SUM z^2
!         szy = szy + z1 * y1 * w1       ! SUM z @ y
!       END DO
!       
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!*          solve the system by Cramer's rule
!*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!       det = sw * sz2 - sz * sz
!       IF (det < EPS) THEN 
!         fault = 3
!         RETURN
!       END IF
!       dydx = (sw * szy - sy * sz) / det         ! dy/dz
!       dydx = dydx * dzdx                        ! dy/dx = dy/dz @ dz/dx
!      RETURN  ! success!
!
!      END  ! of slsl


*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*   a four-quadrant arctangent with range 0...2pi
*     by Richard Pasto
*     part of the Naval Surface Warfare Center library
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      FUNCTION ARTNQ(Y,X)
       IF (X) 1, 2, 5
  1    ARTNQ = ATAN(Y/X) + 3.1415926535898
       RETURN
  2    IF (Y) 3, 8, 4
  3    ARTNQ = 4.7123889803847
       RETURN
  4    ARTNQ = 1.5707963267949
       RETURN
  5    IF (Y) 6, 8, 7
  6    ARTNQ = ATAN(Y/X) + 6.2831853071795
       RETURN
  7    ARTNQ = ATAN(Y/X)
       RETURN
  8    ARTNQ = 0.
       RETURN
      END     ! of ARTNQ


************************************************************************
* interpolate the Klapuri g function
************************************************************************
      SUBROUTINE gfunc(X, Y, A, B, C, n)
      
       IMPLICIT NONE
       INTEGER i, j, n, ibeg, iend, fault
       REAL X(n), Y(n), A(n), B(n), C(n), sr2pi
       REAL farey(23), width(23), height(23), wf, hf
       REAL alpha, beta

************************************************************************
*  These constants represent the strength at which an oscillator latches on
*  to a signal, depending on the ratio of the signal frequency to the 
*  oscillator frequency.  The array farey is made up of fractions from the
*  Stern-Brocott-Farey sequence.  Width and height refer to an "empirical regime
*  diagram" by Large, in which width represents a tolerance around an ideal
*  integer ratio, and height is the a parameter that controls how strongly
*  the oscillator entrains to its input.  These variables can be reinterpreted
*  as intensity (height), and standard deviation (width) around a mean (ratio).
*  Refer to:
*     Edward Wilson Large
*     Dynamic Representation of Musical Structure
*     Ph.D. Thesis, Ohio State University, 1994, Fig. 39
************************************************************************      
       DATA farey /  0.0, .125, .1428571429, 0.1666666667, 0.2, 0.25, 
     &   0.2857142858, 0.3333333333, 0.375, 0.400, 0.4285714286, 0.5, 
     &   0.5714285714, 0.6, 0.625, 0.6666666667, 0.7142857143, 0.75, 
     &   0.8, 0.8333333333, 0.8571428571, 0.875, 1.000 /

       DATA width / 18, 2, 3, 4, 7, 10, 3, 12, 3, 7, 3, 19, 4, 7, 2, 13,
     &   3, 10, 7, 5, 3, 2, 21 /

       DATA height / 5, 2, 2, 2.2, 2.2, 3.3, 1.9, 3.3, 1.7, 2, 1.7, 4, 
     &   1.6, 1.8, 1.6, 3.5, 1.4, 3, 1.6, 1.5, 1.2, 1.2, 5 /
     
*       scale factors
       DATA wf, hf / 720, 3.16 /
       DATA alpha, beta / 0, 0 /       ! not used
       
************************************************************************
* evaluate the function - sum of bell curves with different means and deviations
************************************************************************
       sr2pi = SQRT(2 * ACOS(-1.0))
       DO j = 1, 23
         width(j) = width(j) / wf
         height(j) = height(j) / hf
       END DO
       
       DO i = 1, n
         X(i) = FLOAT(i-1) / (n-1)
         Y(i) = 0
         DO j = 1, 23
           Y(i) = Y(i) + (height(j)**2 / sr2pi) * 
     &        EXP( - (X(i)-farey(j))**2 / (2 * width(j))**2 )
         END DO
       END DO
         
*-----------------------------------------------------------------------
*     call the spline interpolator
*-----------------------------------------------------------------------
       ibeg = 0 
       iend = 0 
       fault = 0
       CALL CBSPL(X, Y, A, B, C, N, ibeg, iend, alpha, beta, fault)
       IF (fault .NE. 0) PRINT *, 'cbspl: error #', fault
               
      END ! of gfunc


*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
*        quartic rational algebraic function approximation of e^x
*   the approximation is excellent (parts per million) for -1 < x < 0
*   and acceptable                 ( ~= 5 per cent )   for -5 < x < 0
*   Outside this domain it is disasterously wrong
*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      FUNCTION SNAP(x)      ! swift Napier
      
       IMPLICIT NONE
       REAL erc(10), x, ax, axb, axb2, fx, fxg, fxg2, numer, denom, SNAP
       SAVE erc
       
       DATA erc /  0.1469443670,      0.5582320252,  1.066831328,
     &  0.5123596415,     -0.1358224253,  0.1662192096, -1.000364885,
     & -0.9095052017E-01,  0.2920193779, -0.1761166417 /

       ax = erc(1) * x
       axb = ax + erc(2)
       axb2 = axb * axb
       numer = (axb2 + ax + erc(3)) * (axb2 + erc(4)) + erc(5)

       fx = erc(6) * x
       fxg = fx + erc(7)
       fxg2 = fxg * fxg
       denom = (fxg2 + fx + erc(8)) * (fxg2 + erc(9)) + erc(10) 
       
       SNAP = numer / denom
      END  ! of SNAP


*#######################################################################
*       quartic rational algebraic function approximation of LOG(x)
*   the approximation is acceptable (within ~= 2 per cent) from .0001 < x < 1
*#######################################################################
      FUNCTION TLN(x)  

       IMPLICIT NONE
       REAL lrc(10), x, ax, axb, axb2, fx, fxg, fxg2, numer, denom, TLN
       SAVE lrc

       DATA lrc /38.32465688,  -7.070035057, 8163.528155,  -8864.222070,
     &           72395842.41,  32.20366109,  11.49453140,  -9105.213437,
     &           8469.641002,  77184407.78 /
      
       ax = lrc(1) * x
       axb = ax + lrc(2)
       axb2 = axb * axb
       numer = (axb2 + ax + lrc(3)) * (axb2 + lrc(4)) + lrc(5)
  
       fx = lrc(6) * x
       fxg = fx + lrc(7)
       fxg2 = fxg * fxg
       denom = (fxg2 + fx + lrc(8)) * (fxg2 + lrc(9)) + lrc(10)

       TLN = numer / denom
      END  ! of TLN
       
*++++++++++++++++++++++++ End of file rhythm.f +++++++++++++++++++++++++
