************************************************************
*
*                                               ROBERT RENKA
*                                       OAK RIDGE NATL. LAB.
*
*   THIS SUBROUTINE USES AN ORDER N*LOG(N) QUICK SORT TO SORT AN INTEGER 
* ARRAY X INTO INCREASING ORDER.  THE ALGORITHM IS AS FOLLOWS.  IND IS 
* INITIALIZED TO THE ORDERED SEQUENCE OF INDICES 1,...,N, AND ALL 
* INTERCHANGES ARE APPLIED TO IND.  X IS DIVIDED INTO TWO PORTIONS BY 
* PICKING A CENTRAL ELEMENT T.  THE FIRST AND LAST ELEMENTS ARE COMPARED 
* WITH T, AND INTERCHANGES ARE APPLIED AS NECESSARY SO THAT THE THREE 
* VALUES ARE IN ASCENDING ORDER.  INTERCHANGES ARE THEN APPLIED SO THAT 
* ALL ELEMENTS GREATER THAN T ARE IN THE UPPER PORTION OF THE ARRAY AND 
* ALL ELEMENTS LESS THAN T ARE IN THE LOWER PORTION.  THE UPPER AND LOWER 
* INDICES OF ONE OF THE PORTIONS ARE SAVED IN LOCAL ARRAYS, AND THE 
* PROCESS IS REPEATED ITERATIVELY ON THE OTHER PORTION.  WHEN A PORTION 
* IS COMPLETELY SORTED, THE PROCESS BEGINS AGAIN BY RETRIEVING THE 
* INDICES BOUNDING ANOTHER UNSORTED PORTION.
*
* INPUT PARAMETERS -   N - LENGTH OF THE ARRAY X.
*
*                      X - VECTOR OF LENGTH N TO BE SORTED.
*
*                    IND - VECTOR OF LENGTH .GE. N.
*
* N AND X ARE NOT ALTERED BY THIS ROUTINE.
*
* OUTPUT PARAMETER - IND - SEQUENCE OF INDICES 1,...,N
*                          PERMUTED IN THE SAME FASHION AS X
*                          WOULD BE.  THUS, THE ORDERING ON
*                          X IS DEFINED BY Y(I) = X(IND(I)).
*
* INTRINSIC FUNCTIONS CALLED BY QSORTI - IFIX, FLOAT
*
************************************************************
*
* NOTE -- IU AND IL MUST BE DIMENSIONED .GE. LOG(N) WHERE
*         LOG HAS BASE 2.
*
************************************************************
      SUBROUTINE QSORTI (X, IND, N)
      INTEGER N, X(N), IND(N)

      INTEGER IU(21), IL(21)
      INTEGER M, I, J, K, L, IJ, IT, ITT, INDX, T
      REAL    R

* LOCAL PARAMETERS -
*
* IU,IL =  TEMPORARY STORAGE FOR THE UPPER AND LOWER
*            INDICES OF PORTIONS OF THE ARRAY X
* M =      INDEX FOR IU AND IL
* I,J =    LOWER AND UPPER INDICES OF A PORTION OF X
* K,L =    INDICES IN THE RANGE I,...,J
* IJ =     RANDOMLY CHOSEN INDEX BETWEEN I AND J
* IT,ITT = TEMPORARY STORAGE FOR INTERCHANGES IN IND
* INDX =   TEMPORARY INDEX FOR X
* R =      PSEUDO RANDOM NUMBER FOR GENERATING IJ
* T =      CENTRAL ELEMENT OF X

      IF (N .LE. 0) RETURN

* INITIALIZE IND, M, I, J, AND R
      DO 1 I = 1,N
  1     IND(I) = I
      M = 1
      I = 1
      J = N
      R = .375

* TOP OF LOOP
  2   IF (I .GE. J) GO TO 10
      IF (R > .5898437) GO TO 3
      R = R + .0390625
      GO TO 4
  3   R = R - .21875

* INITIALIZE K
  4   K = I

* SELECT A CENTRAL ELEMENT OF X AND SAVE IT IN T
      IJ = I + IFIX(R*FLOAT(J-I))
      IT = IND(IJ)
      T = X(IT)

* IF THE FIRST ELEMENT OF THE ARRAY IS GREATER THAN T,
*   INTERCHANGE IT WITH T
      INDX = IND(I)
      IF (X(INDX) .LE. T) GO TO 5
      IND(IJ) = INDX
      IND(I) = IT
      IT = INDX
      T = X(IT)

* INITIALIZE L
  5   L = J

* IF THE LAST ELEMENT OF THE ARRAY IS LESS THAN T,
*   INTERCHANGE IT WITH T
      INDX = IND(J)
      IF (X(INDX) .GE. T) GO TO 7
      IND(IJ) = INDX
      IND(J) = IT
      IT = INDX
      T = X(IT)

* IF THE FIRST ELEMENT OF THE ARRAY IS GREATER THAN T,
*   INTERCHANGE IT WITH T
      INDX = IND(I)
      IF (X(INDX) .LE. T) GO TO 7
      IND(IJ) = INDX
      IND(I) = IT
      IT = INDX
      T = X(IT)
      GO TO 7

* INTERCHANGE ELEMENTS K AND L
  6   ITT = IND(L)
      IND(L) = IND(K)
      IND(K) = ITT

* FIND AN ELEMENT IN THE UPPER PART OF THE ARRAY WHICH IS
*   NOT LARGER THAN T
  7   L = L - 1
      INDX = IND(L)
      IF (X(INDX) > T) GO TO 7

* FIND AN ELEMENT IN THE LOWER PART OF THE ARRAY WHICH IS
*   NOT SMALLER THAN T
  8   K = K + 1
      INDX = IND(K)
      IF (X(INDX) < T) GO TO 8

* IF K .LE. L, INTERCHANGE ELEMENTS K AND L
      IF (K .LE. L) GO TO 6

* SAVE THE UPPER AND LOWER SUBSCRIPTS OF THE PORTION OF THE
*   ARRAY YET TO BE SORTED
      IF (L-I .LE. J-K) GO TO 9
      IL(M) = I
      IU(M) = L
      I = K
      M = M + 1
      GO TO 11

  9   IL(M) = K
      IU(M) = J
      J = L
      M = M + 1
      GO TO 11

* BEGIN AGAIN ON ANOTHER UNSORTED PORTION OF THE ARRAY
 10   M = M - 1
      IF (M .EQ. 0) RETURN
      I = IL(M)
      J = IU(M)

 11   IF (J-I .GE. 11) GO TO 4
      IF (I .EQ. 1) GO TO 2
      I = I - 1

* SORT ELEMENTS I+1,...,J.  NOTE THAT 1 .LE. I < J AND J-I < 11.
 12   I = I + 1
      IF (I .EQ. J) GO TO 10
      INDX = IND(I+1)
      T = X(INDX)
      IT = INDX
      INDX = IND(I)
      IF (X(INDX) .LE. T) GO TO 12
      K = I

   13 IND(K+1) = IND(K)
      K = K - 1
      INDX = IND(K)
      IF (T < X(INDX)) GO TO 13
      IND(K+1) = IT
      GO TO 12
      END ! of QSORTI

************************************************************
*                                               ROBERT RENKA
*                                       OAK RIDGE NATL. LAB.
*
*   THIS SUBROUTINE USES AN ORDER N*LOG(N) QUICK SORT TO SORT THE REAL 
* ARRAY X INTO INCREASING ORDER.  THE ALGORITHM IS AS FOLLOWS.  IND IS 
* INITIALIZED TO THE ORDERED SEQUENCE OF INDICES 1,...,N, AND ALL 
* INTERCHANGES ARE APPLIED TO IND.  X IS DIVIDED INTO TWO PORTIONS BY 
* PICKING A CENTRAL ELEMENT T.  THE FIRST AND LAST ELEMENTS ARE COMPARED 
* WITH T, AND INTERCHANGES ARE APPLIED AS NECESSARY SO THAT THE THREE 
* VALUES ARE IN ASCENDING ORDER.  INTERCHANGES ARE THEN APPLIED SO THAT 
* ALL ELEMENTS GREATER THAN T ARE IN THE UPPER PORTION OF THE ARRAY AND 
* ALL ELEMENTS LESS THAN T ARE IN THE LOWER PORTION.  THE UPPER AND LOWER 
* INDICES OF ONE OF THE PORTIONS ARE SAVED IN LOCAL ARRAYS, AND THE 
* PROCESS IS REPEATED ITERATIVELY ON THE OTHER PORTION.  WHEN A PORTION 
* IS COMPLETELY SORTED, THE PROCESS BEGINS AGAIN BY RETRIEVING THE 
* INDICES BOUNDING ANOTHER UNSORTED PORTION.
*
* INPUT PARAMETERS -   N - LENGTH OF THE ARRAY X.
*
*                      X - VECTOR OF LENGTH N TO BE SORTED.
*
*                    IND - VECTOR OF LENGTH .GE. N.
*
* N AND X ARE NOT ALTERED BY THIS ROUTINE.
*
* OUTPUT PARAMETER - IND - SEQUENCE OF INDICES 1,...,N
*                          PERMUTED IN THE SAME FASHION AS X
*                          WOULD BE.  THUS, THE ORDERING ON
*                          X IS DEFINED BY Y(I) = X(IND(I)).
*
* INTRINSIC FUNCTIONS CALLED BY QSORTR - IFIX, FLOAT
*
************************************************************
*
* NOTE -- IU AND IL MUST BE DIMENSIONED .GE. LOG(N) WHERE
*         LOG HAS BASE 2.
*
************************************************************
      SUBROUTINE QSORTR (X, IND, N)
      INTEGER N, IND(N)
      REAL    X(N)

      INTEGER IU(21), IL(21)
      INTEGER M, I, J, K, L, IJ, IT, ITT, INDX
      REAL    R, T

* LOCAL PARAMETERS -
*
* IU,IL =  TEMPORARY STORAGE FOR THE UPPER AND LOWER
*            INDICES OF PORTIONS OF THE ARRAY X
* M =      INDEX FOR IU AND IL
* I,J =    LOWER AND UPPER INDICES OF A PORTION OF X
* K,L =    INDICES IN THE RANGE I,...,J
* IJ =     RANDOMLY CHOSEN INDEX BETWEEN I AND J
* IT,ITT = TEMPORARY STORAGE FOR INTERCHANGES IN IND
* INDX =   TEMPORARY INDEX FOR X
* R =      PSEUDO RANDOM NUMBER FOR GENERATING IJ
* T =      CENTRAL ELEMENT OF X

      IF (N .LE. 0) RETURN

* INITIALIZE IND, M, I, J, AND R
      DO 1 I = 1, N
  1     IND(I) = I
      M = 1
      I = 1
      J = N
      R = .375

* TOP OF LOOP
  2   IF (I .GE. J) GO TO 10
      IF (R > .5898437) GO TO 3
      R = R + .0390625
      GO TO 4
  3   R = R - .21875

* INITIALIZE K
  4   K = I

* SELECT A CENTRAL ELEMENT OF X AND SAVE IT IN T
      IJ = I + IFIX(R*FLOAT(J-I))
      IT = IND(IJ)
      T = X(IT)

* IF THE FIRST ELEMENT OF THE ARRAY IS GREATER THAN T, INTERCHANGE IT WITH T
      INDX = IND(I)
      IF (X(INDX) .LE. T) GO TO 5
      IND(IJ) = INDX
      IND(I) = IT
      IT = INDX
      T = X(IT)

* INITIALIZE L
  5   L = J

* IF THE LAST ELEMENT OF THE ARRAY IS LESS THAN T, INTERCHANGE IT WITH T
      INDX = IND(J)
      IF (X(INDX) .GE. T) GO TO 7
      IND(IJ) = INDX
      IND(J) = IT
      IT = INDX
      T = X(IT)

* IF THE FIRST ELEMENT OF THE ARRAY IS GREATER THAN T, INTERCHANGE IT WITH T
      INDX = IND(I)
      IF (X(INDX) .LE. T) GO TO 7
      IND(IJ) = INDX
      IND(I) = IT
      IT = INDX
      T = X(IT)
      GO TO 7

* INTERCHANGE ELEMENTS K AND L
  6   ITT = IND(L)
      IND(L) = IND(K)
      IND(K) = ITT

* FIND AN ELEMENT IN THE UPPER PART OF THE ARRAY WHICH IS NOT LARGER THAN T
  7   L = L - 1
      INDX = IND(L)
      IF (X(INDX) > T) GO TO 7

* FIND AN ELEMENT IN THE LOWER PART OF THE ARRAY WHCIH IS NOT SMALLER THAN T
  8   K = K + 1
      INDX = IND(K)
      IF (X(INDX) < T) GO TO 8

* IF K .LE. L, INTERCHANGE ELEMENTS K AND L
      IF (K .LE. L) GO TO 6

* SAVE THE UPPER AND LOWER SUBSCRIPTS OF THE PORTION OF THE
*   ARRAY YET TO BE SORTED
      IF (L-I .LE. J-K) GO TO 9
      IL(M) = I
      IU(M) = L
      I = K
      M = M + 1
      GO TO 11

  9   IL(M) = K
      IU(M) = J
      J = L
      M = M + 1
      GO TO 11

* BEGIN AGAIN ON ANOTHER UNSORTED PORTION OF THE ARRAY
  10  M = M - 1
      IF (M .EQ. 0) RETURN
      I = IL(M)
      J = IU(M)

  11  IF (J-I .GE. 11) GO TO 4
      IF (I .EQ. 1) GO TO 2
      I = I - 1

* SORT ELEMENTS I+1,...,J.  NOTE THAT 1 .LE. I < J AND J-I < 11.
  12  I = I + 1
      IF (I .EQ. J) GO TO 10
      INDX = IND(I+1)
      T = X(INDX)
      IT = INDX
      INDX = IND(I)
      IF (X(INDX) .LE. T) GO TO 12
      K = I

  13  IND(K+1) = IND(K)
      K = K - 1
      INDX = IND(K)
      IF (T < X(INDX)) GO TO 13
      IND(K+1) = IT
      GO TO 12
      END ! of QSORTR


************************************************************
*
*                                               ROBERT RENKA
*                                       OAK RIDGE NATL. LAB.
*
*   THIS ROUTINE APPLIES A SET OF PERMUTATIONS TO A VECTOR.
*
* INPUT PARAMETERS -  N - LENGTH OF A AND IP.
*
*                    IP - VECTOR CONTAINING THE SEQUENCE OF
*                        INTEGERS 1,...,N PERMUTED IN THE
*                         SAME FASHION THAT A IS TO BE PER-
*                         MUTED.
*
*                     A - VECTOR TO BE PERMUTED.
*
* N AND IP ARE NOT ALTERED BY THIS ROUTINE.
*
* OUTPUT PARAMETER -  A - REORDERED VECTOR REFLECTING THE
*                         PERMUTATIONS DEFINED BY IP.
*
************************************************************
      SUBROUTINE IORDER (A, IP, N)
      INTEGER N, A(N), IP(N)

      INTEGER NN, K, J, IPJ, TEMP

* LOCAL PARAMETERS -
*
* NN =   LOCAL COPY OF N
* K =    INDEX FOR IP AND FOR THE FIRST ELEMENT OF A[] IN A PERMUTATION
* J =    INDEX FOR IP AND A, J .GE. K
* IPJ =  IP[J]
* TEMP = TEMPORARY STORAGE FOR A[K]
      NN = N
      IF (NN < 2) RETURN
      K = 1

* LOOP ON PERMUTATIONS
  1   J = K
      TEMP = A(K)

* APPLY PERMUTATION TO A.  IP[J] IS MARKED (MADE NEGATIVE)
*   AS BEING INCLUDED IN THE PERMUTATION.

  2   IPJ = IP(J)
      IP(J) = -IPJ
      IF (IPJ .EQ. K) GO TO 3
      A(J) = A(IPJ)
      J = IPJ
      GO TO 2
  3   A(J) = TEMP

* SEARCH FOR AN UNMARKED ELEMENT OF IP
  4   K = K + 1
      IF (K > NN) GO TO 5
      IF (IP(K) > 0) GO TO 1
      GO TO 4

* ALL PERMUTATIONS HAVE BEEN APPLIED.  UNMARK IP.
  5   DO 6 K = 1,NN
  6     IP(K) = -IP(K)
      RETURN
      END ! of IORDER


************************************************************
*
*                                               ROBERT RENKA
*                                       OAK RIDGE NATL. LAB.
*
*   THIS ROUTINE APPLIES A SET OF PERMUTATIONS TO A VECTOR.
*
* INPUT PARAMETERS -  N - LENGTH OF A AND IP.
*
*                   IP - VECTOR CONTAINING THE SEQUENCE OF
*                         INTEGERS 1,...,N PERMUTED IN THE
*                         SAME FASHION THAT A IS TO BE PER-
*                         MUTED.
*
*                     A - VECTOR TO BE PERMUTED.
*
* N AND IP ARE NOT ALTERED BY THIS ROUTINE.
*
* OUTPUT PARAMETER -  A - REORDERED VECTOR REFLECTING THE
*                         PERMUTATIONS DEFINED BY IP.
*
************************************************************
      SUBROUTINE RORDER (A, IP, N)
      INTEGER N, IP(N)
      REAL    A(N)

      INTEGER NN, K, J, IPJ
      REAL    TEMP

* LOCAL PARAMETERS -
*
* NN =   LOCAL COPY OF N
* K =    INDEX FOR IP AND FOR THE FIRST ELEMENT OF A IN A
*          PERMUTATION
* J =    INDEX FOR IP AND A, J .GE. K
* IPJ =  IP[J]
* TEMP = TEMPORARY STORAGE FOR A[K]

      NN = N
      IF (NN < 2) RETURN
      K = 1

* LOOP ON PERMUTATIONS
  1   J = K
      TEMP = A(K)

* APPLY PERMUTATION TO A.  IP[J] IS MARKED (MADE NEGATIVE)
*   AS BEING INCLUDED IN THE PERMUTATION.
  2   IPJ = IP(J)
      IP(J) = -IPJ
      IF (IPJ .EQ. K) GO TO 3
      A(J) = A(IPJ)
      J = IPJ
      GO TO 2
    3 A(J) = TEMP

* SEARCH FOR AN UNMARKED ELEMENT OF IP
  4   K = K + 1
      IF (K > NN) GO TO 5
      IF (IP(K) > 0) GO TO 1
      GO TO 4

* ALL PERMUTATIONS HAVE BEEN APPLIED.  UNMARK IP.
  5   DO 6 K = 1,NN
  6     IP(K) = -IP(K)
      RETURN
      END ! of RORDER


*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
* rsort - radix sort
* (C) Andy Allinger, 2011
* Free according to the terms of the Allegro Giftware license
* or, at your option, the Gnu General Public License (GPL)
*
* D [] is an integer array of N items
* B the number of significant bits, so the range 0...255 requires 8 bits
*   if B is positive, then the integers will be interpreted as UNSIGNED
*   note that FORTRAN does not support unsigned integers, so make sure
*   you know what you are doing if you use B=32
*   if B is negative, the sign bit will be inspected as a final step
*   So, for the range -256...255 set B = -8
*
* origin:  Herman Hollerith, 1887
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      SUBROUTINE rsort(D, N, B)
      
      IMPLICIT NONE
      INTEGER D, N, B
      DIMENSION D(N)

      INTEGER W ; DIMENSION W(N)          ! declare work space
      INTEGER I, J, K, ILIM    !  counters
      INTEGER P1OLD, P0OLD, P1, P0, S
      LOGICAL INW                         ! data is in W

*          validate
      IF (B .EQ. 0) STOP 'Bit size must not be zero'
      IF (ABS(B) > BIT_SIZE(N)) STOP 'Bit size to large'
*      IF (N < 1) STOP 'Array must not be empty' ! not really necessary

*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*      sort starts on the least significant bit, position 0
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      P1 = N+1 ; P0 = 0
      DO J = 1, N

*      test for negative values
*      comment out the following line for a modest speedup 
*        (esp. for small positive integers)
        IF (D(J) < 0 .AND. B > 0 .AND. B .NE. BIT_SIZE(N)) 
     &      STOP 'Set argument B negative to sort negative numbers'

        IF ( BTEST(D(J), 0) ) THEN
          P1 = P1 - 1
          W(P1) = D(J)
        ELSE 
          P0 = P0 + 1
          W(P0) = D(J)
        END IF
      END DO
      INW = .TRUE.

*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*      now alternate between putting data into D and putting data into W
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      IF (B < 0) ILIM = MIN( ABS(B)-1, BIT_SIZE(I)-2 )
      IF (B > 0) ILIM = B - 1
      DO I = 1, ILIM
        P1OLD = P1 ; P0OLD = P0         ! save the value from previous bit
        P1 = N+1 ; P0 = 0                 ! start a fresh count for next bit

*         odd I
        IF (INW) THEN

*          copy data from the zeros
          DO J = 1, P0OLD, 1
            IF ( BTEST(W(J), I) ) THEN
              P1 = P1 - 1
              D(P1) = W(J)
            ELSE
              P0 = P0 + 1
              D(P0) = W(J)
            END IF
          END DO
          
*          copy data from the ones
          DO J = N, P1OLD, -1           
            IF ( BTEST(W(J), I) ) THEN
              P1 = P1 - 1
              D(P1) = W(J)
            ELSE
              P0 = P0 + 1
              D(P0) = W(J)
            END IF
          END DO
          
*        even I
        ELSE 

*          copy data from the zeros
          DO J = 1, P0OLD, 1
            IF ( BTEST(D(J), I) ) THEN
              P1 = P1 - 1
              W(P1) = D(J)
            ELSE
              P0 = P0 + 1
              W(P0) = D(J)
            END IF
          END DO

*            copy data from the ones
          DO J = N, P1OLD, -1
            IF ( BTEST(D(J), I) ) THEN
              P1 = P1 - 1
              W(P1) = D(J)
            ELSE
              P0 = P0 + 1
              W(P0) = D(J)
            END IF
          END DO
        END IF ! even or odd i
        
        INW = .NOT. INW
      END DO   ! next i

*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*        the sign bit
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      IF (B < 0) THEN
        P1OLD = P1 ; P0OLD = P0
        P1 = N+1 ; P0 = 0 

*          if sign bit is set, send to the zero end of the work array
        IF (INW) THEN
          DO J = 1, P0OLD, 1
            IF ( BTEST(W(J), BIT_SIZE(I)-1) ) THEN 
              P0 = P0 + 1
              D(P0) = W(J)
            ELSE
              P1 = P1 - 1
              D(P1) = W(J)
            END IF
          END DO          
          DO J = N, P1OLD, -1
            IF ( BTEST(W(J), BIT_SIZE(I)-1) ) THEN
              P0 = P0 + 1
              D(P0) = W(J)
            ELSE
              P1 = P1 - 1
              D(P1) = W(J)
            END IF
          END DO
          
        ELSE
          DO J = 1, P0OLD, 1
            IF ( BTEST(D(J), BIT_SIZE(I)-1) ) THEN 
              P0 = P0 + 1
              W(P0) = D(J)
            ELSE
              P1 = P1 - 1
              W(P1) = D(J)
            END IF
          END DO          
          DO J = N, P1OLD, -1
            IF ( BTEST(D(J), BIT_SIZE(I)-1) ) THEN
              P0 = P0 + 1
              W(P0) = D(J)
            ELSE
              P1 = P1 - 1
              W(P1) = D(J)
            END IF
          END DO
        END IF

        INW = .NOT. INW  
      END IF
          
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*      put the data back into the input array
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!      PRINT *, 'P0=', P0, ' P1=', P1
      K = P1
      IF (INW) THEN     
        DO J = 1, P0, 1
          D(J) = W(J)
        END DO
        DO J = N, P1, -1        
          D(K) = W(J)
          K = K + 1
        END DO
       
*      Swap values in the ones portion
      ELSE 
        DO J = N, (N+P1+2)/2, -1
          S = D(J) ; D(J) = D(K) ; D(K) = S
          K = K + 1
        END DO
      END IF

      END ! of rsort


*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*                               Giftware
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
* riid - radix sort and carry along an integer and a double precision array
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      SUBROUTINE riid(D, D2, D3, N, B)
      
       IMPLICIT NONE
       INTEGER D, D2, N, B
       DOUBLE PRECISION D3
       DIMENSION D(N), D2(N), D3(N)

       INTEGER W(N), W2(N)                     ! declare work space
       DOUBLE PRECISION W3(N), S3
       INTEGER I, J, K        !  counters
       INTEGER P1OLD, P0OLD, P1, P0, S
       LOGICAL INW                         ! data is in W

*          validate
       IF (B .EQ. 0) STOP 'Bit size must not be zero'
       IF (ABS(B) > BIT_SIZE(N)) STOP 'Bit size to large'

*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*      sort starts on the least significant bit, position 0
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
       P1 = N+1 ; P0 = 0
       DO J = 1, N

*      test for negative values
*      comment out the following line for a modest speedup 
*        (esp. for small positive integers)
         IF (D(J) < 0 .AND. B .NE. BIT_SIZE(N)) STOP 'negative value'

         IF ( BTEST(D(J), 0) ) THEN
           P1 = P1 - 1
           W(P1) = D(J) ; W2(P1) = D2(J) ; W3(P1) = D3(J)
         ELSE 
           P0 = P0 + 1
           W(P0) = D(J) ; W2(P0) = D2(J) ; W3(P0) = D3(J)
         END IF
       END DO
       INW = .TRUE.

*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*      now alternate between putting data into D and putting data into W
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
       DO I = 1, B-1
         P1OLD = P1 ; P0OLD = P0         ! save the value from previous bit
         P1 = N+1 ; P0 = 0                 ! start a fresh count for next bit

*         odd I
         IF (INW) THEN

*          copy data from the zeros
           DO J = 1, P0OLD, 1
             IF ( BTEST(W(J), I) ) THEN
               P1 = P1 - 1
               D(P1) = W(J) ; D2(P1) = W2(J) ; D3(P1) = W3(J)
             ELSE
               P0 = P0 + 1
               D(P0) = W(J) ; D2(P0) = W2(J) ; D3(P0) = W3(J)
             END IF
           END DO
          
*          copy data from the ones
           DO J = N, P1OLD, -1           
             IF ( BTEST(W(J), I) ) THEN
               P1 = P1 - 1
               D(P1) = W(J) ; D2(P1) = W2(J) ; D3(P1) = W3(J)
             ELSE
               P0 = P0 + 1
               D(P0) = W(J) ; D2(P0) = W2(J) ; D3(P0) = W3(J)
             END IF
           END DO
          
*         even I
         ELSE 

*          copy data from the zeros
           DO J = 1, P0OLD, 1
             IF ( BTEST(D(J), I) ) THEN
               P1 = P1 - 1
               W(P1) = D(J) ; W2(P1) = D2(J) ; W3(P1) = D3(J)
             ELSE
               P0 = P0 + 1
               W(P0) = D(J) ; W2(P0) = D2(J) ; W3(P0) = D3(J)
             END IF
           END DO

*            copy data from the ones
           DO J = N, P1OLD, -1
             IF ( BTEST(D(J), I) ) THEN
               P1 = P1 - 1
               W(P1) = D(J) ; W2(P1) = D2(J) ; W3(P1) = D3(J)
             ELSE
               P0 = P0 + 1
               W(P0) = D(J) ; W2(P0) = D2(J) ; W3(P0) = D3(J)
             END IF
           END DO
         END IF ! even or odd i
        
         INW = .NOT. INW
       END DO   ! next i

          
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*      put the data back into the input array
*~~~~ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
       K = P1
       IF (INW) THEN     
         DO J = 1, P0, 1
           D(J) = W(J) ; D2(J) = W2(J) ; D3(J) = W3(J)
         END DO
         DO J = N, P1, -1        
           D(K) = W(J) ; D2(K) = W2(J) ; D3(K) = W3(J)
           K = K + 1
         END DO
       
*       Swap values in the ones portion
       ELSE 
         DO J = N, (N+P1+2)/2, -1
           S = D(J) ; D(J) = D(K) ; D(K) = S
           S = D2(J) ; D2(J) = D2(K) ; D2(K) = S
           S3 = D3(J) ; D3(J) = D3(K) ; D3(K) = S3
           K = K + 1
         END DO
       END IF

      END ! of riid
      
            
*-----------------------------------------------------------------------
* ISORT - sort integers
*-----------------------------------------------------------------------
*          Sort an array and optionally make the same interchanges in
*            an auxiliary array.  The array may be sorted in increasing
*            or decreasing order.  A slightly modified QUICKSORT
*            algorithm is used.  From the SLATEC library, public domain.
*
*      by   Jones, R. E., (SNLA)
*           Kahaner, D. K., (NBS)
*           Wisniewski, J. A., (SNLA)
*
*   Description of Parameters
*      IX - integer array of values to be sorted
*      N  - number of values in integer array IX to be sorted
*
*  References:  R. C. Singleton, Algorithm 347, An efficient algorithm
*                 for sorting with minimal storage, Communications of
*                 the ACM, 12, 3 (1969), pp. 185-187.
*
*  REVISION HISTORY  (YYMMDD)
*   761118  DATE WRITTEN
*   810801  Modified by David K. Kahaner.
*   890531  Changed all specific intrinsics to generic.  (WRB)
*   890831  Modified array declarations.  (WRB)
*   891009  Removed unreferenced statement labels.  (WRB)
*   891009  REVISION DATE from Version 3.2
*   891214  Prologue converted to Version 4.0 format.  (BAB)
*   900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
*   901012  Declared all variables; changed X,Y to IX,IY. (M. McClain)
*   920501  Reformatted the REFERENCES section.  (DWL, WRB)
*   920519  Clarified error messages.  (DWL)
*   920801  Declarations section rebuilt and code restructured to use
*           IF-THEN-ELSE-ENDIF.  (RWC, WRB)
*   111005  Removed call to XERMSG, simplified. (AJA)
*-----------------------------------------------------------------------
      SUBROUTINE ISORT (IX, N)

*     .. Scalar Arguments ..
      INTEGER N
*     .. Array Arguments ..
      INTEGER IX(N)
*     .. Local Scalars ..
      REAL R
      INTEGER I, IJ, J, K, L, M, NN, T, TT
*     .. Local Arrays ..
      INTEGER IL(21), IU(21)

*        Begin
      NN = N
      IF (NN < 1) THEN
         PRINT *, 'The number of values to be sorted is not positive.'
         RETURN
      END IF

*     Sort IX only
      M = 1
      I = 1
      J = NN
      R = 0.375E0

   20 IF (I .EQ. J) GO TO 60
      IF (R .LE. 0.5898437E0) THEN
         R = R+3.90625E-2
      ELSE
         R = R-0.21875E0
      ENDIF

   30 K = I

*     Select a central element of the array and save it in location T
      IJ = I + INT((J-I)*R)
      T = IX(IJ)

*     If first element of array is greater than T, interchange with T
      IF (IX(I) > T) THEN
         IX(IJ) = IX(I)
         IX(I) = T
         T = IX(IJ)
      ENDIF
      L = J

*     If last element of array is less than than T, interchange with T
      IF (IX(J) < T) THEN
         IX(IJ) = IX(J)
         IX(J) = T
         T = IX(IJ)

*        If first element of array is greater than T, interchange with T
         IF (IX(I) > T) THEN
            IX(IJ) = IX(I)
            IX(I) = T
            T = IX(IJ)
         ENDIF
      ENDIF

*     Find an element in the second half of the array which is smaller than T
   40 L = L-1
      IF (IX(L) > T) GO TO 40

*     Find an element in the first half of the array which is greater than T
   50 K = K+1
      IF (IX(K) < T) GO TO 50

*     Interchange these elements
      IF (K .LE. L) THEN
         TT = IX(L)
         IX(L) = IX(K)
         IX(K) = TT
         GO TO 40
      ENDIF

*     Save upper and lower subscripts of the array yet to be sorted
      IF (L-I > J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 70

*     Begin again on another portion of the unsorted array
   60 M = M-1
      IF (M .EQ. 0) GO TO 200
      I = IL(M)
      J = IU(M)

   70 IF (J-I .GE. 1) GO TO 30
      IF (I .EQ. 1) GO TO 20
      I = I-1

   80 I = I+1
      IF (I .EQ. J) GO TO 60
      T = IX(I+1)
      IF (IX(I) .LE. T) GO TO 80
      K = I

   90 IX(K+1) = IX(K)
      K = K-1
      IF (T < IX(K)) GO TO 90
      IX(K+1) = T
      GO TO 80

 200  END ! of ISORT


*-----------------------------------------------------------------------
* SSORT - sort single precision numbers
*-----------------------------------------------------------------------
*   Sort an array in increasing order.  A slightly modified QUICKSORT 
* algorithm is used.  From the SLATEC library, public domain. 
*
*      By  R.E. Jones, (SNLA) 
*      and J. A. Wisniewski, (SNLA)
*
*   Description of arguments
*      X[]     array of values to be sorted
*      N       number of values in array X to be sorted
*
*  REFERENCES  R. C. Singleton, Algorithm 347, An efficient algorithm
*                 for sorting with minimal storage, Communications of
*                 the ACM, 12, 3 (1969), pp. 185-187.
*
*  REVISION HISTORY  (YYMMDD)
*   761101  DATE WRITTEN
*   761118  Modified to use the Singleton quicksort algorithm.  (JAW)
*   890531  Changed all specific intrinsics to generic.  (WRB)
*   890831  Modified array declarations.  (WRB)
*   891009  Removed unreferenced statement labels.  (WRB)
*   891024  Changed category.  (WRB)
*   891024  REVISION DATE from Version 3.2
*   891214  Prologue converted to Version 4.0 format.  (BAB)
*   900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
*   901012  Declared all variables; changed X,Y to SX,SY. (M. McClain)
*   920501  Reformatted the REFERENCES section.  (DWL, WRB)
*   920519  Clarified error messages.  (DWL)
*   920801  Declarations section rebuilt and code restructured to use
*           IF-THEN-ELSE-ENDIF.  (RWC, WRB)
*   110517  removed call to XERMSG, simplified (AJA)
*-----------------------------------------------------------------------
      SUBROUTINE SSORT (X, N)

*     .. Scalar Arguments ..
      INTEGER  N
*     .. Array Arguments ..
      REAL X(N)
*     .. Local Scalars ..
      REAL R, T, TT
      INTEGER I, IJ, J, K, L, M, NN
*     .. Local Arrays ..
      INTEGER IL(21), IU(21)
*-----------------------------------------------------------------------
      NN = N
      IF (NN < 1) THEN
          PRINT *, 'The number of values to be sorted is not positive.'
         RETURN
      ENDIF

*     Sort X only
      M = 1
      I = 1
      J = NN
      R = 0.375E0

   20 IF (I .EQ. J) GO TO 60
      IF (R .LE. 0.5898437E0) THEN
         R = R+3.90625E-2
      ELSE
         R = R-0.21875E0
      ENDIF

   30 K = I

*     Select a central element of the array and save it in location T
      IJ = I + INT((J-I)*R)
      T = X(IJ)

*     If first element of array is greater than T, interchange with T
      IF (X(I) > T) THEN
         X(IJ) = X(I)
         X(I) = T
         T = X(IJ)
      ENDIF
      L = J

*     If last element of array is less than than T, interchange with T
      IF (X(J) < T) THEN
         X(IJ) = X(J)
         X(J) = T
         T = X(IJ)

*        If first element of array is greater than T, interchange with T
         IF (X(I) > T) THEN
            X(IJ) = X(I)
            X(I) = T
            T = X(IJ)
         ENDIF
      ENDIF

*     Find an element in the second half of the array which is smaller than T
   40 L = L-1
      IF (X(L) > T) GO TO 40

*     Find an element in the first half of the array which is greater than T
   50 K = K+1
      IF (X(K) < T) GO TO 50

*     Interchange these elements
      IF (K .LE. L) THEN
         TT = X(L)
         X(L) = X(K)
         X(K) = TT
         GO TO 40
      ENDIF

*     Save upper and lower subscripts of the array yet to be sorted
      IF (L-I > J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 70

*     Begin again on another portion of the unsorted array
   60 M = M-1
      IF (M .EQ. 0) GO TO 200
      I = IL(M)
      J = IU(M)

   70 IF (J-I .GE. 1) GO TO 30
      IF (I .EQ. 1) GO TO 20
      I = I-1

   80 I = I+1
      IF (I .EQ. J) GO TO 60
      T = X(I+1)
      IF (X(I) .LE. T) GO TO 80
      K = I

   90 X(K+1) = X(K)
      K = K-1
      IF (T < X(K)) GO TO 90
      X(K+1) = T
      GO TO 80

  200 END ! of SSORT
 

*----------------------------------------------------------------------- 
*  Shell sort, according to the Naval Surface Warfare Center
*----------------------------------------------------------------------- 
* The Shell sorting procedure is used to reorder the elements of A
* so that A[i] .LE. A[i+1] for i=1,....,n+1.  It is assumed that N .GE. 1
*-----------------------------------------------------------------------
      SUBROUTINE ISHELL (A, N)
      INTEGER A(N), K(16), S

      DATA K /1, 4, 13, 40, 121, 364, 1093, 3280, 9841, 29524, 88573, 
     &        265720, 797161, 2391484, 7174453, 21523360/

*             Selection of the increments K(i) = (3^i-1)
      IF (N < 2) RETURN
      IMAX = 1
      DO 10 I = 3, 16
         IF (N .LE. K(I)) GO TO 20
         IMAX = IMAX + 1
   10 CONTINUE

*            Stepping through the increments K(imax),...,K(1)
   20 I = IMAX
      DO 40 II = 1,IMAX
         KI = K(I)

*             Sorting elements that are ki positions apart
*               so that A(j) .LE. A(j+ki) for j=1,...n-ki
         JMAX = N - KI
         DO 31 J = 1, JMAX
            L = J
            LL = J + KI
            S = A(LL)
   30          IF (S .GE. A(L)) GO TO 31
               A(LL) = A(L)
               LL = L
               L = L - KI
               IF (L > 0) GO TO 30
   31    A(LL) = S
C
   40 I = I - 1
      END ! of ISHELL


      
*----------------------------------------------------------------------- 
*  Shell sort, according to the Naval Surface Warfare Center
*----------------------------------------------------------------------- 
* The Shell sorting procedure is used to reorder the elements of A
* so that A[i] .LE. A[i+1] for i=1,....,n+1.  It is assumed that N .GE. 1
*-----------------------------------------------------------------------
      SUBROUTINE SHELL (A, N)

      REAL A(N)
      INTEGER K(16)

      DATA K /1, 4, 13, 40, 121, 364, 1093, 3280, 9841, 29524, 88573, 
     &        265720, 797161, 2391484, 7174453, 21523360/

*      small N are special cases
      IF (N < 2) RETURN
      
      IF (N .EQ. 2) THEN         ! swap
        IF ( A(1) > A(2) ) THEN
          SWAP = A(1) ; A(1) = A(2) ; A(2) = SWAP
        END IF
        RETURN
      END IF      

      IF (N .EQ. 3) THEN       ! lexical sort 
        T1 = A(1) ; T2 = A(2) ; T3 = A(3)
        IF (A(1) > A(2)) THEN
          IF (A(2) > A(3)) THEN
            A(1) = T3 ; A(3) = T1  
          ELSE 
            IF (A(1) > A(3)) THEN
               A(1) = T2 ; A(2) = T3 ; A(3) = T1 
            ELSE
              A(1) = T2 ; A(2) = T1
            END IF
          END IF
        ELSE
          IF (A(2) > A(3)) THEN
            IF (A(1) > A(3)) THEN
              A(1) = T3 ; A(2) = T1 ; A(3) = T2
            ELSE
              A(2) = T3 ; A(3) = T2 
            END IF
          END IF
        END IF
        RETURN
      END IF

*             Selection of the increments K[i] = (3^i-1)/2
      IMAX = 1
      DO 10 I = 3, 16
         IF (N .LE. K(I)) GO TO 20
         IMAX = IMAX + 1
   10 CONTINUE

*            Stepping thru the increments K[imax],..., K[1]
   20 I = IMAX
      DO 40 II = 1, IMAX
         KI = K(I)

*             Sorting elements that are ki positions apart
*               so that A[j] .LE. A[j+ki] for j=1,...n-ki
         JMAX = N - KI 
         DO 31 J = 1, JMAX
            L = J
            LL = J + KI
            S = A(LL)
   30          IF (S .GE. A(L)) GO TO 31
               A(LL) = A(L)
               LL = L
               L = L - KI
               IF (L > 0) GO TO 30  
   31    A(LL) = S

   40 I = I - 1
      END ! of SHELL
 
*++++++++++++++++++++++++++ End of file sort.f +++++++++++++++++++++++++
