/* imidiacy - Record a MIDI file from the computer keyboard 
version 2007.9.06 Copyright (C) 2006-2007 Andrew Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 59 
Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program produces standard MIDI files based on presses of the PC
keyboard.  The events Note Off, Note On, Aftertouch, and tempo change
are supported.  Output is to the monitor and soundcard or PC speaker.
 
For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/

/* These defines need to be adjusted at compilation */
// #include <allegro.h>

/* Standard C */
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

/* Strange things */
//#include <sys/time.h>
#include "midi_types.h"   

/**********************************************************************/


/* The record function does the bulk of the work. it returns the number 
of bytes which will have to be written to the file (an amount which must 
go in the track header.  */

long record(long   tempo, 
            int    velocity, 
            float  legato, 
            int    channel, 
            int    key_sig, 
            int    time_sig,  
            bool   use_sound, 
            bool   harmony)

{

/* note that variables suffixed with "_time" are measured in ticks */
unsigned long 
     present_time, prior_time,
     delta_time,         /* the difference in ticks between MIDI events */
     on_time = 0,        /* time in ticks the last note started */
     elapsed_time = 0,   /* ticks gone by BEFORE the recent tempo change */
     adjust_time ,       /* brief storage for conversions */

     msec_orig,         /* what get_time returns */
     msec_new,          /* what get_time returns */
     msec_elapsed,      /* elapsed real time, milliseconds */
     msec_tempo_change = 0 ;  /* elapsed time at the last tempo change. */
     
long delta_off_time,      /* sets a note off event backwards in time
                            so as to create a separation between notes. */
     backoff_time ;       /* it may be necessary to erase some events if 
                            a Note Off is inserted retroactively before them.
                            This sums delta_time's of events, traveling 
                            backward in time, until the inserted event should 
                            be placed in proper sequence */

int event = 0 ;          /* code number for what the keystroke does */

/*
 The event variable may be decoded as follows:  
 event = 0          no keystroke detected
       = 1  to 127  begin MIDI note with corresponding note number, 
                    where N % 12 == 0 is musical pitch of C
       = -1 to -127 turn off the corresponding MIDI notes
       = 128     decrement velocity (ie - volume level)
       = 129     increment velocity 
       = 130     decrement tempo
       = 131     increment tempo
       = 132     set velocity to "piano" preset
       = 133     set velocity to "forte" preset
       = 134     play more staccto (more gap between notes)
       = 135     play more legato  (less gap between notes)
       = 136     Quit.
       = 137     stray keystrokes
       = 138     explicit rest.
*/

long count_bytes = 0 ;   /* the amount of data recorded */
bool reset_tempo = TRUE, 
     beat_shown = FALSE, 
     scrolled = FALSE,
     active_notes[128] ;       /* whether each Midi note is on */
     
int bpm,           /* beats per minute, for the user, not internal */
    note = 0,         /* Midi number of most recently pressed note */
    old_bpm, old_velocity ; 
float old_legato ;    
unsigned char var_bytes[4] ;      /* delta time in variable length format */
DBL_NODE *ptr_active = NULL,    /* tail of the linked list */
         *ptr_early, *ptr_late, *ptr_new ;
int i, j ;

/**********************************************************************/
/* misc. initializations */
bpm = MICROSECONDS_PER_MINUTE / tempo ;
for (j = 0 ; j < 128 ; ++j)           /* Don't use Midi note 0 */
  active_notes[j] = FALSE ;

/* initialize the timer */
prepare_timing() ;

/* Get the starting time */
msec_orig = get_time() ;
prior_time = 0 ;

/* main loop */
while (event != QUIT) {

/* Wait for a while, so not to burn up needless CPU power. */
  delay(DELAY_DURATION) ;    

/* Measure keystroke and time. */
  event = get_input(channel) ;
  msec_new = get_time() ;
  msec_elapsed = msec_new - msec_orig ;

  if (reset_tempo)  {
    elapsed_time = prior_time ;
    reset_tempo = FALSE ;
  }
  present_time = elapsed_time + (unsigned long) (1000 * ((float)TICKS_PER_BEAT / (float)tempo) * (msec_elapsed - msec_tempo_change)) ;

/* should getch() return no keystroke, print the metronome, update the 
screen, and keep looping */
  if (event == NO_KEYSTROKE) {
    if ( (present_time % TICKS_PER_BEAT) < TOL_METRONOME ) {
      if (!beat_shown) {
        metronome(present_time, time_sig) ;
        beat_shown = TRUE ;
      } /* End If (!beat_shown) */
    } 
    else {
      beat_shown = FALSE ;
    } /* End If time is on the beat */

/*Update the screen for the Allegro version - the scroll_roll() function 
doesn't do much in ncurses */
    if ( (msec_elapsed % SCROLL_DELAY) < TOL_SCROLL) {
      if (!scrolled) {
        scroll_roll() ;         /* this function hogs a lot of CPU */
        scrolled = TRUE ;
      } /* End If (!scrolled) */
    } 
    else {
      scrolled = FALSE ;
    } /* End If (time to scroll) */

    continue ;
  }

/* select the event and process it. */
/* first case is a plain Note Off */

  if ( (harmony && (event < 0) ) || ((event == REST) && (note > 0)) )  {

    delta_time = present_time - prior_time ;
    prior_time = present_time ;
    if (harmony && (event < 0) ) 
      note = -(event) ;

/* For an explicit NoteOff, turn off all the presently sounding notes */
    if (event == REST) {
      for (j = 1 ; j < 128 ; ++j) {
        if (!active_notes[j]) continue ;
        
/* turn off all sounding notes */
        active_notes[j] = FALSE ;
        ptr_active = add_node(ptr_active) ;
        count_bytes = count_bytes + write_var_bytes(delta_time, var_bytes) + 3 ;
        for ( i = 0 ; i < 4 ; ++i) 
          ptr_active -> delta_time[i] = var_bytes[i] ;
        ptr_active -> type = 0X8 ;
        ptr_active -> channel = channel ;
        ptr_active -> note =  j ;
        ptr_active -> velocity = 0 ;  /* this is the value used by the Yamaha PSR-410 */
        del_note(j, velocity, channel) ;
        if (use_sound) hush_note(j) ;
      } /* Next j */
    }
    else {                     /* Turn off only one note */
      active_notes[note] = FALSE ;
      ptr_active = add_node(ptr_active) ;
      count_bytes = count_bytes + write_var_bytes(delta_time, var_bytes) + 3 ;
      for (i = 0 ; i < 4 ; ++i) 
        ptr_active -> delta_time[i] = var_bytes[i] ;
      ptr_active -> type = 0X8 ;
      ptr_active -> channel = channel ;
      ptr_active -> note = note ;
      ptr_active -> velocity = 0 ;
      del_note(note, velocity, channel) ;
      if (use_sound) hush_note(note) ;
    } /* End If explicit Note off */
    note = 0 ;
  }  /* End If Note Off event */
    
/* case of a new note.  First turn off the old note, using the retroactive
legato effect, which was devised through much aggravation! */
  if ((event > 0) && (event <= 127)) {
    delta_time = present_time - prior_time ;
    prior_time = present_time ;

/* The retroactive sequence, if a note is already sounding.  Not useful 
for harmony mode. */
    if ( (note > 0) && (!harmony) ) {
      delta_off_time = delta_time - (long)(((float)(present_time - on_time)) * (1.0F - legato)) ;

/* Confusing sequence only is needed to insert a node into the list.  
Note that backoff_time runs NEGATIVE.  Step back to earlier nodes until 
the appropriate time is reached */

      if (delta_off_time <= 0) {
        backoff_time = 0 ;     
        while (backoff_time >= delta_off_time)   {
          if ( ptr_active -> ptr_prev != NULL) {
            backoff_time -= read_var_bytes(ptr_active -> delta_time) ;
            ptr_active = ptr_active -> ptr_prev ;
          }
/* bail out if the beginning of the list is reached */
          else {
            break ;
          }          
        }

/* insert a node, revising the connections */
        ptr_new = (DBL_NODE *)malloc(sizeof(DBL_NODE)) ;
        if (ptr_new == NULL) {
          fprintf(stderr, "Memory allocation error\n") ;
          exit(EXIT_FAILURE) ;               /* flag an allocation error */
        }
        ptr_early = ptr_active ;
        ptr_late = ptr_active -> ptr_next ;
        ptr_new -> ptr_prev = ptr_early ;      
        ptr_new -> ptr_next = ptr_late ;
        ptr_early -> ptr_next = ptr_new ;
        ptr_late -> ptr_prev = ptr_new ;
  
        count_bytes = count_bytes + write_var_bytes( 
                         (delta_off_time - backoff_time), var_bytes) + 3 ;    
        for (i = 0; i < 4; ++i)
          ptr_new -> delta_time[i] = var_bytes[i] ;
        ptr_new -> type = 0X8 ;
        ptr_new -> channel = channel ;
        ptr_new -> note = note ;
        ptr_new -> velocity = 0 ;

/*  Now erase events made irrelevent (ie - aftertouch, explicit note off) 
adjust the times of any remaining tempo changes */
        ptr_active = ptr_late ;

/* This routine is not for the last node in the list */
        while ( ptr_active -> ptr_next != NULL)  {

/* Disconnect Note Off and Aftertouch events */
          if ( (ptr_active -> type == 0X8) || (ptr_active -> type == 0XA) ) {
            ptr_early = ptr_active -> ptr_prev ;
            ptr_late = ptr_active -> ptr_next ;
            ptr_early -> ptr_next = ptr_late ;
            ptr_late -> ptr_prev = ptr_early ;
 
/* modify count bytes in case the delta time has gained a byte, as the 
event being erased has its delta time added to the following event */
            adjust_time = read_var_bytes(ptr_late -> delta_time) ;
            count_bytes -= count_var_bytes(adjust_time) ;
            adjust_time += read_var_bytes(ptr_active -> delta_time) ;
            count_bytes += write_var_bytes(adjust_time, var_bytes) ;
            for (i = 0 ; i < 4 ; ++i)        
              ptr_late -> delta_time[i] = var_bytes[i] ;

/* decrement count_bytes as the node is removed.  Note that the 
adjust_time variable is only for brief storage, and its meaning is 
changed in the following line */
            adjust_time = read_var_bytes(ptr_active -> delta_time) ;
            count_bytes -= count_var_bytes(adjust_time) + 3 ;
            free(ptr_active) ;
            ptr_active = ptr_late ;
          }

/* Leave intact a tempo change. */
          else if (ptr_active -> type == 0X51) {
            ptr_active = ptr_active -> ptr_next ;
          }
          else {
            fprintf(stderr, "Improper data encountered\n") ;
            exit(EXIT_FAILURE) ;         /* invalid type in event structure */
          }
        }    /* Wend the interior-node loop*/

/* disconnecting the last link is a slightly different routine */
        if ( (ptr_active -> type == 0X8) || (ptr_active -> type == 0XA) ) {
          ptr_early = ptr_active -> ptr_prev ;
          ptr_early -> ptr_next = NULL ;

          delta_time = read_var_bytes(ptr_active -> delta_time) ;
          count_bytes -= count_var_bytes(delta_time) + 3 ;
          free(ptr_active) ;
          ptr_active = ptr_early ;
        }
      }

/* for events with the legato effect, but not inserted before a later one */
      else {
        ptr_active = add_node(ptr_active) ;
        count_bytes += write_var_bytes(delta_off_time, var_bytes) + 3 ;
        for (i = 0 ; i < 4 ; ++i)
          ptr_active -> delta_time[i] = var_bytes[i] ;
        ptr_active -> type = 0X8 ;
        ptr_active -> channel = channel ;
        ptr_active -> note = note ;
        ptr_active -> velocity = 0 ;
        delta_time -= delta_off_time ;
      } /* End If retroactive events */

      del_note(note, velocity, channel) ;
      hush_note(note) ;
      active_notes[note] = FALSE ;
    } /* End If (!harmony && note > 0) */
    
/* Process a note on event.  Creation of the implicit note-off event 
(above sequence) flows directly to this step. */

/* indicate that a note has been pressed */
    active_notes[event] = TRUE ;

    ptr_active = add_node(ptr_active) ;
    count_bytes = count_bytes + write_var_bytes(delta_time, var_bytes) + 3 ;
    for (i = 0 ; i < 4 ; ++i)
      ptr_active -> delta_time[i] = var_bytes[i] ;
    ptr_active -> type = 0X9 ;
    ptr_active -> channel = channel ;
    ptr_active -> note = event ;
    ptr_active -> velocity = velocity ;

    note = event ;
    add_note(note, velocity, key_sig, channel) ;

    if (use_sound) {
      if (channel == 9) {
        play_drum(note, velocity) ;
// DEBUG:
//fprintf(stderr, "Called play_drum with note = %d\n", note) ;
      }
      else {     
        play_note(note, velocity) ;
      }
    }
    on_time = present_time ;
  }

/* Check for velocity changes and create After touch events */
  if ( (event == FORTE) || (event == PIANO) || (event == INC_VELOCITY) || (event == DEC_VELOCITY) )   {
    old_velocity = velocity ;
    switch (event) {    
      case FORTE :
        velocity = VEL_FORTE ;
        break ;
      case PIANO :
        velocity = VEL_PIANO ;
        break ;
      case INC_VELOCITY :
        velocity = ((velocity + VEL_STEP) > 127) ? (velocity):(velocity + VEL_STEP) ;
        break ;
      case DEC_VELOCITY :
        velocity = ((velocity - VEL_STEP) < 0  ) ? (velocity):(velocity - VEL_STEP) ;
        break ;
      default:
        ;
    }
    slider(VELOCITY_LINE, old_velocity, velocity) ;
  
/* Execution is diverted back to the top of the while loop if silent */
    if ((note == 0) && (!harmony))  {
      continue ;
    }

/*  if a note is sounding, write an aftertouch event */
    else {
      delta_time = present_time - prior_time ;
      prior_time = present_time ;

      ptr_active = add_node(ptr_active) ;
      if (harmony) {
        count_bytes = count_bytes + write_var_bytes(delta_time, var_bytes) + 2 ;
        for (i = 0 ; i < 4 ; ++i)
          ptr_active -> delta_time[i] = var_bytes[i] ;
        ptr_active -> type = 0XD ;       /* channel aftertouch */
        ptr_active -> channel = channel ;
        ptr_active -> velocity = velocity ;
      }
      else {
        count_bytes = count_bytes + write_var_bytes(delta_time, var_bytes) + 3 ;
        for (i = 0 ; i < 4 ; ++i)
          ptr_active -> delta_time[i] = var_bytes[i] ;
        ptr_active -> type = 0XA ;       /* note aftertouch */
        ptr_active -> channel = channel ;
        ptr_active -> note = note ;
        ptr_active -> velocity = velocity ;
      }

/* Give feedback on velocity for each note sounding */
      for ( j = 0 ; j < 128 ; ++j) {
/* Skip ahead if the note isn't sounding */
        if (!active_notes[j]) continue ;

        del_note(j, old_velocity, channel) ;
        add_note(j, velocity, key_sig, channel) ;
        if (use_sound) adjust_volume(j, velocity) ;
      } /* Next j */
    } /* End If (!harmony && note == 0) */
  } /* End If aftertouch */

/* Process and write tempo events */
  if ( (event == INC_TEMPO) || (event == DEC_TEMPO) )  {
    old_bpm = bpm ;
    if (event == INC_TEMPO)
      bpm = ((bpm + BPM_STEP) > MAX_BPM) ? (bpm) : (bpm + BPM_STEP) ;
    if (event == DEC_TEMPO)
      bpm = ((bpm - BPM_STEP) < MIN_BPM) ? (bpm) : (bpm - BPM_STEP) ;
    slider(TEMPO_LINE, old_bpm, bpm) ;
    tempo = MICROSECONDS_PER_MINUTE / bpm ;

    delta_time = present_time - prior_time ;
    prior_time = present_time ;
 
    ptr_active = add_node(ptr_active) ;
    count_bytes = count_bytes + write_var_bytes(delta_time, var_bytes) + 6 ;
    for (i = 0 ; i < 4 ; ++i)
      ptr_active -> delta_time[i] = var_bytes[i] ;
    ptr_active -> type = 0X51 ;
    ptr_active -> channel = (tempo >> 16) & 0XFF ;
    ptr_active -> note = (tempo >> 8) & 0XFF ;
    ptr_active -> velocity = (tempo) & 0XFF ;

    msec_tempo_change = msec_elapsed ;
    reset_tempo = TRUE ;
  } /* End If tempo */

/* process legato or staccato events */
  if ( (event == LEGATO) || (event == STACCATO) ) {
    old_legato = legato ;
    if (event == LEGATO) 
      legato = ( (legato * ((float)LEG_PERCENTAGE / 100.0 )) > 1.0) ? (legato) : (legato * ((float)LEG_PERCENTAGE / 100.0 )) ;
    if (event == STACCATO) 
      legato = ( (legato / ((float)LEG_PERCENTAGE / 100.0 )) < 0.5) ? (legato) : (legato / (float)(LEG_PERCENTAGE / 100.0) ) ;
    sliderf(LEGATO_LINE, old_legato, legato) ;
/* this is not a MIDI event, and there is nothing to put in the file */
  }

/* Termination */
  if (event == QUIT) {

/* shut off any notes still playing */
    for (j = 1 ; j < 128 ; ++j) {
      if (!active_notes[j]) continue ;

      delta_time = present_time - prior_time ;
      prior_time = present_time ;
      
      ptr_active = add_node(ptr_active) ;
      count_bytes = count_bytes + write_var_bytes(delta_time, var_bytes) + 3 ;
      for ( i = 0 ; i < 4 ; ++i) 
        ptr_active -> delta_time[i] = var_bytes[i] ;
      ptr_active -> type = 0X8 ;
      ptr_active -> channel = channel ;
      ptr_active -> note = j ;
      ptr_active -> velocity = 0 ;        /* this is the value used by the Yamaha PSR-410 */

      del_note(j, velocity, channel) ;
      if (use_sound) 
        hush_note(j) ;
      note = 0 ;
    } /* Next j */

/* Leave the main loop */
    break ;
  } /* End If termination */

}                /* Wend main loop */

end_timing() ;
return count_bytes ;
}                   /* End Function record */

